//AnsiString Class
//WARNING!
//If you are in Linux, then define LINUX=1 in compiler (-D Linux)
//If you are in Windows, do not define LINUX directive!
#ifndef AnsiStringClass
#define AnsiStringClass
#ifdef LINUX
    #include <stdio.h>
    #include <stdlib.h>
    #include <errno.h>
    #include <unistd.h>
    #include <signal.h>
    #include <string.h>
    typedef unsigned long long int QWORD;
#else
    #include <windows.h>
#endif
#include <string.h>
#include <stdio.h>

#define AnsiStringMAXLEN 500
#define NumberConvertException 10
#define Exception int

//typedef unsigned __int64 QWORD;

#ifndef LINUX
static HWND Owner=NULL;
#endif

class AnsiString{
  public:
      int ConvertToInt();
      double ConvertToDouble();
      char NumberConvError;
      char String[AnsiStringMAXLEN];
      AnsiString();
#ifndef LINUX
	  AnsiString(HWND Hwnd);
#endif
      AnsiString(const char* str);
      ~AnsiString();
      AnsiString operator=(const AnsiString &str);
      AnsiString operator=(const char * str);
      AnsiString operator+(const AnsiString &str);
      AnsiString operator+(const char * str);
	  int operator==(const AnsiString &str);
	  int operator==(const char * str);
      char& operator[] (unsigned i);
      int AnsiCompare( const AnsiString & str);
      int AnsiCompare( const char * str);
	  int AnsiCompareIC( const AnsiString & str);
      char* c_str();
      int IsEmpty();
      int Length();
      int Pos(const AnsiString & str);
      int Pos(const char * str);
      int AnsiPos(const AnsiString & str);
      int AnsiPos(const char * str);
      int ToInt();
      double ToDouble();
      AnsiString Delete(int index, int count);
      AnsiString Insert(const AnsiString& str, int index);
      AnsiString Insert(const char * str, int index);
      AnsiString SubString (int index, int count);
      void IntToStr(int x);
      void FloatToStr(double x);
      void Show();
	  void ShowT(char* Title);
	  void ShowT(AnsiString Title);
      void LoadStr(const char* str);
      void LoadStr(const AnsiString &str);
	  void UpperCase();
	  bool Load(AnsiString &fname);
	  bool Save(AnsiString &fname);
	  bool SaveAdd(AnsiString &fname);
	  bool Load(char* fname);
	  bool Save(char* fname);
	  bool SaveAdd(char* fname);
	  void ShowErrorMsg(AnsiString Title);
#ifndef LINUX
	  void Sprintf(char*fmt,int n);
	  void Sprintf(char*fmt,float n);
	  void Sprintf(char*fmt,double n);
	  void Sprintf(char*fmt,char* str);
	  void Sprintf(char*fmt,AnsiString& str);
#endif
	  void LowerCase();
};


//Linux specific functions
#ifdef LINUX
typedef void (*sigfunc)(int, siginfo_t *, void *);
//initialize signal sending that will be received in main thread to print out error message
//SendFailure will print out the error message
static void InitErrorSignal(sigfunc sfc){
    struct sigaction act;
	memset (&act, 0, sizeof(act));
	act.sa_sigaction = sfc;
	act.sa_flags = SA_SIGINFO;
	if (sigaction(SIGUSR2, &act, NULL) < 0) {
		printf ("FATAL: sigaction ERROR\n");
		exit(0);
	}
}
//Print out error message. tis function has a Windows and a Linux version
static void ErrorAlert(char* msg,char* title){
    AnsiString* err=new AnsiString;
    sprintf(err->String,"%s - %s",msg,title);
	union sigval value;
	value.sival_ptr=(void*)err;
	if(sigqueue(getpid(), SIGUSR2, value)==-1){
		//printf("ERROR sigqueue\n");
		return;
	}
}
#endif

//Windows specific functions
#ifndef LINUX
static void ShowMessage(AnsiString str){
     str.Show();
}

static void ShowMessage(char* str){
     AnsiString strR(str);
     strR.Show();
}

static void ShowMessageT(char* str,char* Title){
     AnsiString strR(str);
     strR.ShowT(Title);
}

static void ShowMessageT(char* str,AnsiString Title){
     AnsiString strR(str);
     strR.ShowT(Title);
}

static void ShowError(char* msg,char* title){
	AnsiString Title(title);
	AnsiString MSG(msg);
	MSG.ShowErrorMsg(Title);
}

static bool YesNo(char* str){
	if(MessageBox(NULL,str,"Question",MB_YESNO) == IDYES)
		return true;
	else
		return false;
}

static bool YesNo(AnsiString str){
	return YesNo(str.String);
}

static bool OkCancel(char* str){
	if(MessageBox(NULL,str,"Question",MB_OKCANCEL|MB_ICONINFORMATION)==IDOK)
		return true;
	else
		return false;
}

static bool OkCancel(AnsiString str){
	return OkCancel(str.String);
}

//Gets file size: returns -1 if file size> 32000byte
static QWORD FileSize(char* fname){
	DWORD sHi,sLo;
	QWORD size;
	
	HANDLE hFile=CreateFile(fname,GENERIC_READ|GENERIC_WRITE,0,NULL,OPEN_EXISTING,0,NULL);
	if(!hFile){
		return 0;
	}
	sLo=GetFileSize(hFile,&sHi);
	CloseHandle(hFile);
	
	size=sLo+sHi<<32;
	
	if(GetLastError()){
		return 0;
	}else{
		return size;
	}
}

static QWORD FileSize(AnsiString fname){
	return FileSize(fname.String);
}

//for random number generation
static int time(int*p){
	SYSTEMTIME time;
	int T;
	GetLocalTime(&time);
	T=time.wYear;
	T+=time.wMonth;
	T+=time.wDay;
	T+=time.wHour;
	T+=time.wMinute;
	T+=time.wSecond;
	return T;
}

//Print out error message. tis function has a Windows and a Linux version
static void ErrorAlert(char* msg,char* title){
    MessageBox(Owner, msg, title, MB_APPLMODAL | MB_ICONSTOP);
}
#endif


//Common functions in Linux and in Windows

static AnsiString ExtractFilePath(AnsiString& str){
	AnsiString f=str;
	int x;
	for(x=f.Length()-1;x>=0;x--){
		if((f.String[x]=='\\')||(f.String[x]=='/'))break;
	}
	if(x<0){
		f.String[0]=0;
		return f;
	}
	if(x==f.Length()-1)return f;
	f.String[x+1]=0;
	return f;
}

static AnsiString ExtractFileName(AnsiString& str){
	int x;
	AnsiString f=str;
	for(x=f.Length()-1;x>=0;x--){
		if((f.String[x]=='\\')||(f.String[x]=='/'))break;
	}
	if(x<0)return f;
	strcpy(f.String,f.String+x+1);
	return f;
}

static AnsiString ChangeFileExt(AnsiString& str,AnsiString& Ext){
	int x;
	AnsiString f=str;
	for(x=f.Length()-1;x>=0;x--){
		if(f.String[x]=='.')break;
	}
	if(x<0){
		f=f+"."+Ext;
		return f;
	}
	strcpy(f.String+x+1,Ext.String);
	return f;
}

static AnsiString ChangeFileExt(AnsiString& str,char* Ext){
	AnsiString EXT(Ext);
	return ChangeFileExt(str,EXT);
}

static bool FileExists(char* fname){
	FILE*f;
	f=fopen(fname,"r");
	if(f==NULL){
		return false;
	}else{
		fclose(f);
		return true;
	}
}

static bool FileExists(AnsiString fname){
	return FileExists(fname.String);
}

#endif
