//MainForm
#ifndef MainFormH
#define MainFormH

#include <vector>

#include "..\Client.h"
#include "..\FormClass.h"
#include "..\ButtonClass.h"
#include "..\LabelClass.h"
#include "..\CanvasClass.h"
#include "..\TimerClass.h"
#include "..\TabControl.h"
#include "..\TScrollBar.h"
#include "..\MemoClass.h"
#include "..\EditClass.h"
#include "..\OpenFileClass.h"
#include "resource.h"

#include "MainFormParams.h"

//ENable GUI related popup error messages
//#define GUIERR

//Animation changeist file
#define CHANGELISTFILE "Anim.txt"
#define CHANGELISTFILEDEACT "_Anim.txt"

struct GIF{
	int x;
	int y;
	int phnum;
	int phact;
	char name[4];
	int valid;
	AnsiString* kep[50];
};

struct Session;

struct Partner{
	int ID;
	AnsiString name;	//name
	unsigned char* pic;	//picture data
	int online;			//online/offline
	Session* ses;
	int IconPos;		//Icon Y position
	int cskey;			//send key
	int crkey;			//receive key
};

struct Session{
	int tab;			//session opened on which tab
	unsigned char* win;	//window bmp data
	std::vector<GIF*> Objlist;
	Partner* part;
	int CX;				//Cursor
	int CY;
	int valid;
	char* buf;			//last 2048 char
	int bpoi;			//buffer pointer
	int blink;			//blinking the tab
	int scnt;			//counter for re-sending the key
	int rscnt;			//resending counter on failures
};

//size of Main bitmap
const int MBW=300;
const int MBH=600;//300;
//size of Editor bitmap
const int EBW=MBW;
const int EBH=200;//100;
//size of viewport of Main bitmap
const int MBVX=50;
const int MBVY=60;
const int MBVW=MBW;
const int MBVH=300;//MBH;
//size of viewport of Editor bitmap
const int EBVX=50;
const int EBVY=370;
const int EBVW=EBW;
const int EBVH=100;//EBH;
//Viewports
static int MainViewY=0;//MBH-MBVH;	//Main viewport start Y coordinate
static int EditViewY=0;//EBH-EBVH;	//Edit viewport start Y coordinate
//Color of main window
static int mColor=0xFFF0F0;
//Color of editor window
const int eColor=0xF0FFF0;
//Frame picture sizes
const int FRX=0;
const int FRY=20;
const int FRW=410;
const int FRH=500;
//Partner picture sizes
const int PW=50;
const int PH=50;
const int PSOR=MBVY;			//partnerek rajzolsnak kezdsora
const int PSORMAX=PSOR+MBVH-20;	//partnerek rajzolsnak vgsora
const int POSZ=100;				//partnerek rajzolsnak oszlopa

static volatile int extrarow=0;
static volatile int intimer=1;
static volatile int inOnChar=0;
static volatile int CTRLpressed=0;
static std::vector<AnsiString*> Frames;
static std::vector<Partner*> Partners;
static std::vector<Session*> Sessions;
static unsigned char* DefPartnerPic;
static long DefPartnerPicSize;

//Felugr dinamikus editboxok
static int M_edit=0;
static int E_edit=0;
static int EditMParam[4]={MBVX,MBVY,MBVW,MBVH};
static int EditEParam[4]={EBVX,EBVY,EBVW,EBVH};

class MainForm : FormClass, Client {
public:
	MainForm();
	~MainForm();
	void BuildForm(HWND hwnd,HINSTANCE hInst);
	void DestroyForm();
	void OnPaint();
	void OnCommand(WORD command);
	void OnTimer(unsigned TimerID);
	void OnChar(char ch);
	void OnKeyDown(WPARAM w);
	void OnKeyUp(WPARAM w);
	void OnNotify(NMHDR*notif);
	void Vscroll(DWORD s_info,DWORD s_handle);
	void OnMouseDbClick(int X,int Y);
	void OnMouseWheel(int xPos,int yPos,WORD delta,WORD buttons);
	void OnMouseButton(int xPos,int yPos);
	void OnMouseRightBtn(int X,int Y);
	void ButtonXClick();
	void ButtonF1Click();
	void ButtonF2Click();
	void ButtonH1Click();
	void ButtonH2Click();
	void ButtonKepClick();
	
	HINSTANCE hinst;
	
	TimerClass* Timer1;
	TabControl* TabControl1;
	CanvasClass* Canvas;
	TScrollBar* TScrollBar1;
	TScrollBar* TScrollBar2;
	
	Button*  ButtonX;
	
	//httr<>
	Label* LabelH;
	Button* ButtonH1;
	Button* ButtonH2;
	//Font <>
	Label* LabelF;
	Button* ButtonF1;
	Button* ButtonF2;
	//Add partner
	Label* LabelAddP;
	EditBox* EditAddP;
	//kpcsere
	Button* ButtonKep;
	
	//F megjelent bitmap
	HDC		Mhdc;
	HBITMAP Mhbmp;
	void* 	DIBport;
	//szerkeszt bitmap
	HDC		MhdcSz;
	HBITMAP MhbmpSz;
	void* 	DIBportSz;
	//Small partner icon clipping bitmap
	HDC		MhdcClip;
	HBITMAP MhbmpClip;
	void* 	DIBportClip;
	//Double buffering bitmap
	HDC		hdcDoubleBuf;
	HBITMAP bmpDoubleBuf;
	//Dinamic Edit / memo area
	Memo*	EditM;
	Memo*	EditE;
	//File send selector box
	TOpenDialog *OpenDialog1;
	TOpenPictureDialog* Openpic;
	
	//Used Fonts
	HFONT SysFont;		//Used for communicating with server
	HFONT ErrorFont;	//Error messages
	HFONT mFont1;
	HFONT mFont2;
	HFONT mFont3;
	HFONT mFont4;
	HFONT mFont5;
	HFONT mFont6;
	HFONT mFont7;
	HFONT mFont8;
	HFONT mFont9;
	HFONT mFont10;
	HFONT NameF;

private:	// User declarations
	int mcx,mcy;	//main cursors
	int ecx,ecy;	//edit cursors
	int vecx,vecy;	//visible edit cursors
	int echar;		//Character to be edited in editor
	void PutStrToMain(char*str);
	void PutCharToMain(char c);
	void PutObjectToMain(char* obj);
	void DrawGIF();
	void ScrollGIFs(int move);
	void PutCharToEdit(char c);
	char Editbuf[1024];
	int Epoi;
	void FlushEdit();
	void ClearEdit();
	int cursorState;
	void ShowEditCursor();
	void HideEditCursor();
	void ComputeVisibleCursor();
	void CopyFromClipboard();
	void SetFrame();
	void SwitchToTab(int ToTAB);
	void SwitchToTabHidden(int ToTAB);
	//Create new partner - session
	Session* ActSes;	//TAB-al egytt redundns!
	void AddSession(char* name,int DRAW);
	void AddSession(Partner* p,int DRAW);
	void AddSessionSpec(Partner* p,int DRAW);
	void RemoveSession();
	void ResyncAllSession();
	Partner* AddPartner(int id,char* pname,int active);
	Partner* GetPartner(int id);
	Partner* GetPartner(char* Name);
	//Draw main panel with partners
	void DrawMainPanel();
	void DrawMainPanelUp();
	void HideMainPanel();
	void DrawPartnerIcon(Partner*p,int sor);
	void SetMainFont(int f);
	void SetEditFont();
	void ChangeEditFont();
	void ShowSetFont();
	int  IsBMPfile(char* fname);
	int SpecifyINIfile();
	
	BYTE* ConvertRGBToBMPBuffer ( BYTE* Buffer, int width, int height, long* newsize );
	BYTE* ConvertBMPToRGBBuffer ( BYTE* Buffer, int width, int height );
	
	//GIF list
	AnsiString GIFRoot;
	AnsiString FrameRoot;
	//Ini file
	AnsiString INIfile;
	
	int ScrollPos;
	int TAB;	//ActSes -el egytt redundns!
	
	int CreateClientCalled;
	int RouteEnabled;
	int InRepaint;
	int Blinking;	//application caption is blinking
	int Blinkstate;
	int AuthPassed;
	int SpecialGUI;	//This for drawing with PutStrToMain without refreshing GUI
	int TryReconnect;	//time counter for reconnecting network if connection lost
	
	//server alive checking
	int AliveCnt;
	int AliveRet;
	
	//flood protection and GUI overlapping protection
	Semaphore* GUIsema;
	void EnterGUICritical();
	void LeaveGUICritical();
	
public:
	//INTERFACE to network (Client)
	void GuiPrint(char* message);
	void ProcessPartner(char*uname, int id, int ON_OFF, unsigned char* photo);	//process partner info loaded from server
	void NotifAuthPassed();
	void NotifUserOffline();
	void NotifUnameChg(int ID,char*name);
	void NotifUremoved(int ID);
	void NotifUstatus(int ID,int state);
	void RouteNotifMsg(int From,int Font,char*message);		//Route command from another client
	void GuiUpdateNotif();
	void NetDisconnectNotif();
	void ServerAlive();		//Echo for alive command
	//key setter, getter functions for ClientApp
	int& getCSkey(int ID);
	int& getCRkey(int ID);
	int  setCSRkey(int KEY,int ID);
	void NotitErrorCode(int From,int ecode);
	void FileHasSentNotif(int ID,int res);
	void FileReceivingNotif(int F,char*name);
	void FileReceivedNotif(int F,char*name);
	//----REQUESTS TO SERVER-----
	void SendScrMsgToServer(char*msg);
	void ReqSendFileToClient();
	void ServerReqAddPartner(AnsiString part);
	void ServerReqChangePic();
	int  AdminRequest(char* adm);
	
	void BlinkingTab(Session*ses);
	void BlinkingOFF();
	void Blink1();
	void Blink2();
	
	//GIF aniation handling
	char*sch;
	void LoadChangeFile();
	void ChangeToAnim();
	void SetAnimate();
	char** chnum;
	char** chstr;
	int chlen;
	
	//scheduled automatic GUI updating counter
	int AutGuiUpdt;
};

#endif
