#include "Client_Linux.h"

Client::Client(){
	conn=INVALID_SOCKET;
	level=0;
	FileSendError=0;
	ClientThreadRun=0;
	FsendThreadRun=0;
	ClientThread.v=0;
	FsendThread.v=0;
	Fsend[0]=0;
	NEWKEY=0;
	int res=pthread_mutex_init(&clientCriticalSection, NULL);
	if(res){
		ClientPrintf("FATAL: pthread_mutex_init(1) failed with %d\n",res);
		exit(0);
	}
	res=pthread_mutex_init(&CriticalSection, NULL);
	if(res){
		ClientPrintf("FATAL: pthread_mutex_init(2) failed with %d\n",res);
		exit(0);
	}
	recbuf= new unsigned char[4096];
	sendbuf= new unsigned char[4096];
	filesendbuf=NULL;
	//ini file name
	AnsiString PATH=GetPath();
	AnsiString FILENAME=INIFILENAME;
	INIFILE=PATH+FILENAME;
}

Client::~Client(){
	if(recbuf)
		delete[] recbuf;
	recbuf=NULL;
	if(sendbuf)
		delete[] sendbuf;
	sendbuf=NULL;
	int res=pthread_mutex_destroy(&clientCriticalSection);
	if(res){
		ClientPrintf("pthread_mutex_destroy(1) failed with %d\n",res);
	}
	res=pthread_mutex_destroy(&clientCriticalSection);
	if(res){
		ClientPrintf("pthread_mutex_destroy(2) failed with %d\n",res);
	}
	if(filesendbuf)
		delete[] filesendbuf;
}

int Client::CreateClient(){
	if(i==NULL)
		i=new Initializable();
	AdminPrintf("Starting admin client\n");

	//Connect to server
	Connect(INIFILE.String);
	if(!m_bIsConnected){
		ClientPrintf("ERROR: create() failed\n");
		return 0;
	}

	//Create Client thread
	AdminPrintf("Starting Client Thread\n");
	ClientThreadRun=1;
	int res=pthread_create( &ClientThread.t, NULL, MessageRecThread, (void*)this);
	if(res){
		ClientPrintf("\nFailed to create Client thread (%d)\n",res);
		return 0;
	}
	ClientThread.v=1;

	//First message to server
	sleep(1);
	SendHello();
	return 1;
}

void Client::DestroyClient(){
	//stop sending file
	ImmStopFsend=1;
	sleep(1);
	FsendThreadRun=0;
	FsendThGarbageColl();
	StrmIn.clear();
	//Exiting
	AdminPrintf("Exiting...\n");
	Disconnect();
	//Close thread
	AdminPrintf("Waiting for Client Thread to be terminated\n");
	ClientThreadRun=0;
	CloseThread(ClientThread, "ClientThread");
	AdminPrintf("Admin client successfully terminated\n");
	if(i){
		delete i;
		i=NULL;
	}
}

void Client::Connect(char* Inifile){
	AdminPrintf("Connect\n");
	m_bIsConnected = false;
	
	if(LoadFromIni(Inifile)){
		ClientPrintf("ERROR: ini file loading failed!\n");
		return;
	}
	AdminPrintf("Server IP=%s\n",i->sIpAddress);
	AdminPrintf("PORT=%d\n",i->iPort);
	AdminPrintf("ID=%d\n",i->ID);
	AdminPrintf("NAME=%s\n",i->Uname);
	
	//Initialize socket
	struct hostent *hp;
	unsigned int addr;
	struct sockaddr_in server;
	
	conn=socket(AF_INET,SOCK_STREAM,0);
	if(conn==INVALID_SOCKET){
		ClientPrintf("ERROR: INVALID_SOCKET\n");
		PrintError();
		return;
	}

	//NOW 'gethostbyname' is used instead of 'inet_addr' and 'gethostbyaddr'
	//addr=inet_addr(i->sIpAddress);
	//hp=gethostbyaddr((char*)&addr,sizeof(addr),AF_INET);
	hp=gethostbyname(i->sIpAddress);
	if(hp==NULL){
		ClientPrintf("ERROR gethostbyaddr FAILED\n");
		PrintError();
		close(conn);
		conn=INVALID_SOCKET;
		return;
	}

	server.sin_addr.s_addr=*((unsigned long*)hp->h_addr);
	server.sin_family=AF_INET;
	server.sin_port=htons(i->iPort);
	if(connect(conn,(struct sockaddr*)&server,sizeof(server))){
		ClientPrintf("ERROR connect() FAILED\n");
		PrintError();
		close(conn);
		conn=INVALID_SOCKET;
		return;	
	}
	m_bIsConnected = true;
	sKEY=0;
	rKEY=0;
#ifdef ADMN
	CSkey=0;
	CRkey=0;
#endif
	bufp=0;
	AuthOk=0;
	return;
}

void Client::Disconnect(){
	AdminPrintf("Disconnect\n");
	FsendThreadRun=0;
	CloseThread(FsendThread,"FsendThread");
	EnterCriticalSection();
	if(m_bIsConnected){
		ClientPrintf("Closing client socket\n");
		if(shutdown(conn, SHUT_RDWR)==-1){
			ClientPrintf("ERROR shutdown on Client socket\n");
			PrintError();
		}
		if(close(conn)==-1){
			ClientPrintf("ERROR closesocket on Client socket\n");
			PrintError();
		}
		conn=INVALID_SOCKET;
		m_bIsConnected=false;
	}else{
		ClientPrintf("Socket was not connected\n");
	}
	LeaveCriticalSection();
}

//Get working directory
AnsiString Client::GetPath(){
	AnsiString path;
    if(getcwd(path.String, 200)==NULL){
        PrintError("FATAL: getcwd() failed!\n");
        exit(0);
    }
    strcat(path.String,"/");
	return path;
}

void Client::CloseThread(THREAD Thread, char* Text){
	if(Thread.v==0){
		//AdminPrintf("%s is already closed\n",Text);
		return;
	}
	struct timespec tim;
	int ret;
	AdminPrintf("Closing %s\n",Text);
	if (clock_gettime(CLOCK_REALTIME, &tim) == -1) {
		ClientPrintf("ERROR getting Epoch time!\n");
		tim.tv_sec =0;
	}else{
		tim.tv_sec += 3;
	}
    ret=pthread_timedjoin_np( Thread.t, NULL, &tim);
	if(ret){
		ClientPrintf("ERROR closing %s\n",Text);
		ret=pthread_cancel(Thread.t);
		if(ret){
			ClientPrintf("ERROR cancelling %s\n",Text);
		}
	}
	Thread.v=0;
}

void Client::AdminPrintf(const char *fmt, ...){
#ifdef ADMN
	va_list va;
	va_start(va,fmt);
	char buff[500];
	vsprintf(buff,fmt,va);
	GuiPrint(buff);
	va_end(va);
#endif
}

void Client::ClientPrintf(const char *fmt, ...){
	va_list va;
	va_start(va,fmt);
	char buff[500];
	vsprintf(buff,fmt,va);
	GuiPrint(buff);
	va_end(va);
}

//convert MD5 code from integer array to string
void Client::MD5_IntToStr(unsigned char* m, char* s){
	int x;
	int w;
	for(x=0;x<16;x++){
		for(w=0;w<2;w++){
			switch( w? m[x]&0x0F : (m[x]&0xF0)>>4 ){
				case 0:
					s[x*2+w]='0';
					break;
				case 1:
					s[x*2+w]='1';
					break;
				case 2:
					s[x*2+w]='2';
					break;
				case 3:
					s[x*2+w]='3';
					break;
				case 4:
					s[x*2+w]='4';
					break;
				case 5:
					s[x*2+w]='5';
					break;
				case 6:
					s[x*2+w]='6';
					break;
				case 7:
					s[x*2+w]='7';
					break;
				case 8:
					s[x*2+w]='8';
					break;
				case 9:
					s[x*2+w]='9';
					break;
				case 10:
					s[x*2+w]='A';
					break;
				case 11:
					s[x*2+w]='B';
					break;
				case 12:
					s[x*2+w]='C';
					break;
				case 13:
					s[x*2+w]='D';
					break;
				case 14:
					s[x*2+w]='E';
					break;
				case 15:
					s[x*2+w]='F';
					break;
			}
		}
	}
	s[32]=0;
}

int Client::SendToServer(unsigned char* sMessage, int len){
	int iStat = 0;
	EnterCriticalSection();
	
#ifdef LOGSEND
	FILE*o;
	AnsiString PATH=GetPath();
	AnsiString FILENAME="sendlog.txt";
	AnsiString fname=PATH+FILENAME;
	o=fopen(fname.String,"a+");
	if(o){
		fprintf(o,"SENT(%d): ",len);
		for(int x=0;x<len;x++){
			fprintf(o,"%X ",sMessage[x]);
		}
		fprintf(o,"\n");
		fclose(o);
	}
#endif
	
	Crypt(SERVER_SCRIPT_KEY, sKEY, sMessage, len);
	if(conn!=INVALID_SOCKET)
		iStat = send(conn,(char*)sMessage,len,0);
	LeaveCriticalSection();
	if(iStat == -1){
		ImmStopFsend=1;
		ClientPrintf("ERROR: Send failed\n");
		PrintError();
		return 1;
	}
	return 0;
}

void Client::EnterSendCritical(){
	struct timespec tim;
	if (clock_gettime(CLOCK_REALTIME, &tim) == -1) {
		ClientPrintf("ERROR getting Epoch time!(1)\n");
		tim.tv_sec =0;
	}else{
		tim.tv_sec += 10;
	}
	int res=pthread_mutex_timedlock (&clientCriticalSection,&tim);
	if(res){
		if(res==ETIMEDOUT){
			ClientPrintf("EnterSendCritical timeout!\n");
		}else{
			ClientPrintf("pthread_mutex_lock(1) failed with %d\n",res);
		}
	}
}

void Client::LeaveSendCritical(){
	int res=pthread_mutex_unlock (&clientCriticalSection);
	if(res){
		ClientPrintf("pthread_mutex_unlock(1) failed with %d\n",res);
	}
}

void Client::EnterCriticalSection(){
	struct timespec tim;
	if (clock_gettime(CLOCK_REALTIME, &tim) == -1) {
		ClientPrintf("ERROR getting Epoch time!(2)\n");
		tim.tv_sec =0;
	}else{
		tim.tv_sec += 10;
	}
	int res=pthread_mutex_timedlock (&CriticalSection,&tim);
	if(res){
		if(res==ETIMEDOUT){
			ClientPrintf("EnterCriticalSection timeout!\n");
		}else{
			ClientPrintf("pthread_mutex_lock(2) failed with %d\n",res);
		}
	}
}

void Client::LeaveCriticalSection(){
	int res=pthread_mutex_unlock (&CriticalSection);
	if(res){
		ClientPrintf("pthread_mutex_unlock(2) failed with %d\n",res);
	}
}

void Client::SendHello(){
	unsigned char resp[]={0xBA,0,26,0,0,0,0,0,0,0,'H','e','l','l','o','M','e','s','s','e','n','g','e','r','B','A'};
	int x;
	for(x=10;x<26;x++){
		resp[x]^=0xA5;
	}
	SendToServer(resp,26);
}

//Send route command to server: route script to another client
void Client::CRouteScript(char* scr, int font, int TOC, int& CsKEY){
	EnterSendCritical();
	int n=strlen(scr);
	n+=15;
	sendbuf[0]=0xBA;
	sendbuf[1]=(n&0xFF00)>>8;
	sendbuf[2]=n&0xFF;
	sendbuf[4]=(i->ID&0xFF00)>>8;
	sendbuf[5]=i->ID&0xFF;
	sendbuf[6]=(TOC&0xFF00)>>8;
	sendbuf[7]=TOC&0xFF;
	sendbuf[8]=6;
	sendbuf[9]=0;
	sendbuf[10]=0;
	sendbuf[11]=font;
	sendbuf[12]=0;
	sendbuf[13]=0;
	strcpy((char*)(sendbuf+14),scr);
	sendbuf[n-1]=0;
	sendbuf[10]=Calculate_CRC8(sendbuf+10, n-10);
	Crypt(CLIENT_SCRIPT_KEY, CsKEY, sendbuf+10, n-10);
	sendbuf[3]=0;
	sendbuf[3]=Calculate_CRC8(sendbuf,n);
	SendToServer(sendbuf,n);
	LeaveSendCritical();
}

//Send route command to server: route key to another client
void Client::CRouteCmdKey(int TOC, int& CsKEY){
	EnterSendCritical();
	sendbuf[0]=0xBA;
	sendbuf[1]=0;
	sendbuf[2]=110;
	sendbuf[4]=(i->ID&0xFF00)>>8;
	sendbuf[5]=i->ID&0xFF;
	sendbuf[6]=(TOC&0xFF00)>>8;
	sendbuf[7]=TOC&0xFF;
	sendbuf[8]=6;
	sendbuf[9]=6;

	if(Random(sendbuf+10,100)){
		ClientPrintf("Random generator failure\n");
		LeaveSendCritical();
		return;
	}
	int KEY=CalcClientKey(sendbuf+10);
	sendbuf[3]=0;
	sendbuf[3]=Calculate_CRC8(sendbuf,110);
	SendToServer(sendbuf,110);
#ifdef ADMN
	CSkey=KEY;
	CRkey=KEY;
#else
	setCSRkey(KEY,TOC);
#endif
	LeaveSendCritical();
}

void Client::CRouteFileSend(int TOC, char* fname){
	if(FsendThreadRun){
		ClientPrintf("ERROR: only one file can be sent in the same time. Please wait for finishing.\n");
		return;
	}
	FsendThGarbageColl();
	
	if(strlen(fname)>98){
		ClientPrintf("ERROR: file name too long\nTl hossz a megadott fjlnv\n");
		return;
	}
	strcpy(Fsend,fname);
	FTOC=TOC;
	int res=pthread_create( &FsendThread.t, NULL, FileSendRunner, (void*)this);
	if(res){
		ClientPrintf("ERROR: File sending thread creation failed (%d)\n",res);
		FsendThreadRun=0;
		return;
	}
	FsendThread.v=1;
}

//File send thread garbage collection: closes the not running file sending thread
void Client::FsendThGarbageColl(){
	if(!FsendThreadRun){
		CloseThread(FsendThread,"FsendThread");
		sleep(1);
	}
}

//File sending thread function
void Client::FileSendRunTh(){
	FileSendBlockCnt=0;
	CAliveRaw();
	FILE*f;
	f=fopen(Fsend,"rb");
	if(f==NULL){
		ClientPrintf("ERROR: file can not be opened\n");
		return;
	}
	int cnt=0;
	char*C=Fsend;
	while(*C&&(*C!='/'))C++;
	if(*C==0){
		C=Fsend;
	}else{
		C=Fsend+strlen(Fsend);
		while(*C!='/')C--;
		C++;
	}
	filesendbuf=new unsigned char[4096];
	FsendStreamOpen(1,C);
	int w;
	
	while(!ImmStopFsend){
		pthread_testcancel();
		w=fgetc(f);
		if(feof(f))
			break;
		FsendStreamAdd(1,cnt,w);
		pthread_testcancel();
		if(ImmStopFsend)
			break;
	}
	
	FsendStreamClose(1,cnt);
	delete[] filesendbuf;
	filesendbuf=NULL;
	fclose(f);
	CAliveRaw();
	if(ImmStopFsend){
		AdminPrintf("File sending was stopped due to errors\n");
		FileHasSentNotif(FTOC,0);
	}else{
		AdminPrintf("File has been sent\n");
		FileHasSentNotif(FTOC,1);
	}
}

void Client::FsendStreamOpen(int Sord,char* Sname){
	EnterSendCritical();
	int n=strlen(Sname);
	n+=15;
	filesendbuf[0]=0xBA;
	filesendbuf[1]=(n&0xFF00)>>8;
	filesendbuf[2]=n&0xFF;
	filesendbuf[4]=(i->ID&0xFF00)>>8;
	filesendbuf[5]=i->ID&0xFF;
	filesendbuf[6]=(FTOC&0xFF00)>>8;
	filesendbuf[7]=FTOC&0xFF;
	filesendbuf[8]=6;
	filesendbuf[9]=7;
	filesendbuf[10]=0;
	filesendbuf[11]=1;
	filesendbuf[12]=0;
	filesendbuf[13]=Sord;
	strcpy((char*)(filesendbuf+14),Sname);
	filesendbuf[n-1]=0;
	filesendbuf[10]=Calculate_CRC8(filesendbuf+10, n-10);
#ifdef ADMN
	Crypt(CLIENT_SCRIPT_KEY, CSkey, filesendbuf+10, n-10);
#else
	Crypt(CLIENT_SCRIPT_KEY, getCSkey(FTOC), filesendbuf+10, n-10);
#endif
	filesendbuf[3]=0;
	filesendbuf[3]=Calculate_CRC8(filesendbuf,n);
	SendToServer(filesendbuf,n);
	LeaveSendCritical();
}

void Client::FsendStreamAdd(int Sord,int& CNT,int data){
	EnterSendCritical();
	if(CNT==0){
		filesendbuf[0]=0xBA;
		filesendbuf[1]=0x10;
		filesendbuf[2]=0x00;
		filesendbuf[4]=(i->ID&0xFF00)>>8;
		filesendbuf[5]=i->ID&0xFF;
		filesendbuf[6]=(FTOC&0xFF00)>>8;
		filesendbuf[7]=FTOC&0xFF;
		filesendbuf[8]=6;
		filesendbuf[9]=8;
		filesendbuf[10]=0;
		filesendbuf[11]=1;
		filesendbuf[12]=0;
		filesendbuf[13]=Sord;
		CNT=14;
	}
	filesendbuf[CNT++]=data;
	if(CNT>=4096){
		filesendbuf[10]=0;
		filesendbuf[10]=Calculate_CRC8(filesendbuf+10, 4086);
#ifdef ADMN
		Crypt(CLIENT_SCRIPT_KEY, CSkey, filesendbuf+10, 4086);
#else
		Crypt(CLIENT_SCRIPT_KEY, getCSkey(FTOC), filesendbuf+10, 4086);
#endif
		filesendbuf[3]=0;
		filesendbuf[3]=Calculate_CRC8(filesendbuf,4096);
		SendToServer(filesendbuf,4096);
		
		FileSendBlockCnt++;
		CNT=0;
	}
	LeaveSendCritical();
	
	if(FileSendBlockCnt>100){
		CAliveRaw();
		FileSendBlockCnt=0;
	}
}

void Client::FsendStreamClose(int Sord,int& CNT){
	EnterSendCritical();
	if(CNT){
		if(CNT>4096)CNT=4096;
		filesendbuf[1]=(CNT&0xFF00)>>8;
		filesendbuf[2]=CNT&0xFF;
		filesendbuf[10]=0;
		filesendbuf[10]=Calculate_CRC8(filesendbuf+10, CNT-10);
#ifdef ADMN
		Crypt(CLIENT_SCRIPT_KEY, CSkey, filesendbuf+10, CNT-10);
#else
		Crypt(CLIENT_SCRIPT_KEY, getCSkey(FTOC), filesendbuf+10, CNT-10);
#endif
		filesendbuf[3]=0;
		filesendbuf[3]=Calculate_CRC8(filesendbuf,CNT);
		SendToServer(filesendbuf,CNT);
		CNT=0;
	}
	filesendbuf[0]=0xBA;
	filesendbuf[1]=0;
	filesendbuf[2]=14;
	filesendbuf[4]=(i->ID&0xFF00)>>8;
	filesendbuf[5]=i->ID&0xFF;
	filesendbuf[6]=(FTOC&0xFF00)>>8;
	filesendbuf[7]=FTOC&0xFF;
	filesendbuf[8]=6;
	filesendbuf[9]=9;
	filesendbuf[10]=0;
	filesendbuf[11]=1;
	filesendbuf[12]=0;
	filesendbuf[13]=Sord;
	filesendbuf[10]=Calculate_CRC8(filesendbuf+10, 4);
#ifdef ADMN
	Crypt(CLIENT_SCRIPT_KEY, CSkey, filesendbuf+10, 4);
#else
	Crypt(CLIENT_SCRIPT_KEY, getCSkey(FTOC), filesendbuf+10, 4);
#endif
	filesendbuf[3]=0;
	filesendbuf[3]=Calculate_CRC8(filesendbuf,14);
	SendToServer(filesendbuf,14);
	LeaveSendCritical();
}

//Send error to another client
void Client::CRouteError(int TOC, int error){
	EnterSendCritical();
	sendbuf[0]=0xBA;
	sendbuf[1]=0;
	sendbuf[2]=12;
	sendbuf[4]=(i->ID&0xFF00)>>8;
	sendbuf[5]=i->ID&0xFF;
	sendbuf[6]=(TOC&0xFF00)>>8;
	sendbuf[7]=TOC&0xFF;
	sendbuf[8]=6;
	sendbuf[9]=10;
	sendbuf[10]=(error&0xFF00)>>8;
	sendbuf[11]=error&0xFF;
	
	sendbuf[3]=0;
	sendbuf[3]=Calculate_CRC8(sendbuf,12);
	SendToServer(sendbuf,12);
	LeaveSendCritical();
}

//Send Alive to server
void Client::CAliveRaw(){
	if( !i || !sendbuf )
		return;
	EnterSendCritical();
	sendbuf[0]=0xBA;			//start
	sendbuf[1]=0;			//size low
	sendbuf[2]=20;			//size high
	sendbuf[4]=(i->ID&0xFF00)>>8;	//from
	sendbuf[5]=i->ID&0xFF;			//from
	sendbuf[6]=0;					//to
	sendbuf[7]=0;					//to
	sendbuf[8]=10;	//type
	sendbuf[9]=1;	//subtype
	//ballast content (for security)
	Random(sendbuf+10,10);
	sendbuf[3]=0;
	sendbuf[3]=Calculate_CRC8(sendbuf,20);
	SendToServer(sendbuf,20);
	LeaveSendCritical();
}

//Send Alive to server
void Client::CAlive(){
	if(FsendThreadRun)
		return;
	CAliveRaw();
}

int Client::RecMessagePort(){
	int len;
	int size;
	int x;

	len = recv(conn,(char*)(recbuf+bufp),4096-bufp,0);
	if(len <= 0){
		return 1;
	}else{
		Encrypt(SERVER_SCRIPT_KEY, rKEY, (unsigned char*)(recbuf+bufp), len);
		
#ifdef LOGREC
		FILE*o;
		AnsiString PATH=GetPath();
		AnsiString FILENAME="reclog.txt";
		AnsiString fname=PATH+FILENAME;
		o=fopen(fname.String,"a+");
		unsigned char*w=(recbuf+bufp);
		if(o){
			fprintf(o,"REC(%d): ",len);
			for(int x=0;x<len;x++){
				fprintf(o,"%X ",w[x]);
			}
			fprintf(o,"\n");
			fclose(o);
		}
#endif

		bufp+=len;
		while(bufp){
			if(recbuf[0]!=(unsigned char)0xBA){
				ClientPrintf("ERROR 0xBA missing from header start\n");
				bufp=0;
				break;
			}
			size=recbuf[1]*256+recbuf[2];
			if((size<9)||(size>4096)){
				ClientPrintf("ERROR message size is out of range\n");
				bufp=0;
				break;
			}
			if(bufp<size)
				break;
			if(ProcSrvMessage(size)){
				bufp=0;
				break;
			}
			for(x=0;x<4096-size;x++)
				recbuf[x]=recbuf[x+size];
			bufp-=size;
			if(bufp<0)bufp=0;
		}
	}
	return 0;
}

int Client::ProcSrvMessage(int size){
	int x;
	//check message header
	BYTE CRC=recbuf[3]; recbuf[3]=0;
	BYTE CRCexp=Calculate_CRC8(recbuf, size);
	if(CRC!=CRCexp){
		ClientPrintf("CRC error in received message\n");
		return 1;
	}
	int FromID=recbuf[4]*256+recbuf[5];
	int ToID=  recbuf[6]*256+recbuf[7];
	int type=recbuf[8];
	int subtype=recbuf[9];
	if(AuthOk&&( !i || (i->ID!=ToID) )){
		ClientPrintf("ToID is wrong, message sent to wrong client\n");
		return 1;
	}
	if(FromID==0){
		//server message
		switch(type){
			//Question RND
			case 1:
				if(AuthOk)
					break;
				resp[0]=0xBA;
				resp[1]=0;
				resp[2]=40;
				resp[3]=0;
				resp[4]=0;
				resp[5]=0;
				resp[6]=0;
				resp[7]=0;
				resp[8]=2;
				resp[9]=0;
				NEWKEY=KeyGen(resp+10, recbuf+10);

				resp[3]=Calculate_CRC8(resp, 40);
				SendToServer(resp,40);
				break;
			//Auth yourself
			case 3:
				if(AuthOk)
					break;
				for(x=10;x<22;x++)
					recbuf[x]^=0x5A;
				recbuf[22]=0;
				if(strcmp((char*)(recbuf+10),"AuthYourself")){
					ClientPrintf("AuthYourself message failed from server\n");
					break;
				}
				//change KEY
				sKEY=NEWKEY;
				rKEY=NEWKEY;
				//send auth
				resp[0]=0xBA;
				resp[1]=0;
				resp[2]=47;
				resp[3]=0;
				resp[4]=(i->ID&0xFF00)>>8;
				resp[5]=i->ID&0xFF;
				resp[6]=0;
				resp[7]=0;
				resp[8]=4;
				resp[9]=0;
				for(x=0;x<21;x++)
					resp[x+10]=i->Uname[x];
				for(x=0;x<16;x++)
					resp[x+31]=i->Upass[x];
				resp[3]=Calculate_CRC8(resp, 47);
				SendToServer(resp,47);
				break;
			//Plist
			case 5:
				ProcessPartner((char*)(recbuf+10),(int)(recbuf[31]*256+recbuf[32]),(int)recbuf[33],(unsigned char*)(recbuf+34));
				break;
			//notification
			case 8:
				if(ToID!=i->ID)
					break;
				if(subtype==1){
					//error occured
					x=recbuf[10]*256+recbuf[11];
					switch(x){
						case 100:
							ImmStopFsend=1;
							ClientPrintf("Error: server full, try to connect later\n");
							break;
						case 200:
							ImmStopFsend=1;
							ClientPrintf("ERROR user is already logged in. Doubled auth. not enabled\n");
							break;
						case 250:
							ImmStopFsend=1;
							ClientPrintf("CRC ERROR returned from server\n");
							break;
						case 251:
							ImmStopFsend=1;
							ClientPrintf("Invalid request error returned from server\n");
							break;
						case 252:
							ImmStopFsend=1;
							ClientPrintf("Invalid FromID or ToID\n");
							break;
						case 260:
							ClientPrintf("Invalid Admin request\n");
							break;
						case 261:
							ClientPrintf("User name requested is not existing\n");
							break;
						case 262:
							ClientPrintf("Invalid user ID requested\n");
							break;
						case 263:
							ClientPrintf("Admin level authentication failed. Access refused\n");
							break;
						case 264:
							ClientPrintf("Admin request contains invalid element\n");
							break;
						case 265:
							ClientPrintf("Admin request failed due to server errors\n");
							break;
						case 266:
							ClientPrintf("File upload error to server\n");
							FileSendError=1;
							break;
						default:
							ImmStopFsend=1;
							ClientPrintf("Error occured, code from server=%d\n",x);
							break;
					}
				}else if(subtype==2){
					//route enabled
					NotifAuthPassed();
					AuthOk=1;
				}else if(subtype==3){
					ImmStopFsend=1;
					AdminPrintf("Routing failed: user is offline\n");
					NotifUserOffline();
				}else if(subtype==4){
					ToID=recbuf[10]*256+recbuf[11];
					recbuf[32]=0;
					NotifUnameChg(ToID,(char*)(recbuf+12));
				}else if(subtype==5){
					ToID=recbuf[10]*256+recbuf[11];
					NotifUremoved(ToID);
				}else if(subtype==6){
					ToID=recbuf[10]*256+recbuf[11];
					NotifUstatus(ToID, recbuf[12]);
					if(FsendThreadRun&&(ToID==FTOC))
						ImmStopFsend=1;
#ifdef ADMN
				}else if(subtype==9){
					ToID=recbuf[10]*256+recbuf[11];
					AdminPrintf("UNEWID: user added successfully, new ID=%d\n",ToID);
				}else if(subtype==10){
					AdminPrintf("NOTIF: SECGRANT, requested admin level granted\n");
					level=recbuf[10];
				}else if(subtype==11){
					//GETUSTATUID
					recbuf[32]=0;
					AdminPrintf("GETUSTATUID: UID=%d, name=%s, online=%d, state=%d, ThRun=%d, ecnt=%d\n",
							recbuf[10]*256+recbuf[11],
							(char*)(recbuf+12),recbuf[33],recbuf[34],recbuf[35],
							recbuf[36]*256+recbuf[37]
							);
				}else if(subtype==12){
					//GETUSECUSTAT
					CryptMD5(recbuf+20);
					MD5_IntToStr(recbuf+20, (char*)resp);
					AdminPrintf("GETUSECUSTAT: stimer=%d, skey=%d, rkey=%d, PW=%s\n",
							recbuf[10]*256+recbuf[11],
							recbuf[12]*16777216+recbuf[13]*65536+recbuf[14]*256+recbuf[15],
							recbuf[16]*16777216+recbuf[17]*65536+recbuf[18]*256+recbuf[19],
							resp
							);
				}else if(subtype==14){
					//GETUSUID
					AdminPrintf("GETUSUID: UstatConn=%d, UstatSent=%d, SumErrors=%d\n",
							recbuf[10]*16777216+recbuf[11]*65536+recbuf[12]*256+recbuf[13],
							recbuf[14]*16777216+recbuf[15]*65536+recbuf[16]*256+recbuf[17],
							recbuf[18]*16777216+recbuf[19]*65536+recbuf[20]*256+recbuf[21]
							);
				}else if(subtype==15){
					//GETSERST
					AdminPrintf("GETSERST: m_vClientList=%d, zombieClientList=%d, unusedClientList=%d, timesupervisClient=%d, Users=%d\n",
						recbuf[10]*16777216+recbuf[11]*65536+recbuf[12]*256+recbuf[13],
						recbuf[14]*16777216+recbuf[15]*65536+recbuf[16]*256+recbuf[17],
						recbuf[18]*16777216+recbuf[19]*65536+recbuf[20]*256+recbuf[21],
						recbuf[22]*16777216+recbuf[23]*65536+recbuf[24]*256+recbuf[25],
						recbuf[26]*16777216+recbuf[27]*65536+recbuf[28]*256+recbuf[29]
						);
#endif
				}else if(subtype==16){
					OpenStream(FromID,recbuf[40],recbuf[41],recbuf[42],(char*)(recbuf+10));
				}else if(subtype==17){
					RecStream(FromID,recbuf[10],recbuf[11],recbuf[12],recbuf+13,size-13);
				}else if(subtype==18){
					CloseStream(FromID,recbuf[10],recbuf[11],recbuf[12]);
#ifdef ADMN
				}else if(subtype==19){
					recbuf[29]=0;
					AdminPrintf("GETIP: %s\n",(char*)(recbuf+10));
#endif
				}else if(subtype==20){
					GuiUpdateNotif();
				}
				break;
			//resend
			case 9:
				if(ToID!=i->ID)
					break;
				ImmStopFsend=1;
				ClientPrintf("ERROR: resend from server\n");
				break;
			//CAlive
			case 10:
				if(ToID!=i->ID)
					break;
				ServerAlive();
				break;
			default:
				ImmStopFsend=1;
				ClientPrintf("Invalid command from server: %d\n",type);
				break;
		}
	}else{
		//message from client
		switch(type){
			//route
			case 6:
				if(ToID!=i->ID)
					break;
				//Client decrypt
				switch(recbuf[9]){
				case 0:
#ifdef ADMN
					Encrypt(CLIENT_SCRIPT_KEY, CRkey, recbuf+10, size-10);
#else
					Encrypt(CLIENT_SCRIPT_KEY, getCRkey(FromID), recbuf+10, size-10);
#endif
					CRC=recbuf[10]; recbuf[10]=0;
					CRCexp=Calculate_CRC8(recbuf+10, size-10);
					if(CRC!=CRCexp){
						ClientPrintf("CRC error in Client script message\n");
						CRouteError(FromID, 5);
						return 0;
					}else{
						recbuf[size-1]=0;
						RouteNotifMsg(FromID,recbuf[11],(char*)(recbuf+14));
					}
					break;
				case 6:
#ifdef ADMN
					CSkey=CalcClientKey(recbuf+10);
					CRkey=CSkey;
					AdminPrintf("ROUTE key sync command: from=%d, KEY=%d\n",FromID,CSkey);
#else
					setCSRkey(CalcClientKey(recbuf+10),FromID);
#endif
					break;
				case 7:
				case 8:
				case 9:
#ifdef ADMN
					Encrypt(CLIENT_SCRIPT_KEY, CRkey, recbuf+10, size-10);
#else
					Encrypt(CLIENT_SCRIPT_KEY, getCRkey(FromID), recbuf+10, size-10);
#endif
					CRC=recbuf[10]; recbuf[10]=0;
					CRCexp=Calculate_CRC8(recbuf+10, size-10);
					if(CRC!=CRCexp){
						ClientPrintf("CRC error in Client script message\n");
						CRouteError(FromID, 5);
						return 0;
					}else{
						switch(recbuf[9]){
						case 7:
							OpenStream(FromID,recbuf[11],recbuf[12],recbuf[13],(char*)recbuf+14);
							break;
						case 8:
							RecStream(FromID,recbuf[11],recbuf[12],recbuf[13],recbuf+14, size-14);
							break;
						case 9:
							CloseStream(FromID,recbuf[11],recbuf[12],recbuf[13]);
							break;
						}
					}
					break;
				case 10:
					//esetleges file klds azonnali lelltsa
					if(FsendThreadRun&&(FromID==FTOC))
						ImmStopFsend=1;
					//Hibakd rkezett a msik klienstl
					NotitErrorCode(FromID,recbuf[10]*256+recbuf[11]);
					break;
				default:
					ClientPrintf("Invalid routing command from%d, to%d\n",FromID,ToID);
					break;
				}
				break;
			//resend
			case 9:
				//Obsolete. Resend mechanism changed. If error returned from client, the last message immedeately resent
				if(ToID!=i->ID)
					break;
				ClientPrintf("ERROR: resend from client\n");
				break;
			default:
				ClientPrintf("Invalid command from client %d, type=%d\n",FromID, type);
				break;
		}
	}

	
	return 0;
}

void Client::CryptMD5(unsigned char* code){
	unsigned char PASSMD5_scry[16]={0xAB,0xEB,0xC3,0x7A,0x91,0xAF,0x4E,0x12,0x82,0xE4,0x22,0x5A,0xF2,0xC7,0xB9,0xE9};
	int x;
	for(x=0;x<16;x++)
		code[x]=code[x]^PASSMD5_scry[x];
}

Streams* Client::LookUpStream(int F,int T,int S,int O){
	list<Streams*>::iterator itl;
	Streams*str;
	for(itl = StrmIn.begin();itl != StrmIn.end();itl++){
		str=*itl;
		if( (str->FromID==F) && (str->type==T) && (str->subtype==S) && (str->ord==O) ){
			return str;
		}
	}
	return NULL;
}

//From, Type, Subtype, Ordinary
//Type:  stream type, 1=file, 2=audio, 3=video, ...
//Subtype: stream subtype (0:default)
//Ordinary: tbb azonos stream jhet be egyszerre s ezeknek a sorszma
void Client::OpenStream(int F,int T,int S,int O,char* name){
	AdminPrintf("Open Stream: F=%d,T=%d,S=%d,O=%d,name=%s\n",F,T,S,O,name);
	FileReceivingNotif(F,name);
	if(LookUpStream(F,T,S,O)!=NULL){
		ClientPrintf("ERROR: stream wanted to be opened twice\n");
		CRouteError(F, 5);
		return;
	}
	AnsiString fplace;
	AnsiString fdir;
	fdir=GetPath()+"Download";
	fplace=fdir+"/"+name;
	
	//open file (checking if it can be opened and set file size to zero)
	FILE*f;
	f=fopen(fplace.String,"wb");
	if(f==NULL){
		if(mkdir(fdir.String, S_IRWXU | S_IRWXG | S_IROTH | S_IXOTH)==-1){
			ClientPrintf("ERROR: Can not create directory 'Download'\n");
			CRouteError(F, 5);
			return;
		}
		f=fopen(fplace.String,"wb");
		if(f==NULL){
			ClientPrintf("ERROR: Stream file writing failed: %s\n",name);
			CRouteError(F, 5);
			return;
		}
	}
	fclose(f);
	
	Streams*Str=new Streams();
	if(!Str)
		return;
	Str->FromID=F;
	Str->type=T;
	Str->subtype=S;
	Str->ord=O;
	strcpy(Str->Sname,fplace.String);
	StrmIn.push_back(Str);
}

void Client::RecStream(int F,int T,int S,int O,unsigned char* data, int size){
	AdminPrintf("Receive Stream: F=%d,T=%d,S=%d,O=%d\n",F,T,S,O);
	Streams*Str=LookUpStream(F,T,S,O);
	if(Str==NULL){
		ClientPrintf("ERROR: stream is not opened yet\n");
		CRouteError(F, 5);
		return;
	}
	//just save the data to file named same as the stream
	FILE*f;
	f=fopen(Str->Sname,"ab+");
	if(f==NULL){
		ClientPrintf("ERROR: Stream file writing failed: %s\n",Str->Sname);
		CRouteError(F, 5);
		return;
	}
	int x;
	for(x=0;x<size;x++){
		fputc(data[x],f);
	}
	fclose(f);
}

void Client::CloseStream(int F,int T,int S,int O){
	AdminPrintf("Close Stream: F=%d,T=%d,S=%d,O=%d\n",F,T,S,O);
	Streams*Str=LookUpStream(F,T,S,O);
	if(Str==NULL){
		ClientPrintf("ERROR: stream is not opened yet\n");
		CRouteError(F, 5);
		return;
	}
	AdminPrintf("File is received: ");
	AdminPrintf(Str->Sname);
	AdminPrintf("\n");
	FileReceivedNotif(F,Str->Sname);
	StrmIn.remove(Str);
	delete Str;
}

//calculate client script key from Client random seed
int Client::CalcClientKey(unsigned char* seed){
	//Client Crypto-random seed generator
	int x;
	int Cryran=-17;
	Cryran+=djb2hash(seed+23,56);
	Cryran+=checksum(seed+5,88);
	Cryran+=seed[0]+seed[3]+seed[6]+seed[98]+seed[97];
	Cryran+=Calculate_CRC8(seed+41,53);
	for(x=1;x<42;x++)
		Cryran^=seed[x];
	Cryran^=checksum(seed,100);
	if(Cryran<0)Cryran*=-1;
	return Cryran&0x7FF;
}

//print out SOCKET error
void Client::PrintError(){
	ClientPrintf("NETWORK ERROR (code=%d)\n",errno);
}
//print out other error
void Client::PrintError(char*err){
	ClientPrintf("ERROR: %s (code=%d)\n",err,errno);
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//ADMIN commands

void Client::SendPhoto(int A, char* fname, int ID){
	int x, y;
	long s;
	BYTE* a = LoadBMP ( &x, &y, &s, fname );
	if(!a){
		ClientPrintf("ERROR: LoadBMP failed\n");
		return;
	}
	if((x!=50)||(y!=50)){
		ClientPrintf("ERROR: BMP size must be 50X50 !\n");
		delete[] a;
		return;
	}
	BYTE* b = ConvertBMPToRGBBuffer ( a, x, y );
	if(!b){
		ClientPrintf("ERROR: ConvertBMPToRGBBuffer failed\n");
		delete[] a;
		return;
	}
	unsigned char* ph=new unsigned char[3750];
	int p=0;
	
	unsigned int t1,t2;
	
	for(x=0;x<3750;x++){
		t1=b[p++]&0xF0;
		t2=(b[p++]&0xF0)>>4;
		ph[x]=t1|t2;
		//ph[x]=(b[p++]&0xF0)|((b[p++]&0xF0)>>4);
	}
	if(ID==-1)
		SendAdm(A, ph, 3750);
	else
		SendAdm(A, ID, ph, 3750);
	delete[] ph;
	delete[] a;
	delete[] b;
}

void Client::SendAdm(int a){
	EnterSendCritical();
	int len=10;
	sendbuf[0]=0xBA;
	sendbuf[1]=(len&0xFF00)>>8;
	sendbuf[2]=len&0xFF;
	sendbuf[3]=0;//CRC
	sendbuf[4]=(i->ID&0xFF00)>>8;
	sendbuf[5]=i->ID&0xFF;
	sendbuf[6]=0;
	sendbuf[7]=0;
	sendbuf[8]=7;
	sendbuf[9]=a;
	sendbuf[3]=Calculate_CRC8(sendbuf, len);
	SendToServer(sendbuf,len);
	LeaveSendCritical();
}

void Client::SendAdm(int a, unsigned char* buf, int len){
	EnterSendCritical();
	len+=10;
	sendbuf[0]=0xBA;
	sendbuf[1]=(len&0xFF00)>>8;
	sendbuf[2]=len&0xFF;
	sendbuf[3]=0;//CRC
	sendbuf[4]=(i->ID&0xFF00)>>8;
	sendbuf[5]=i->ID&0xFF;
	sendbuf[6]=0;
	sendbuf[7]=0;
	sendbuf[8]=7;
	sendbuf[9]=a;
	memcpy(sendbuf+10,buf,len-10);
	sendbuf[3]=Calculate_CRC8(sendbuf, len);
	SendToServer(sendbuf,len);
	LeaveSendCritical();
}

void Client::SendAdm(int a, int ID){
	EnterSendCritical();
	int len=12;
	sendbuf[0]=0xBA;
	sendbuf[1]=(len&0xFF00)>>8;
	sendbuf[2]=len&0xFF;
	sendbuf[3]=0;//CRC
	sendbuf[4]=(i->ID&0xFF00)>>8;
	sendbuf[5]=i->ID&0xFF;
	sendbuf[6]=0;
	sendbuf[7]=0;
	sendbuf[8]=7;
	sendbuf[9]=a;
	sendbuf[10]=(ID&0xFF00)>>8;
	sendbuf[11]=ID&0xFF;
	sendbuf[3]=Calculate_CRC8(sendbuf, len);
	SendToServer(sendbuf,len);
	LeaveSendCritical();
}

void Client::SendAdm(int a, int ID, unsigned char* buf, int len){
	EnterSendCritical();
	len+=12;
	sendbuf[0]=0xBA;
	sendbuf[1]=(len&0xFF00)>>8;
	sendbuf[2]=len&0xFF;
	sendbuf[3]=0;//CRC
	sendbuf[4]=(i->ID&0xFF00)>>8;
	sendbuf[5]=i->ID&0xFF;
	sendbuf[6]=0;
	sendbuf[7]=0;
	sendbuf[8]=7;
	sendbuf[9]=a;
	sendbuf[10]=(ID&0xFF00)>>8;
	sendbuf[11]=ID&0xFF;
	memcpy(sendbuf+12,buf,len-12);
	sendbuf[3]=Calculate_CRC8(sendbuf, len);
	SendToServer(sendbuf,len);
	LeaveSendCritical();
}

void Client::SendFileToServer(char*fname){
	int len;
	FileSendError=0;
	FILE*f;
	f=fopen(fname,"rb");
	if(f==NULL){
		ClientPrintf("File open error\n");
		return;
	}
	
	//truncate path
	char*C=fname;
	while(*C&&(*C!='/'))C++;
	if(*C!=0){
		C=fname+strlen(fname);
		while(*C!='/')C--;
		C++;
		fname=C;
	}
	
	EnterSendCritical();
	//Send open file
	len=strlen(fname)+12;
	sendbuf[0]=0xBA;
	sendbuf[1]=(len&0xFF00)>>8;
	sendbuf[2]=len&0xFF;
	sendbuf[3]=0;//CRC
	sendbuf[4]=(i->ID&0xFF00)>>8;
	sendbuf[5]=i->ID&0xFF;
	sendbuf[6]=0;
	sendbuf[7]=0;
	sendbuf[8]=7;
	sendbuf[9]=33;
	sendbuf[10]=1;
	strcpy((char*)(sendbuf+11), fname);
	sendbuf[3]=Calculate_CRC8(sendbuf, len);
	SendToServer(sendbuf,len);

	//Send file content
	len=11;
	
	int w;
	while(!FileSendError){
		w=fgetc(f);
		if(feof(f))
			break;
		
		sendbuf[len++]=w;
		if(len>=4096){
			len=4096;
			sendbuf[0]=0xBA;
			sendbuf[1]=(len&0xFF00)>>8;
			sendbuf[2]=len&0xFF;
			sendbuf[3]=0;//CRC
			sendbuf[4]=(i->ID&0xFF00)>>8;
			sendbuf[5]=i->ID&0xFF;
			sendbuf[6]=0;
			sendbuf[7]=0;
			sendbuf[8]=7;
			sendbuf[9]=33;
			sendbuf[10]=2;
			sendbuf[3]=Calculate_CRC8(sendbuf, len);
			SendToServer(sendbuf,len);
			len=11;
		}
	}
	if((len>11)&&!FileSendError){
		sendbuf[0]=0xBA;
		sendbuf[1]=(len&0xFF00)>>8;
		sendbuf[2]=len&0xFF;
		sendbuf[3]=0;//CRC
		sendbuf[4]=(i->ID&0xFF00)>>8;
		sendbuf[5]=i->ID&0xFF;
		sendbuf[6]=0;
		sendbuf[7]=0;
		sendbuf[8]=7;
		sendbuf[9]=33;
		sendbuf[10]=2;
		sendbuf[3]=Calculate_CRC8(sendbuf, len);
		SendToServer(sendbuf,len);
	}

	//Close file
	len=11;
	sendbuf[0]=0xBA;
	sendbuf[1]=(len&0xFF00)>>8;
	sendbuf[2]=len&0xFF;
	sendbuf[3]=0;//CRC
	sendbuf[4]=(i->ID&0xFF00)>>8;
	sendbuf[5]=i->ID&0xFF;
	sendbuf[6]=0;
	sendbuf[7]=0;
	sendbuf[8]=7;
	sendbuf[9]=33;
	sendbuf[10]=3;
	sendbuf[3]=Calculate_CRC8(sendbuf, len);
	SendToServer(sendbuf,len);
	fclose(f);
	LeaveSendCritical();
	
	if(!FileSendError){
		ClientPrintf("File sending finished\n");
	}
}

int Client::GetIntFromAdmCmd(char*& cmd){
	if(*cmd==0){
		ClientPrintf("ERROR: number argument expected\n");
		return -1;
	}
	char* s=cmd;
	while((*cmd!=0)&&(*cmd!=32)){
		cmd++;
	}
	if(*cmd==32){
		*cmd=0;
		cmd++;
	}
	int num;
	if(sscanf(s,"%d",&num)<1){
		ClientPrintf("ERROR: number argument expected\n");
		return -1;
	}
	return num;
}

char* Client::GetStrFromAdmCmd(char*& cmd){
	if(*cmd==0){
		ClientPrintf("ERROR: string argument expected\n");
		return NULL;
	}
	char* s=cmd;
	while((*cmd!=0)&&(*cmd!=32)){
		cmd++;
	}
	if(*cmd==32){
		*cmd=0;
		cmd++;
	}
	return s;
}

void Client::ExecAdmCmd(int x,char* cmd){
	char*s;
	char*z;
	int n,m;
	if( (x>=9)&&(x<=27) )
		if(level<1){
			ClientPrintf("Admin Authentication failure: level=%d\n",level);
			return;
		}
	if(x==28)
		if(level<2){
			ClientPrintf("Admin Authentication failure: level=%d\n",level);
			return;
		}
	if(x==29)
		if(level<3){
			ClientPrintf("Admin Authentication failure: level=%d\n",level);
			return;
		}
	if( (x>=30)&&(x<=35) )
		if(level<4){
			ClientPrintf("Admin Authentication failure: level=%d\n",level);
			return;
		}
	switch(x){
		case 0://modify name
			if(!(s=GetStrFromAdmCmd(cmd))){
				return;
			}
			if(strlen(s)>20){
				ClientPrintf("ERROR: name is too long!\n");
				return;
			}
			strcpy(i->Uname,s);
			if(SaveToIni(INIFILE.String)){
				ClientPrintf("ERROR: ini file saving failed!\n");
				return;
			}
			SendAdm(x, (unsigned char*)s, strlen(s)+1);
			break;
		case 1://modify password
			if(!(s=GetStrFromAdmCmd(cmd))){
				return;
			}
			MD5hash(s, i->Upass);
			if(SaveToIni(INIFILE.String)){
				ClientPrintf("ERROR: ini file saving failed!\n");
				return;
			}
			SendAdm(x, i->Upass, 16);
			break;
		case 2:
			if(!(s=GetStrFromAdmCmd(cmd))){
				return;
			}
			SendPhoto(x,s,-1);
			break;
		case 3:
			if(!(s=GetStrFromAdmCmd(cmd))){
				return;
			}
			if(strlen(s)>20){
				ClientPrintf("ERROR: name is too long!\n");
				return;
			}
			SendAdm(x, (unsigned char*)s, strlen(s)+1);
			break;
		case 4:
			if(!(s=GetStrFromAdmCmd(cmd))){
				return;
			}
			if(strlen(s)>20){
				ClientPrintf("ERROR: name is too long!\n");
				return;
			}
			SendAdm(x, (unsigned char*)s, strlen(s)+1);
			break;
		case 5:
			n=GetIntFromAdmCmd(cmd);if(n==-1)return;
			SendAdm(x, n);
			break;
		case 6:
			SendAdm(x);
			break;
		case 7:
			SendAdm(x);
			break;
#ifdef LEVEL1
		case 8:
			SendAdm(x);
			break;
		case 9:
			n=GetIntFromAdmCmd(cmd);if(n==-1)return;
			SendAdm(x, n);
			break;
		case 10:
			SendAdm(x);
			break;
		case 11:
			SendAdm(x);
			break;
		case 12:
			SendAdm(x);
			break;
		case 13:
			SendAdm(x);
			break;
		case 14:
			n=GetIntFromAdmCmd(cmd);if(n==-1)return;
			SendAdm(x, n);
			break;
		case 15:
			n=GetIntFromAdmCmd(cmd);if(n==-1)return;
			SendAdm(x, n);
			break;
		case 16:
			n=GetIntFromAdmCmd(cmd);if(n==-1)return;
			SendAdm(x, n);
			break;
		case 17:
			n=GetIntFromAdmCmd(cmd);if(n==-1)return;
			SendAdm(x, n);
			break;
		case 18:
			n=GetIntFromAdmCmd(cmd);if(n==-1)return;
			SendAdm(x, n);
			break;
		case 19:
			n=GetIntFromAdmCmd(cmd);if(n==-1)return;
			if(!(s=GetStrFromAdmCmd(cmd))){
				return;
			}
			if(strlen(s)>20){
				ClientPrintf("ERROR: name is too long!\n");
				return;
			}
			SendAdm(x, n,(unsigned char*)s, strlen(s)+1);
			break;
		case 20:
			n=GetIntFromAdmCmd(cmd);if(n==-1)return;
			if(!(s=GetStrFromAdmCmd(cmd))){
				return;
			}
			MD5hash(s, resp);
			SendAdm(x, n, resp, 16);
			break;
		case 21:
			n=GetIntFromAdmCmd(cmd);if(n==-1)return;
			if(!(s=GetStrFromAdmCmd(cmd))){
				return;
			}
			SendPhoto(x,s,n);
			break;
		case 22:
			n=GetIntFromAdmCmd(cmd);if(n==-1)return;
			m=GetIntFromAdmCmd(cmd);if(m==-1)return;
			resp[0]=(m&0xFF00)>>8;
			resp[1]=m&0xFF;
			SendAdm(x, n, resp, 2);
			break;
		case 23:
			n=GetIntFromAdmCmd(cmd);if(n==-1)return;
			SendAdm(x, n);
			break;
		case 24:
			n=GetIntFromAdmCmd(cmd);if(n==-1)return;
			SendAdm(x, n);
			break;
		case 25:
			if(!(s=GetStrFromAdmCmd(cmd))){
				return;
			}
			if(strlen(s)>20){
				ClientPrintf("ERROR: name is too long!\n");
				return;
			}
			if(!(z=GetStrFromAdmCmd(cmd))){
				return;
			}
			strcpy((char*)resp,s);
			MD5hash(z, resp+21);
			SendAdm(x, resp, 37);
			break;
		case 26:
			SendAdm(x);
			break;
#endif
#ifdef LEVEL2
		case 27:
			SendAdm(x);
			break;
#endif
#ifdef LEVEL3
		case 28:
			SendAdm(x);
			break;
		case 29:
			if(!(s=GetStrFromAdmCmd(cmd))){
				return;
			}
			SendAdm(x, (unsigned char*)s, strlen(s)+1);
			break;
		case 30:
			SendAdm(x);
			break;
		case 31:
			if(!(s=GetStrFromAdmCmd(cmd))){
				return;
			}
			SendAdm(x,(unsigned char*)s, strlen(s)+1);
			break;
		case 32:
			if(!(s=GetStrFromAdmCmd(cmd))){
				return;
			}
			SendAdm(x, (unsigned char*)s, strlen(s)+1);
			break;
		case 33:
			if(!(s=GetStrFromAdmCmd(cmd))){
				return;
			}
			SendFileToServer(s);
			break;
		case 34:
			n=GetIntFromAdmCmd(cmd);if(n==-1)return;
			SendAdm(x, n);
			break;
		case 35:
			n=GetIntFromAdmCmd(cmd);if(n==-1)return;
			SendAdm(x, n);
			break;
		case 36:
			n=GetIntFromAdmCmd(cmd);if(n==-1)return;
			SendAdm(x, n);
			break;
#endif
		default:
			break;
	}
}

int Client::AdminCmdInterp(char* cmd){
	int x=0;
	while(strcmp(ADMcmd[x],"END")){
		if(!strncmp(ADMcmd[x],cmd,strlen(ADMcmd[x]))){
			ExecAdmCmd(x,cmd+strlen(ADMcmd[x])+1);
			return 1;
		}
		x++;
	}
	return 0;
}

//INTERFACE
//empty implementation of virtual functions
int ReturnDefault=0;
void Client::GuiPrint(char* message){}
void Client::ProcessPartner(char*uname, int id, int ON_OFF, unsigned char* photo){}
void Client::NotifAuthPassed(){}
void Client::NotifUserOffline(){}
void Client::NotifUnameChg(int ID,char*name){}
void Client::NotifUremoved(int ID){}
void Client::NotifUstatus(int ID,int state){}
void Client::RouteNotifMsg(int From,int Font,char*message){}
void Client::GuiUpdateNotif(){}
void Client::NetDisconnectNotif(){}
//key setter, getter functions for ClientApp
int& Client::getCSkey(int ID){return ReturnDefault;}
int& Client::getCRkey(int ID){return ReturnDefault;}
int  Client::setCSRkey(int KEY,int ID){return ReturnDefault;}
void Client::NotitErrorCode(int From,int ecode){}
void Client::FileHasSentNotif(int ID,int res){}
void Client::FileReceivingNotif(int F,char*name){}
void Client::FileReceivedNotif(int F,char*name){}
void Client::ServerAlive(){}		//Echo for alive command
