//Admin Client
#include "BotChat.h"

AdminClient* ACTHIS=NULL;
THREAD TimerThread;
static volatile int TimerThreadRun=0;
int AuthPassed;
int AliveCnt;
AdminClient* ADM;
void ServiceStop();

void LOG(const char *fmt, ...){
	va_list va;
	va_start(va,fmt);
	char buff[256];
	vsprintf(buff,fmt,va);
#ifdef LOG_CONSOLE
	printf("%s",buff);
#else
	FILE*f;
	f=fopen(LOG_FILE,"a+");
	if(f){
		fprintf(f,"%s",buff);
		fclose(f);
	}
#endif
	va_end(va);
}

//Use only sleep below 1sec!
void Sleep(int msec){
   struct timespec tim;
   tim.tv_sec = 0;
   tim.tv_nsec = 1000000L*msec;
   if(nanosleep(&tim , NULL) < 0 ){
      LOG("Nano sleep system call failed %d\n",errno);
      return;
   }
}

static void* TimerThreadFunc(void* pParam){
	while(TimerThreadRun){
		Sleep(300);
		if(ACTHIS&&AuthPassed){
			if(AliveCnt==0){
				ACTHIS->CAlive();
			}
			AliveCnt++;
			if(AliveCnt>20)
				AliveCnt=0;
		}
	}
	return 0;
}

AdminClient::AdminClient(){
	ACTHIS=this;
	tmp=0;
	AuthPassed=0;
	AliveCnt=-80;
	
	//create timer thread
	TimerThreadRun=1;
	TimerThread.v=0;
	int res=pthread_create( &TimerThread.t, NULL, TimerThreadFunc, NULL );
	if(res){
		LOG("Failed to create timer thread (%d)\n",res);
		exit(0);
	}
	TimerThread.v=1;
}

AdminClient::~AdminClient(){
	list<Questioner*>::iterator itl;
	for(itl = QC.begin();itl != QC.end();itl++){
		delete *itl;
	}
	QC.clear();
	
	//closing timer thread
	TimerThreadRun=0;
	TimerThread.v=0;
	LOG("Closing Timer Thread\n");
	struct timespec tim;
	int ret;
	if (clock_gettime(CLOCK_REALTIME, &tim) == -1) {
		LOG("ERROR getting Epoch time!\n");
		tim.tv_sec =0;
	}else{
		tim.tv_sec += 3;
	}
    ret=pthread_timedjoin_np( TimerThread.t, NULL, &tim);
	if(ret){
		LOG("ERROR closing TimerThread\n");
		ret=pthread_cancel(TimerThread.t);
		if(ret){
			ClientPrintf("ERROR cancelling TimerThread\n");
		}
	}
}

//INTERFACE FUNCTIONS

void AdminClient::GuiPrint(char* message){
	LOG("%s",message);
}

//process partner info loaded from server
void AdminClient::ProcessPartner(char*uname, int id, int ON_OFF, unsigned char* photo){
	LOG("Loaded partner: %s, ID=%d, ",uname,id);
	if(ON_OFF)
		LOG("online\n");
	else
		LOG("offline\n");
}

void AdminClient::NotifAuthPassed(){
	LOG("Authentication passed, route enabled\n");
	AuthPassed=1;
}

void AdminClient::NotifUstatus(int ID,int state){
	LOG("NOTIF: user status ID=%d ",ID);
	if(state){
		LOG("online\n");
	}else{
		LOG("offline\n");
		Questioner*Q = LookUpClient(ID);
		if(Q==NULL)
			return;
		QC.remove(Q);
		delete Q;
	}
}

//Route command from another client
void AdminClient::RouteNotifMsg(int From,int Font,char*message){
	LOG("%d-> %s\n",From,message);
	Questioner*Q = LookUpClient(From);
	if(Q==NULL){
		return;
	}
	message[100]=0;
	AnsiString str=message;
	Bot->sendline(str);
	CRouteScript(Bot->replytext.String, 0, From, Q->skey);
	LOG("Bot-> %s\n",Bot->replytext.String);
	Bot->replytext="";
}

void AdminClient::NetDisconnectNotif(){
	LOG("Network is disconnected\n");
	AuthPassed=0;
	kill(getpid(),SIGTERM);
}

void AdminClient::NotitErrorCode(int ecode){
	char code[100];
	sprintf(code,"Error from client, code=%d\n",ecode);
	LOG("%s",code);
}

int& AdminClient::getCSkey(int ID){
	Questioner*Q = LookUpClient(ID);
	if(Q==NULL){
		return tmp;
	}
	return Q->skey;
}

int& AdminClient::getCRkey(int ID){
	Questioner*Q = LookUpClient(ID);
	if(Q==NULL){
		return tmp;
	}
	return Q->rkey;
}

int  AdminClient::setCSRkey(int KEY,int ID){
	Questioner*Q = LookUpClient(ID);
	if(Q==NULL){
		Q=new Questioner();
		Q->ID=ID;
		Q->skey=KEY;
		Q->rkey=KEY;
		QC.push_back(Q);
	}else{
		Q->skey=KEY;
		Q->rkey=KEY;
	}
	return KEY;
}

void AdminClient::FileReceivingNotif(int F,char*name){
}

void AdminClient::FileReceivedNotif(int F,char*name){
	Questioner*Q = LookUpClient(F);
	if(Q==NULL){
		return;
	}
	CRouteScript("Ne kldj nekem fjlokat!\n", 0, F, Q->skey);
}

Questioner* AdminClient::LookUpClient(int id){
	list<Questioner*>::iterator itl;
	for(itl = QC.begin();itl != QC.end();itl++){
		if( (*itl)->ID == id )
			return *itl;
	}
	return NULL;
}


//*********************************************************************************************************
//ERROR SIGNAL HANDLER
static void ErrorSignal (int sig, siginfo_t *siginfo, void *context){
	AnsiString* e = (AnsiString*)siginfo->si_value.sival_ptr;
	LOG ("%s\n",e->String);
	delete e;
}

//*********************************************************************************************************
//Service/Daemon related
int ServiceStart(){
    InitErrorSignal(&ErrorSignal);
	ADM=new AdminClient();
	ADM->Bot=new ChatBot();
	if(!ADM->CreateClient()){
		return 0;
	}
	return 1;
}
	
void ServiceRun(){
#ifdef SERVICEMODE
    LOG("Service main thread is running (ServiceRun)\n");
	while (1){
		sleep(10);
	}
#else
	//Message loop
	char* buf=new char[100];
	LOG("Message loop\n");
	while(gets(buf)){
		if(strlen(buf) == 0)
			break;
	}
	delete[] buf;
#endif
}

void ServiceStop(){
	delete ADM->Bot;
	ADM->DestroyClient();
	delete ADM;
	exit(0);
}

//*******************************************************************************************************

#ifdef SERVICEMODE
//It is a daemon

void signal_handler(int sig){
	switch(sig) {
	case SIGHUP:
	    LOG("SIGHUP received. Nothing to do\n");
		break;
	case SIGTERM:
	    LOG("SIGTERM received. Stopping server\n");
        ServiceStop();
        exit(0);
		break;
	}
}

int main(){
    int i,lfp;
    char str[10];
	if(getppid()==1)exit(0); /* already a daemon */
	i=fork();
	if (i<0) exit(1); /* fork error */
	if (i>0) exit(0); /* parent exits */
	/* child (daemon) continues */
	if(setsid()<0)exit(1); /* obtain a new process group */
	for (i=getdtablesize();i>=0;--i) close(i); /* close all descriptors */
	i=open("/dev/null",O_RDWR); dup(i); dup(i); /* handle standart I/O */
	umask(033); /* set newly created file permissions */
	if(chdir(RUNNING_DIR)<0)exit(1); /* change running directory */
	lfp=open(LOCK_FILE,O_RDWR|O_CREAT,0640);
	if (lfp<0) exit(1); /* can not open */
	if (lockf(lfp,F_TLOCK,0)<0) exit(0); /* can not lock */
	/* first instance continues */
	sprintf(str,"%d\n",getpid());
	write(lfp,str,strlen(str)); /* record pid to lockfile */
	signal(SIGCHLD,SIG_IGN); /* ignore child */
	signal(SIGTSTP,SIG_IGN); /* ignore tty signals */
	signal(SIGTTOU,SIG_IGN);
	signal(SIGTTIN,SIG_IGN);
	signal(SIGHUP,signal_handler); /* catch hangup signal */
	signal(SIGTERM,signal_handler); /* catch kill signal */

    LOG("--------------------------------------\n");
	if(!ServiceStart()){
		LOG("Start FAILED!\n");
		return 0;
	}
	ServiceRun();
	LOG("Service main thread has exited!\n");
	return 0;
}

#else

//It is an application
int main(int argc, char* argv[]){
    if(!ServiceStart()){
    	LOG("Start FAILED!\n");
		return 0;
    }
	ServiceRun();
	ServiceStop();
    return 0;
}
#endif
