//This Linux daemon helps for PHP engine for implementing "Long Polling" method.
//It is used in shoutforum (/var/www/shoutforum) for real time chat like operation.
//wrutten by Andras Bognar 2014
 
#include <stdio.h>
#include <fcntl.h>
#include <signal.h>
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#include <sys/msg.h>
#include <pthread.h>
#include <list>

using namespace std;

#define RUNNING_DIR	"/bin/bamessenger/longpolld"
#define LOCK_FILE	"lpoll.lock"
#define LOG_FILE	"lpoll.log"
#define INI_FILE	"lpoll.ini"
const unsigned int MAX_LEN = 512;
const unsigned int MSG_TYPE = 1;
const unsigned int LPOLL_CNT_START = 8;
const unsigned int UPRESCNTINIT = 60;
const unsigned int VERSIONLIFETIME = 120;

int mq1 = 0;
int mq2 = 0;
volatile int MQ1ThreadRun = 1;
volatile int MQ2ThreadRun = 1;
pthread_t tm1;
pthread_t tm2;
static pthread_mutex_t serverCriticalSection;

volatile int VER=0;
volatile int GTIME=0;

//Forum modification enty noticement
struct Version{
    unsigned int ver;	//version of the modification in the forum
    unsigned int epoch;	//time when it occured
    unsigned int table;	//forum table that has been modified
    unsigned int who;	//The person's ID
    unsigned int action;//action that was done
    unsigned int msgid;	//Message ID that has been modified (only used for likes)
    char del;           //Delete flag
};
//User present database
struct Upres{
    unsigned int cntbk;	//Back counter.
    unsigned int who;	//user ID
    char del;           //Delete flag
};
//Long poll request database, this is the waiting chain for long poll reply
struct Req{
	unsigned int who;	//user ID
	int tcnt;			//Delayed long poll reply timer count down
	unsigned int table;	//forum table that user actually browses
	unsigned int ver;	//requested version
	char del;           //Delete flag
};

list<Version*> Versions;    //Actions / History list
list<Upres*> Upresence;     //User presence list
list<Req*> Request;         //Pending requests queue

void log_message(char *message){
    FILE *logfile;
	logfile=fopen(LOG_FILE,"a");
	if(!logfile) return;
	fprintf(logfile,"%d: %s\n",GTIME,message);
	fclose(logfile);
}

void EnterCriticalSection(){
	struct timespec tim;
	if (clock_gettime(CLOCK_REALTIME, &tim) == -1) {
		log_message("ERROR getting Epoch time!");
		tim.tv_sec =0;
	}else{
		tim.tv_sec += 6;
	}
	int res=pthread_mutex_timedlock (&serverCriticalSection,&tim);
	if(res){
		if(res==ETIMEDOUT){
			log_message("EnterCriticalSection timeout!");
		}else{
			log_message("pthread_mutex_lock() failed");
		}
	}
}
void LeaveCriticalSection(){
	int res=pthread_mutex_unlock (&serverCriticalSection);
	if(res){
		log_message("pthread_mutex_unlock() failed");
	}
}
void InitializeCriticalSection(){
	int res=pthread_mutex_init(&serverCriticalSection, NULL);
	if(res){
		log_message("pthread_mutex_init() failed");
	}
}
void DeleteCriticalSection(){
	int res=pthread_mutex_destroy(&serverCriticalSection);
	if(res){
		log_message("pthread_mutex_destroy() failed");
	}
}

void Cleanup(){
    //delete Message Queues
	msgctl( mq1, IPC_RMID, 0 );
	msgctl( mq2, IPC_RMID, 0 );
	//closing threads
	MQ1ThreadRun = 0;
    MQ2ThreadRun = 0;
	struct timespec tim;
	int ret;
	if (clock_gettime(CLOCK_REALTIME, &tim) == -1) {
		log_message("ERROR getting Epoch time!");
		tim.tv_sec =0;
	}else{
		tim.tv_sec += 3;
	}
    ret=pthread_timedjoin_np( tm1, NULL, &tim);
	if(ret){
		log_message("ERROR closing TM1");
		ret=pthread_cancel(tm1);
		if(ret){
			log_message("ERROR cancelling TM1");
		}
	}
	if (clock_gettime(CLOCK_REALTIME, &tim) == -1) {
		log_message("ERROR getting Epoch time!");
		tim.tv_sec =0;
	}else{
		tim.tv_sec += 3;
	}
    ret=pthread_timedjoin_np( tm2, NULL, &tim);
	if(ret){
		log_message("ERROR closing TM2");
		ret=pthread_cancel(tm2);
		if(ret){
			log_message("ERROR cancelling TM2");
		}
	}
	//delete lists
	list<Version*>::iterator itl;
	for(itl = Versions.begin();itl != Versions.end();itl++){
		delete *itl;
	}
	Versions.clear();
	list<Upres*>::iterator ite;
	for(ite = Upresence.begin();ite != Upresence.end();ite++){
		delete *ite;
	}
	Upresence.clear();
	list<Req*>::iterator itr;
	for(itr = Request.begin();itr != Request.end();itr++){
		delete *itr;
	}
	Request.clear();
	DeleteCriticalSection();
	log_message("Exited");
	exit(0);
}

//Increment version numnber
void VersionPlus(){
    VER++;
    FILE*f=fopen("version.txt","w");
	if(f){
		fprintf(f,"%d",VER);
		fclose(f);
	}
}

//Debug function
//print out all lists to a file
void DebugList(){
	FILE*f;
	f=fopen("debuglist.txt","w");
	if(!f){
		log_message("ERROR can not open debug log file");
		return;
	}
	fprintf(f,"Versions(%d)\n",Versions.size());
	EnterCriticalSection();
	list<Version*>::iterator itl;
	for(itl = Versions.begin();itl != Versions.end();itl++){
		fprintf(f,"%d,%d,%d,%d,%d,%d,%d\n",(*itl)->ver,(*itl)->epoch,(*itl)->table,(*itl)->who,(*itl)->action,(*itl)->msgid,(*itl)->del);
	}
	fprintf(f,"Upresence(%d)\n",Upresence.size());
	list<Upres*>::iterator ite;
	for(ite = Upresence.begin();ite != Upresence.end();ite++){
		fprintf(f,"%d,%d,%d\n",(*ite)->cntbk,(*ite)->who,(*ite)->del);
	}
	fprintf(f,"Request(%d)\n",Request.size());
	list<Req*>::iterator itr;
	for(itr = Request.begin();itr != Request.end();itr++){
		fprintf(f,"%d,%d,%d,%d,%d\n",(*itr)->who,(*itr)->tcnt,(*itr)->table,(*itr)->ver,(*itr)->del);
	}
	LeaveCriticalSection();
	fclose(f);
}

//Service a request
//sendout==1: send out "0" if no message otherwise no
//return 0 if no new message in list
int ServeRequest(Req*R,int sendout){
    struct message {
    	long type;
    	char resp[1024];	//request service response
	}msg;
	msg.type=R->who;
	int n=0;
	char* c=msg.resp;
	*c=0;
	int ok=1;
	
	if((R->ver==VER)||(Versions.size()==0)){
		n=0;
	}else{
		list<Version*>::iterator itv;
		for(itv = Versions.begin();itv != Versions.end();itv++){
			if((*itv)->who == R->who)continue;
			if((*itv)->ver <= R->ver)continue;
			/*
			if((*itv)->table != R->table){
				sprintf(c,"%d",(*itv)->table);
				while(*c){c++;n++;}
				if(n>1000){ok=0;break;}
				*c=';';c++;
				continue;
			}
			*/
			sprintf(c,"%d",(*itv)->ver);
			while(*c){c++;n++;}
			if(n>1000){ok=0;break;}
			*c=':';c++;
	
			sprintf(c,"%d",(*itv)->epoch);
			while(*c){c++;n++;}
			if(n>1000){ok=0;break;}
			*c=':';c++;
        		
			sprintf(c,"%d",(*itv)->table);
			while(*c){c++;n++;}
			if(n>1000){ok=0;break;}
			*c=':';c++;
		        		
    		sprintf(c,"%d",(*itv)->who);
        	while(*c){c++;n++;}
			if(n>1000){ok=0;break;}
			*c=':';c++;
		        		
			sprintf(c,"%d",(*itv)->action);
			while(*c){c++;n++;}
			if(n>1000){ok=0;break;}
			*c=':';c++;
	        		
			sprintf(c,"%d",(*itv)->msgid);
			while(*c){c++;n++;}
			if(n>1000){ok=0;break;}
			*c=';';c++;
		}
		*c=0;
	}
	if(ok==0){
		strcpy(msg.resp,"1");
	}
	if(n==0){
		strcpy(msg.resp,"0");
	}
	
	if(sendout||n){
        if(msgsnd(mq2,(void *)&msg,strlen(msg.resp)+1,IPC_NOWAIT)==-1){
    	    log_message("ERROR: msgsnd() failed");
        }
	}
	return n;
}

//Get online user list
void SendOnlineUserList(int who){
    struct message {
    	long type;
    	char resp[1024];	//request service response
	}msg;
	msg.type=who;
	char*c= msg.resp;
	*c='L';c++;

	list<Upres*>::iterator ite;
	for(ite = Upresence.begin();ite != Upresence.end();ite++){
		sprintf(c,"%d",(*ite)->who);
		while(*c)c++;
		*c=':';c++;
	}
	*c='U';c++;*c=';';c++;*c=0;
	if(msgsnd(mq2,(void *)&msg,strlen(msg.resp)+1,IPC_NOWAIT)==-1){
		log_message("ERROR: msgsnd() failed");
	}
}

//condition function
bool ver_purge_condition(Version*v){
	if(v->del){
		return true;
	}else{
		return false;
	}
}
bool upres_purge_condition(Upres*u){
	if(u->del){
		return true;
	}else{
		return false;
	}
}
bool req_purge_condition(Req*r){
	if(r->del){
		return true;
	}else{
		return false;
	}
}

//Return back all request immediately
void TriggerRequests(){
	list<Req*>::iterator itr;
	for(itr = Request.begin();itr != Request.end();itr++){
		(*itr)->tcnt=0;
	}
}

//Here arrive all of the action noticements just for register them
//No response sent back
//Data block format:
//epoch:table:who:action:msgid
static void* MQ1ThreadFunc(void* pParam){
	char buffer[ MAX_LEN ];
	char str[500];
	int bytes;
	char* c;
	char* d;
	while(MQ1ThreadRun){
		bytes = 0;
		if ( ( bytes = msgrcv( mq1, buffer, MAX_LEN, MSG_TYPE, 0 ) ) < 0 ) {
			if(errno==EIDRM)return NULL;
			log_message("FATAL: msgrcv failed");
	    	Cleanup();
		}
		sprintf(str,"%s",((msgbuf*)buffer)->mtext);
		c=str;
		while((*c)&&(*c!='"'))c++;
		if(*c==0){
		    c=str;
        }else{
            c++;
        }
        d=c;
        bytes=0;
        while(*d){
        	if(*d==':')bytes++;
        	d++;
        }
        if(bytes!=4){
        	log_message("ERROR: invalid noticement");
        	continue;
        }
		d=c;
		VersionPlus();
        Version*v=new Version();
		v->ver=VER;
		while(*d&&(*d!=':'))d++;*d=0;
		strcpy(str,c);d++;c=d;
		v->epoch=atoi(str);
		while(*d&&(*d!=':'))d++;*d=0;
		strcpy(str,c);d++;c=d;
		v->table=atoi(str);
		while(*d&&(*d!=':'))d++;*d=0;
        strcpy(str,c);d++;c=d;
        v->who=atoi(str);
        while(*d&&(*d!=':'))d++;*d=0;
        strcpy(str,c);d++;c=d;
        v->action=atoi(str);
        while(*d&&(*d!=':'))d++;*d=0;
        strcpy(str,c);d++;c=d;
        v->msgid=atoi(str);
        v->del=0;
        EnterCriticalSection();
        Versions.push_back(v);
		LeaveCriticalSection();
        //Return back all request immediately
        TriggerRequests();
	}
}

//Here the requests arrive for processing and respond to them 
//Data block format:
//who:table:version
static void* MQ2ThreadFunc(void* pParam){
	char buffer[ MAX_LEN ];
	char str[500];
	int bytes;
	char* c;
	char* d;
	struct timespec tim;
	int who=0;
	int ok;
	int table;
	int ver;
	while(MQ2ThreadRun){
    	bytes = 0;
		if ( ( bytes = msgrcv( mq2, buffer, MAX_LEN, MSG_TYPE, 0 ) ) < 0 ) {
			if(errno==EIDRM)return NULL;
			log_message("FATAL: msgrcv failed");
	    	Cleanup();
		}
		sprintf(str,"%s",((msgbuf*)buffer)->mtext);
		if (clock_gettime(CLOCK_REALTIME, &tim) != -1) {
			GTIME=tim.tv_sec;
		}
		c=str;
		while((*c)&&(*c!='"'))c++;
		if(*c==0){
		    c=str;
        }else{
            c++;
        }
        //Debug function
        if(*c=='D'){
        	DebugList();
        	continue;
        }
		//Get online user list 
		if(*c=='L'){
			c++;
			who=atoi(c);
        	SendOnlineUserList(who);
        	continue;
        }
		d=c;
        bytes=0;
        while(*d){
        	if(*d==':')bytes++;
        	d++;
        }
        if(bytes!=2){
        	log_message("ERROR: invalid request");
        	continue;
        }
        d=c;
        while(*d&&(*d!=':'))d++;*d=0;
        who=atoi(c);
        d++;c=d;
        while(*d&&(*d!=':'))d++;*d=0;
        table=atoi(c);
        d++;c=d;
        while(*d&&(*d!=':'))d++;*d=0;
        ver=atoi(c);

		Req*R=new Req();
		R->who=who;
		R->tcnt=LPOLL_CNT_START;
		R->del=0;
		R->table=table;
		R->ver=ver;
		
		//pre-check request
		struct message {
    	    long type;
    	    char resp[4];
	    }msg;
		msg.type=who;
		int send=0;
		if((ver+20)<VER){
        	strcpy(msg.resp,"1");
        	send=1;
        }else if(ver>VER){
        	strcpy(msg.resp,"-1");
        	send=1;
        }else{
            if(!ServeRequest(R,0)){
                //request is pending
                EnterCriticalSection();
		        Request.push_back(R);
		        LeaveCriticalSection();
            }else{
                //request is served
                delete R;
            }
        }
        if(send){
            if(msgsnd(mq2,(void *)&msg,strlen(msg.resp)+1,IPC_NOWAIT)==-1){
                log_message("ERROR: msgsnd() failed");
            }
        	delete R;
        }

		EnterCriticalSection();
		ok=0;
		list<Upres*>::iterator ite;
		for(ite = Upresence.begin();ite != Upresence.end();ite++){
			if((*ite)->who==who){
				ok=1;
				(*ite)->cntbk=UPRESCNTINIT;
			}
		}
		LeaveCriticalSection();
		if(!ok){
			Upres*U=new Upres();
			U->cntbk=UPRESCNTINIT;
			U->who=who;
			U->del=0;
			Upresence.push_back(U);
            VersionPlus();
            Version*v=new Version();
            v->ver=VER;
            v->epoch=GTIME;
            v->table=0;
            v->who=who;
            v->action=5;
            v->msgid=0;
            v->del=0;
            EnterCriticalSection();
            Versions.push_back(v);
			LeaveCriticalSection();
			//Return back all request immediately
			TriggerRequests();
		}
	}
}

void signal_handler(int sig){
	switch(sig) {
	case SIGHUP:
		log_message("SIGNAL: hangup signal catched");
		Cleanup();
		break;
	case SIGTERM:
		log_message("SIGNAL: terminate signal catched");
		Cleanup();
		break;
	}
}

int main(){
    int i,lfp;
    char str[10];
	if(getppid()==1)exit(0); /* already a daemon */
	i=fork();
	if (i<0) exit(1); /* fork error */
	if (i>0) exit(0); /* parent exits */
	/* child (daemon) continues */
	if(setsid()<0)exit(1); /* obtain a new process group */
	for (i=getdtablesize();i>=0;--i) close(i); /* close all descriptors */
	i=open("/dev/null",O_RDWR); dup(i); dup(i); /* handle standart I/O */
	umask(027); /* set newly created file permissions */
	if(chdir(RUNNING_DIR)<0)exit(1); /* change running directory */
	lfp=open(LOCK_FILE,O_RDWR|O_CREAT,0640);
	if (lfp<0) exit(1); /* can not open */
	if (lockf(lfp,F_TLOCK,0)<0) exit(0); /* can not lock */
	/* first instance continues */
	sprintf(str,"%d\n",getpid());
	write(lfp,str,strlen(str)); /* record pid to lockfile */
	signal(SIGCHLD,SIG_IGN); /* ignore child */
	signal(SIGTSTP,SIG_IGN); /* ignore tty signals */
	signal(SIGTTOU,SIG_IGN);
	signal(SIGTTIN,SIG_IGN);
	signal(SIGHUP,signal_handler); /* catch hangup signal */
	signal(SIGTERM,signal_handler); /* catch kill signal */

    //initialize
    log_message("STARTUP: Starting LongPollDaemon service");
    char path[500];
    //Read in forum PHP sources directory from INI file
    FILE *f;
	f=fopen(INI_FILE,"r");
	if(!f){
	    log_message("FATAL: could not open INI file");
	    exit(0);
	}
	if(!fgets(path,200,f)){
	    log_message("FATAL: could not read INI file");
	    exit(0);
	}
	fclose(f);
	//create message queues keys and intercommunicate it with PHP forum code
	umask(022);
	key_t key1 = ftok(LOCK_FILE,0);
	key_t key2 = ftok(LOCK_FILE,1);
	int len=strlen(path);
	strcat(path,"/mq1.txt");
	f=fopen(path,"w");
	if(!f){
	    log_message("FATAL: could not open mq1.txt file for writing");
	    exit(0);
	}
	fprintf(f,"%d",key1);
	fclose(f);
	path[len]=0;
	strcat(path,"/mq2.txt");
	f=fopen(path,"w");
	if(!f){
	    log_message("FATAL: could not open mq1.txt file for writing");
	    exit(0);
	}
	fprintf(f,"%d",key2);
	fclose(f);
	//create message queues
	if ( ( mq1 = msgget( key1, 0666 | IPC_CREAT ) ) < 0 ) {
		log_message("FATAL: could not open message queue 1");
	    exit(0);
	}
	if ( ( mq2 = msgget( key2, 0666 | IPC_CREAT ) ) < 0 ) {
		log_message("FATAL: could not open message queue 2");
	    exit(0);
	}
	//Create message threads
	MQ1ThreadRun = 1;
    MQ2ThreadRun = 1;
	int res=pthread_create( &tm1, NULL, MQ1ThreadFunc, NULL );
	if(res){
		log_message("Failed to create tm1 thread");
		exit(0);
	}
	res=pthread_create( &tm2, NULL, MQ2ThreadFunc, NULL );
	if(res){
		log_message("Failed to create tm2 thread");
		exit(0);
	}
	//Create synchronizing object
	InitializeCriticalSection();
    //Read in actual Version
    f=fopen("version.txt","r");
    if(f){
    	fscanf(f,"%d",&VER);
    	fclose(f);
	}else{
		VER=0;
	}
    //Get epoch
    struct timespec tim;
	if (clock_gettime(CLOCK_REALTIME, &tim) == -1) {
		log_message("ERROR getting Epoch time!");
		Cleanup();
	}
    GTIME=tim.tv_sec;
    
    //***************************************
	//Daemon cycle
	list<Version*>::iterator itv;
	list<Upres*>::iterator ite;
	list<Req*>::iterator itr;
	int gc=0;
	int ok;
	while(1){
	    sleep(1);
	    gc++;
	    if(gc>10){
	    	gc=0;
	    	if (clock_gettime(CLOCK_REALTIME, &tim) == -1) {
				log_message("ERROR getting Epoch time!");
				GTIME++;
			}
	    }else{
	    	GTIME++;
	    }
	    //Presence list counter decrement
	    ok=0;
	    EnterCriticalSection();
		for(ite = Upresence.begin();ite != Upresence.end();ite++){
			(*ite)->cntbk--;
			if((*ite)->cntbk<=0){
                VersionPlus();
                Version*v=new Version();
                v->ver=VER;
                v->epoch=GTIME;
                v->table=0;
                v->who=(*ite)->who;
                v->action=6;
                v->msgid=0;
                v->del=0;
                Versions.push_back(v);
				//Return back all request immediately
				TriggerRequests();
                (*ite)->del=1;
                ok=1;
			}
		}
		if(ok){
		    Upresence.remove_if(upres_purge_condition);
		}
		ok=0;
        for(itv = Versions.begin();itv != Versions.end();itv++){
            if(((*itv)->epoch+VERSIONLIFETIME) < GTIME){
                (*itv)->del=1;
                ok=1;
            }
        }
        if(ok){
            Versions.remove_if(ver_purge_condition);
        }
        LeaveCriticalSection();
        ok=0;
        EnterCriticalSection();
        for(itr = Request.begin();itr != Request.end();itr++){
		    (*itr)->tcnt--;
		    if((*itr)->tcnt<=0){
		        //servering request
		        ServeRequest(*itr,1);
		        (*itr)->del=1;
		        ok=1;
		    }
	    }
	    if(ok){
	        Request.remove_if(req_purge_condition);
	    }
	    LeaveCriticalSection();
	    
	}
	
	Cleanup();
	return 0;
}
