//Memo Class

#include "MemoClass.h"
#include "Windowsx.h"

extern int IDord;


Memo::Memo(HWND hwnd,HINSTANCE hinst,char* Caption,int* Size){
    MemoHwnd =
    CreateWindowEx(/*WS_EX_DLGMODALFRAME | */WS_EX_CLIENTEDGE,
                   "EDIT",
                   Caption,
                   WS_CHILD | WS_VISIBLE | WS_HSCROLL | WS_VSCROLL | ES_LEFT | ES_MULTILINE | ES_AUTOHSCROLL | ES_AUTOVSCROLL | ES_NOHIDESEL,
                   Size[0],Size[1],Size[2],Size[3],
                   hwnd,(HMENU)IDord,hinst,0);
			   
	MemoInit();
}

Memo::Memo(HWND hwnd,HINSTANCE hinst,char* Caption,int* Size,int Mode){
	//Create Memo without scrollbars
	if(Mode==1){
		MemoHwnd =
		CreateWindowEx(/*WS_EX_DLGMODALFRAME | */WS_EX_CLIENTEDGE,
                   "EDIT",
                   Caption,
                   WS_CHILD | WS_VISIBLE | ES_LEFT | ES_MULTILINE | ES_AUTOHSCROLL | ES_AUTOVSCROLL | ES_NOHIDESEL,
                   Size[0],Size[1],Size[2],Size[3],
                   hwnd,(HMENU)IDord,hinst,0);
			   
		MemoInit();	
	}
}

void Memo::MemoInit(){
    ID=IDord++;
    
    Strings.OwnerMemo=this;
    Lines=new MemoLines();
    Lines->OwnerMemo=this;
    Lines->Strings.OwnerMemo=this;
	
	//compute visible line count
	RECT r;
	Edit_GetRect(MemoHwnd,&r);
	VisibleLineCnt=r.bottom/14;
	hFont=NULL;
	
	SendMessage(MemoHwnd,EM_SETLIMITTEXT,-1,0);
	MaxCharCnt=0x7FFFFFFE;
	TextBuf=NULL;
	ListBuf=NULL;
	Update=0;
	LastSelectedLine=-1;
	
	MemoProcAddr = (WNDPROC)GetWindowLong(MemoHwnd, GWL_WNDPROC);
	SetWindowLong(MemoHwnd, GWL_WNDPROC, reinterpret_cast<LONG_PTR>(&Memo::StatWndProc));
}

Memo::~Memo(){
    delete Lines;
	//if(hFont)DeleteObject(hFont);
	if(TextBuf)delete TextBuf;
	if(ListBuf)delete ListBuf;
}

void Memo::Clear(){
     Edit_SetText(MemoHwnd,"");
}

void Memo::LoadFromFile(AnsiString& FileName){
	LoadFromFile(FileName.String);
}

void Memo::LoadFromFile(char* FileName){
	Clear();
	DWORD sHi,sLo;
	
	HANDLE hFile=CreateFile(FileName,GENERIC_READ|GENERIC_WRITE,0,NULL,OPEN_EXISTING,0,NULL);
	sLo=GetFileSize(hFile,&sHi);
	CloseHandle(hFile);
	
	if(sHi!=0 ){
		ShowError("ERROR: Too big file to load !","TStringList::LoadFromFile");
		return;
	}
	if(sLo>=MaxCharCnt){
		ShowError("ERROR: Too big file to load !", "TStringList::LoadFromFile");
		return;
	}

	FILE*f;
	if(!( f=fopen(FileName,"rb"))){
		ShowError("ERROR: File open error!", "TStringList::LoadFromFile");
		return;
	}
	TextBuf= new char[sLo+2];
	if(!TextBuf){
		ShowError("ERROR: Text buffer memory allocation error", "TStringList::LoadFromFile");
		return;
	}
	DWORD x;
	for(x=0;x<sLo;x++)
		TextBuf[x]=fgetc(f);
	TextBuf[sLo]=0;
	
	Edit_SetText(MemoHwnd,TextBuf);

	delete[] TextBuf;
	TextBuf=NULL;
	fclose(f);
}

void Memo::SaveToFile(AnsiString& FileName){
	SaveToFile(FileName.String);
}

void Memo::SaveToFile(char* FileName){
	FILE*f;
	if(!( f=fopen(FileName,"wb"))){
		ShowError("ERROR: File open error!", "TStringList::SaveToFile");
		return;
	}

    DWORD lentxt=Edit_GetTextLength(MemoHwnd);
    TextBuf= new char[lentxt+2];
    Edit_GetText(MemoHwnd,TextBuf,lentxt+2);

	int x;
	for(x=0;x<lentxt;x++)
		fputc(TextBuf[x],f);

	delete[] TextBuf;
	TextBuf=NULL;
	fclose(f);	
}

void Memo::Add(AnsiString str){
	if(Update){
		if(ListBuf)ListBuf->Add(str);
		return;
	}
    DWORD lentxt=Edit_GetTextLength(MemoHwnd);
	if((lentxt+str.Length())>=MaxCharCnt){
		ShowError("ERROR: Too much characters in Edit box", "TStringList::Add");
		return;
	}
    TextBuf= new char[lentxt+(DWORD)(strlen(str.String)+3)];
	if(!TextBuf){
		ShowError("ERROR: Memory allocation error", "TStringList::Add");
		return;
	}
    Edit_GetText(MemoHwnd,TextBuf,lentxt+3);
    strcpy(TextBuf+lentxt,str.String);
    strcpy(TextBuf+lentxt+strlen(str.String),"\r\n");
    Edit_SetText(MemoHwnd,TextBuf);
    delete[] TextBuf;
	TextBuf=NULL;
	
	SendMessage(MemoHwnd, WM_VSCROLL , SB_BOTTOM, 0); 
}

void Memo::Add(char* str){
    AnsiString temp(str);
    Add(temp);
}

void Memo::BeginUpdate(){
	if(!Update){
		Update=1;
		if(!ListBuf){
			ListBuf=new TStringList();
		}
	}
}

void Memo::EndUpdate(){
	if(Update){
		Update=0;
		if(ListBuf){
			LoadText(ListBuf);
			delete ListBuf;
			ListBuf=NULL;
		}
	}
}

void Memo::EnsureVisible(unsigned line){
	if(line>=Count()){
		ShowError("ERROR: index exceeds bounds", "TStringList::EnsureVisible");
		return;
	}
	
	//int max,min,npos;
	//GetScrollRange(MemoHwnd,SB_VERT,&min,&max);
	
	SetScrollPos(MemoHwnd,SB_VERT,line,true);
	Edit_ScrollCaret(MemoHwnd);
	/*
	vis=Edit_GetFirstVisibleLine(MemoHwnd);
	int fast=1;
	while(!((vis<=line)&&((vis+VisibleLineCnt)>line))){
		if(abs( vis+VisibleLineCnt/2-line )<VisibleLineCnt)fast=0;
		if(vis<line){
			if(fast)
				SendMessage(MemoHwnd, EM_SCROLL, SB_PAGEDOWN, 0);
			else
				SendMessage(MemoHwnd, EM_SCROLL, SB_LINEDOWN, 0);
		}else{
			if(fast)
				SendMessage(MemoHwnd, EM_SCROLL, SB_PAGEUP, 0);
			else
				SendMessage(MemoHwnd, EM_SCROLL, SB_LINEUP, 0);
		}
		vis=Edit_GetFirstVisibleLine(MemoHwnd);
	}
	*/
	SetFocus(MemoHwnd);
	
}

void Memo::SelectLine(unsigned line){
	if(line>=Count()){
		ShowError("ERROR: index exceeds bounds", "Memo::SelectLine");
		return;
	}
	LastSelectedLine=line;
	int st=Edit_LineIndex(MemoHwnd,line);
	int len=Edit_LineLength(MemoHwnd,st);
	SendMessage(MemoHwnd, EM_SETSEL, st+len, st); 
	SetFocus(MemoHwnd);
	EnsureVisible(line);
	
	//SetScrollPos(MemoHwnd,SB_HORZ,0,true);
	//UpdateWindow(MemoHwnd);
}

void Memo::SetDefFont(){
	hFont = CreateFont(-11,0,0,0,FW_BOLD,FALSE,FALSE,FALSE,DEFAULT_CHARSET,OUT_OUTLINE_PRECIS,
                CLIP_DEFAULT_PRECIS,DEFAULT_QUALITY, VARIABLE_PITCH,TEXT("Tahoma"));
	SendMessage(MemoHwnd,WM_SETFONT ,(WPARAM)hFont,true);
}

void Memo::ReadOnly(bool RO){
	Edit_SetReadOnly(MemoHwnd,RO);
}

void Memo::LoadText(TStringList* Text){
	unsigned x;
	Clear();
	DWORD chnum=0;
	for(x=0;x<Text->Count;x++){
		chnum+=strlen((Text->Strings[x]).String)+3;
	}
	chnum++;
	TextBuf=new char[chnum];
	chnum=0;
	int n=0;
	for(x=0;x<Text->Count;x++){
		n=0;
		while((Text->Strings[x]).String[n]){
			TextBuf[chnum++]=(Text->Strings[x]).String[n++];
		}
		TextBuf[chnum++]='\r';
		TextBuf[chnum++]='\n';
	}
	TextBuf[chnum]=0;
	Edit_SetText(MemoHwnd,TextBuf);
	delete TextBuf;
	TextBuf=NULL;
}

//******************************************************************************
//      MemoStrings stuktura
MemoStrings::MemoStrings(){
}

MemoStrings::~MemoStrings(){
}

AnsiString& MemoStrings::operator[] (unsigned i){
//!!!!Egyenlore csak sor kiolvasasara jo, sort irni vele nem lehet !!!!
    return GetLine(i);
}

AnsiString& MemoStrings::GetLine(unsigned i){
	if(i>=OwnerMemo->Count()){
		ShowError("ERROR: index exceeds bounds", "MemoStrings::GetLine");
		LineStr="";
		return LineStr;
	}
    int len=Edit_GetLine(OwnerMemo->MemoHwnd,i,LineStr.String,AnsiStringMAXLEN-1);
    LineStr.String[len]=0;
    return LineStr;
}

//******************************************************************************
//      MemoLines stuktura
MemoLines::MemoLines(){
}

MemoLines::~MemoLines(){
}
/*
void MemoLines::Add(AnsiString& str){
    OwnerMemo->Add(str);
}
*/
void MemoLines::Add(AnsiString str){
    OwnerMemo->Add(str);
}

void MemoLines::Add(char* str){
    OwnerMemo->Add(str);
}

void MemoLines::Clear(){
    OwnerMemo->Clear();
}

void MemoLines::BeginUpdate(){
	OwnerMemo->BeginUpdate();
}

void MemoLines::EndUpdate(){
	OwnerMemo->EndUpdate();
}

void MemoLines::LoadFromFile(AnsiString& FileName){
	OwnerMemo->LoadFromFile(FileName);
}

void MemoLines::LoadFromFile(char* FileName){
	OwnerMemo->LoadFromFile(FileName);
}

void MemoLines::SaveToFile(AnsiString& FileName){
	OwnerMemo->SaveToFile(FileName);
}

void MemoLines::SaveToFile(char* FileName){
	OwnerMemo->SaveToFile(FileName);
}

int MemoLines::Count(){
	return OwnerMemo->Count();
}

int Memo::Count(){
	return Edit_GetLineCount(MemoHwnd);
}
