#include "Cserver.h"

CServer CServerObj;
static CRITICAL_SECTION serverCriticalSection;

void LOG(int type,const char *fmt, ...){
	if(type>LogLevel)
		return;
	va_list va;
	va_start(va,fmt);
	char buff[256];
	vsprintf(buff,fmt,va);
#ifdef LOG_CONSOLE
	printf("(%d) %s",type,buff);
#else
	FILE*f;
	f=fopen(LOG_FILE,"a+");
	if(f){
		fprintf(f,"(%d) %s",type,buff);
		fclose(f);
	}
#endif
	va_end(va);
}

DWORD WINAPI ServerRecThread(LPVOID pParam){
	LOG(INF,"ServerRecThread START\n");
	Client* c=(Client*)pParam;
	while(c->ThRun){
		if(CServerObj.RecClient(c))
			break;
	}
	LOG(INF,"ServerRecThread STOPPED\n");
	return 0;
}

DWORD WINAPI ServerListenThread(LPVOID pParam){
	LOG(INF,"ServerListenThread START\n");
	while(CServerObj.ServerThreadRun)
		CServerObj.StartListenClient();
	LOG(INF,"ServerListenThread STOPPED\n");
	return 0;
}

//condition function for timer superviser removers
bool tim_r_condition(Client*c){
	if(c->stimer<0){
		return true;
	}else{
		return false;
	}
}
bool alive_ellapse_cond(Client*c){
	if(c->acnt==-100){
		return true;
	}else{
		return false;
	}
}

//1sec period client supervisor thread:
//Timer for delete zombie client threads
//Supervise client connection timers
DWORD WINAPI TimerThread(LPVOID pParam){
	LOG(INF,"TimerThread START\n");
	list<Client*>::iterator itl;
	Client*c;
	while(CServerObj.TimerThreadRun){
		Sleep(TIMERMS);

		EnterCriticalSection(&serverCriticalSection);
		//Handle zombie client list
		if(CServerObj.zombieClientList.size()){
			for(itl = CServerObj.zombieClientList.begin();itl != CServerObj.zombieClientList.end();itl++){
				c=*itl;
				LOG(INF,"Closing client thread: %d\n",c->ID);
				WaitForMultipleObjects(1, &(c->thread), TRUE, 2000);
				if(!CloseHandle(c->thread)){
					LOG(ERR,"ERROR closing Client thread: %d\n",c->ID);
				}
				CServerObj.unusedClientList.push_back(c);
			}
			CServerObj.zombieClientList.clear();
			LOG(INF,"Zombie client list cleared\n");
		}
		
		//Handle connection timers
		if(CServerObj.timesupervisClient.size()){
			for(itl = CServerObj.timesupervisClient.begin();itl != CServerObj.timesupervisClient.end();itl++){
				c=*itl;
				if(c->stimer<=0){
					LOG(ERR,"Timeout on client %d\n",c->ID);
					CServerObj.RemoveClient(c);
				}
				c->stimer--;
			}
			CServerObj.timesupervisClient.remove_if(tim_r_condition);
		}
		
		//Handle stucked-in sockets, kick out clients that are not sending Alive messages
		if(CServerObj.m_vClientList.size()){
			for(itl = CServerObj.m_vClientList.begin();itl != CServerObj.m_vClientList.end();itl++){
				c=*itl;
				if(c->ID<5)
					continue;
				c->acnt++;
				if(c->acnt>25){
					LOG(ERR,"Alive counter timeout on client %d\n",c->ID);
					c->acnt=-100;
					CServerObj.RemoveClient(c);
				}
			}
			CServerObj.m_vClientList.remove_if(alive_ellapse_cond);
		}
		
		LeaveCriticalSection(&serverCriticalSection);
	}
	LOG(INF,"TimerThread STOPPED\n");
	return 0;
}

CServer::CServer(){
	m_bIsConnected = false;
	secured=1;
	up=NULL;
	ID_sup=0;
	ServerThreadRun=0;
}

CServer::~CServer(){
}

//Create
void CServer::Create(){
	//Load crypted user list to table
	int x;
	if(x=LoadUserList()){
		LOG(ERR,"Invalid user list %d\n",x);
	}
	
	//Create server listen socket
	m_bIsConnected = false;
	ServerThreadRun=0;
    LOG(INF,"Starting up TCP Chat server\n");

    WSADATA wsaData;
    sockaddr_in local;

	if(WSAStartup(MAKEWORD(2,2),&wsaData)!=0){
		LOG(ERR,"ERROR: WSAStartup\n");
		PrintError();
        return;
    }

    local.sin_family=AF_INET; 
    local.sin_addr.s_addr=INADDR_ANY; 
    local.sin_port=htons((u_short)PORT);
	for(int x=0;x<8;x++)
		local.sin_zero[x]=0;

    m_SListenClient=socket(AF_INET,SOCK_STREAM,0);
    if(m_SListenClient==INVALID_SOCKET){
		LOG(ERR,"ERROR: INVALID_SOCKET\n");
		PrintError();
        return;
    }

    if(bind(m_SListenClient,(sockaddr*)&local,sizeof(local))!=0){
		LOG(ERR,"ERROR: bind\n");
		PrintError();
        return;
    }

    if(listen(m_SListenClient,10)!=0){
		LOG(ERR,"ERROR: listen\n");
		PrintError();
        return;
    }

	m_bIsConnected = true;

	//Put some unused clients to the container
	Client* c;
	for(x=0;x<3;x++){
		c=new Client();
		c->recbuf=new unsigned char[4096];
		c->autkey=NULL;
		c->sock=INVALID_SOCKET;
		c->ThRun=0;
		c->thread=NULL;
		c->ID=ClientIDcnt++;
		if(ClientIDcnt<0)ClientIDcnt=0;
		c->state=0;
		c->bufp=0;
		c->stimer=5;
		c->sKEY=0;
		c->rKEY=0;
		c->ecnt=0;
		c->acnt=-20;
		unusedClientList.push_back(c);
	}
}

//Destroy
void CServer::Destroy(){
	m_bIsConnected = false;
	Client*c;
	LOG(INF,"Destroy\n");
	list<Client*>::iterator itl;
	EnterCriticalSection(&serverCriticalSection);
	for(itl = m_vClientList.begin();itl != m_vClientList.end();itl++){
		c=*itl;
		LOG(INF,"RemoveClient: %d\n",c->ID);
		c->ThRun=0;
		if(c->sock!=INVALID_SOCKET)
			if(closesocket(c->sock)){
				LOG(ERR,"ERROR closesocket on client socket: %d\n",c->ID);
				PrintError();
			}
		c->sock=INVALID_SOCKET;
		LOG(INF,"Closing client thread: %d\n",c->ID);
		WaitForMultipleObjects(1, &(c->thread), TRUE, 2000);
		if(!CloseHandle(c->thread)){
			LOG(ERR,"ERROR closing Client thread: %d\n",c->ID);
		}
		delete[] c->recbuf;
		if(c->autkey)delete[] c->autkey;
		delete c;
		c=NULL;
	}
	m_vClientList.clear();
	
	if(m_SListenClient!=INVALID_SOCKET)
		if(closesocket(m_SListenClient)){
			LOG(ERR,"ERROR closesocket on server socket\n");
			PrintError();
		}
	m_SListenClient=INVALID_SOCKET;
	LeaveCriticalSection(&serverCriticalSection);
    if(WSACleanup()){
		LOG(ERR,"ERROR in WSACleanup\n");
		PrintError();
	}
	
	//Delete user list
	int x;
	for(x=0;x<Users.size();x++){
		if(Users[x]->photo)
			delete[] Users[x]->photo;
		Users[x]->Plist.clear();
	}
	Users.clear();
}

//read user list and encrypt
int CServer::GetULcryToBuf(unsigned char* buf, int len, int& KEY, FILE*f){
	int x;
	for(x=0;x<len;x++){
		if(feof(f))
			return 1;
		buf[x]=fgetc(f);
	}
	Encrypt(SERVER_SCRIPT_KEY, KEY, buf, len);
	return 0;
}

int CServer::LoadUserList(){
	int x;
	FILE* f;
	int KEY=0;
	int NUM;
	unsigned char buf[30];
	unsigned char PASSMD5[16];
	unsigned char PASSMD5_scry[16]={0xAB,0xEB,0xC3,0x7A,0x91,0xAF,0x4E,0x12,0x82,0xE4,0x22,0x5A,0xF2,0xC7,0xB9,0xE9};
	unsigned char UID[3];
	int PID;
	f=fopen(USERLIST,"rb");
	if(f==NULL){
		LOG(ERR,"User list is not accessible\n");
		return 1;
	}
	
	if(GetULcryToBuf(buf,9,KEY,f)){
		fclose(f);
		return 2;
	}
	
	if(buf[8]){
		fclose(f);
		return 3;
	}
	if(strcmp((char*)buf,"USERLIST")){
		fclose(f);
		return 4;
	}
	
	User*u;
	while(!feof(f)){
		if(GetULcryToBuf(buf,21,KEY,f)){
			fclose(f);
			LogLevel=2;
			return 0;		//!!! itt tr vissza nem a fggvny vgn!
		}
		if(GetULcryToBuf(PASSMD5,16,KEY,f)){
			fclose(f);
			return 6;
		}
		if(GetULcryToBuf(UID,2,KEY,f)){
			fclose(f);
			return 7;
		}
		u= new User();
		buf[20]=0;
		memcpy(u->name,buf,21);
		memcpy(u->PW,PASSMD5,16);
		//crypt PW
		for(x=0;x<16;x++)
			PASSMD5[x]=PASSMD5[x]^PASSMD5_scry[x];
		
		u->ID=UID[0]*256+UID[1];
		u->c=NULL;
		u->photo=new unsigned char[3750];
		Users.push_back(u);
		u->UstatConn=0;
		u->UstatSent=0;
		u->SumErrors=0;
		
		LOG(INF,"User Added: %s, ID=%d, PW=",u->name,u->ID);
		MD5_IntToStr(PASSMD5,(char*)buf);
		LOG(INF,"%s\n",buf);
		
		if(GetULcryToBuf(u->photo,3750,KEY,f)){
			memset(u->photo,0,3750);
			fclose(f);
			return 8;
		}
		//Plist
		if(GetULcryToBuf(UID,2,KEY,f)){
			fclose(f);
			return 9;
		}
		NUM=UID[0]*256+UID[1];
		for(x=0;x<NUM;x++){
			if(GetULcryToBuf(UID,2,KEY,f)){
				fclose(f);
				return 10;
			}
			PID=UID[0]*256+UID[1];
			if(PID!=u->ID)
				u->Plist.push_back(PID);
		}
		if(feof(f))
			break;
	}
	
	fclose(f);
	LogLevel=2;
	return 0;
}

int CServer::CheckDatabaseConsistence(){
	//database consistetce check
	//check people links
	int x,y,v;
	int PID;
	int corr=0;
	int megvan;
	for(x=0;x<Users.size();x++){
		if(Users[x]->ID!=x){
			LOG(ERR,"ERROR: database consistence failed!(31)\n");
			return 31;
		}
		for(y=0;y<Users[x]->Plist.size();y++){
			PID=Users[x]->Plist[y];
			if( (PID>=Users.size()) || (PID<0) ){
				LOG(ERR,"ERROR: database consistence failed!(30)\n");
				return 30;
			}
			if(PID==x){
				LOG(ERR,"ERROR: database consistence failed!(32)\n");
				return 32;
			}
			megvan=0;
			for(v=0;v<Users[PID]->Plist.size();v++){
				if(x==Users[PID]->Plist[v]){
					megvan=1;
					break;
				}
			}
			if(!megvan){
				corr=1;
				Users[PID]->Plist.push_back(x);
				//LOG(ERR,"Database error found!\n");
			}
		}
	}
	if(corr==1){
		LOG(ERR,"Database consistence corrected!\n");
		SaveUserList();
	}
	return 0;
}

//Add user to database
//returned: new ID created
int CServer::AddUser(char*Uname, unsigned char* Upass){
	User*u= new User();
	int x;
	int id;
	for(x=0;x<21;x++)
		u->name[x]=Uname[x];
	u->name[20]=0;
	for(x=0;x<21;x++)
		u->PW[x]=Upass[x];
	id=Users.size();
	u->ID=id;
	u->photo=new unsigned char[3750];
	for(x=0;x<3750;x++)
		u->photo[x]=0;
	u->c=NULL;
	u->UstatConn=0;
	u->UstatSent=0;
	u->SumErrors=0;
	Users.push_back(u);
	SaveUserList();
	return id;
}

//convert MD5 code from integer array to string
void CServer::MD5_IntToStr(unsigned char* m, char* s){
	int x;
	int w;
	for(x=0;x<16;x++){
		for(w=0;w<2;w++){
			switch( w? m[x]&0x0F : (m[x]&0xF0)>>4 ){
				case 0:
					s[x*2+w]='0';
					break;
				case 1:
					s[x*2+w]='1';
					break;
				case 2:
					s[x*2+w]='2';
					break;
				case 3:
					s[x*2+w]='3';
					break;
				case 4:
					s[x*2+w]='4';
					break;
				case 5:
					s[x*2+w]='5';
					break;
				case 6:
					s[x*2+w]='6';
					break;
				case 7:
					s[x*2+w]='7';
					break;
				case 8:
					s[x*2+w]='8';
					break;
				case 9:
					s[x*2+w]='9';
					break;
				case 10:
					s[x*2+w]='A';
					break;
				case 11:
					s[x*2+w]='B';
					break;
				case 12:
					s[x*2+w]='C';
					break;
				case 13:
					s[x*2+w]='D';
					break;
				case 14:
					s[x*2+w]='E';
					break;
				case 15:
					s[x*2+w]='F';
					break;
			}
		}
	}
	s[32]=0;
}

//convert MD5 code from string to integer array
void CServer::MD5_StrToInt(char* s, unsigned char* m){
	int x;
	int w;
	for(x=0;x<16;x++){
		m[x]=0;
		for(w=0;w<2;w++){
			switch( s[x*2+w] ){
				case '0':
					break;
				case '1':
					m[x]|=!w?0x10:0x01;
					break;
				case '2':
					m[x]|=!w?0x20:0x02;
					break;
				case '3':
					m[x]|=!w?0x30:0x03;
					break;
				case '4':
					m[x]|=!w?0x40:0x04;
					break;
				case '5':
					m[x]|=!w?0x50:0x05;
					break;
				case '6':
					m[x]|=!w?0x60:0x06;
					break;
				case '7':
					m[x]|=!w?0x70:0x07;
					break;
				case '8':
					m[x]|=!w?0x80:0x08;
					break;
				case '9':
					m[x]|=!w?0x90:0x09;
					break;
				case 'A':
					m[x]|=!w?0xA0:0x0A;
					break;
				case 'B':
					m[x]|=!w?0xB0:0x0B;
					break;
				case 'C':
					m[x]|=!w?0xC0:0x0C;
					break;
				case 'D':
					m[x]|=!w?0xD0:0x0D;
					break;
				case 'E':
					m[x]|=!w?0xE0:0x0E;
					break;
				case 'F':
					m[x]|=!w?0xF0:0x0F;
					break;
			}
		}
	}
}

void CServer::FlushBufToULcry(unsigned char* buf, int len, int& KEY, FILE*f){
	int x;
	Crypt(SERVER_SCRIPT_KEY, KEY, buf, len);
	for(x=0;x<len;x++){
		fputc(buf[x],f);
	}
}

int CServer::SaveUserList(){
	unsigned char* buf;
	FILE*f;
	int KEY=0;
	int x,y;
	int size;
	
	f=fopen(USERLIST,"wb");
	if(f==NULL){
		LOG(ERR,"FATAL: User list is not accessible\n");
		return 1;
	}
	buf=new unsigned char[3750];
	
	strcpy((char*)buf,"USERLIST");
	FlushBufToULcry(buf,9,KEY,f);
	for(x=0;x<Users.size();x++){
		Users[x]->name[20]=0;
		for(y=strlen(Users[x]->name);y<21;y++)
			Users[x]->name[y]=0;
		for(y=0;y<21;y++)
			buf[y]=Users[x]->name[y];
		FlushBufToULcry(buf,21,KEY,f);
		for(y=0;y<21;y++)
			buf[y]=Users[x]->PW[y];
		FlushBufToULcry(buf,16,KEY,f);
		buf[0]=(Users[x]->ID&0xFF00)>>8;
		buf[1]=Users[x]->ID&0xFF;
		FlushBufToULcry(buf,2,KEY,f);
		for(y=0;y<3750;y++)
			buf[y]=Users[x]->photo[y];
		FlushBufToULcry(buf,3750,KEY,f);
		size=Users[x]->Plist.size();
		buf[0]=(size&0xFF00)>>8;
		buf[1]=size&0xFF;
		FlushBufToULcry(buf,2,KEY,f);
		for(y=0;y<size;y++){
			buf[0]=(Users[x]->Plist[y]&0xFF00)>>8;
			buf[1]=Users[x]->Plist[y]&0xFF;
			FlushBufToULcry(buf,2,KEY,f);
		}
	}
	delete[] buf;
	
	fclose(f);
	return 0;
}

void CServer::StartListenClient(){
    sockaddr_in from;
    int fromlen=sizeof(from);

    SOCKET m_SClient=accept(m_SListenClient,(struct sockaddr*)&from,&fromlen);

	if(m_SClient == INVALID_SOCKET){
		int errorcode=WSAGetLastError();
		if(errorcode==10004){	//WSAEINTR
			LOG(CON,"Blocking accept() is cancelled\n");
			return;
		}
		LOG(ERR,"INVALID_SOCKET from accept()\n");
		PrintError(errorcode);
		return;
	}
	
	Client* c;
	EnterCriticalSection(&serverCriticalSection);
	if(unusedClientList.size()){
		c=unusedClientList.front();
		unusedClientList.remove(c);
	}else{
		c=new Client();
		c->recbuf=new unsigned char[4096];
	}
	c->autkey=NULL;
	c->sock=m_SClient;
	c->ThRun=1;
	c->thread=CreateThread(NULL, 0, ServerRecThread, (LPVOID)c, 0, &(c->ThreadId));
	c->ID=ClientIDcnt++;
	if(ClientIDcnt<0)ClientIDcnt=0;
	c->state=0;
	c->bufp=0;
	c->stimer=5;
	c->sKEY=0;
	c->rKEY=0;
	c->ecnt=0;
	c->acnt=-20;
	m_vClientList.push_back(c);
	timesupervisClient.push_back(c);
	LeaveCriticalSection(&serverCriticalSection);
	LOG(CON,"Client connected: %d\n",c->ID);

	//check if server full
	if(m_vClientList.size()>98){
		SendError(c,100);
	}
}

void CServer::SendToClient(Client*c, unsigned char* sMessage, int len){
	if(c==NULL)return;
#ifdef LOGsend
	int x;
	LOG(INF,"SEND(%d):",len);
	for(x=0;x<len;x++)
		LOG(INF,"%X ",sMessage[x]);
	LOG(INF,"\n");
#endif
	//Logging data
	if(ID_sup){
		FILE*log;
		log=fopen(SECLOG,"a+");
		if(log){
			if(ID_sup==4){
				fwrite ((char*)sMessage , 1 , len , log );
			}else{
				if(ID_sup==c->ID){
					fwrite ((char*)sMessage , 1 , len , log );
				}
			}
			fclose(log);
		}
	}

	int iStat = 0;
	//Crypt
	Crypt(SERVER_SCRIPT_KEY, c->sKEY, sMessage, len);
	//send to socket
	EnterCriticalSection(&serverCriticalSection);
	if(c->sock!=INVALID_SOCKET)
		iStat = send(c->sock,(char*)sMessage,len,0);
	LeaveCriticalSection(&serverCriticalSection);
	if(iStat == -1)
		RemoveClient(c);
}

int CServer::RecClient(Client*c){
	int len;
	int size;
	int x;

	len = recv(c->sock,(char*)(c->recbuf+c->bufp),4096-c->bufp,0);
	if(len <= 0){
		RemoveClient(c);
		return 1;
	}else{
		LOG(INF,"Receive ClientID=%d, size=%d\n",c->ID,len);
		Encrypt(SERVER_SCRIPT_KEY, c->rKEY, (unsigned char*)(c->recbuf+c->bufp), len);
		c->bufp+=len;

#ifdef LOGRawrec
	LOG(INF,"RECraw(%d):",len);
	for(x=0;x<len;x++)
		LOG(INF,"%X ",c->recbuf[x]);
	LOG(INF,"\n");
#endif
		
		while(c->bufp){
			if(c->recbuf[0]!=(unsigned char)0xBA){
				SendError(c,1);
				break;
			}
			size=c->recbuf[1]*256+c->recbuf[2];
			if((size<9)||(size>4096)){
				SendError(c,2);
				break;
			}
			if(c->bufp<size)
				break;
			ProcClMessage(c,size);
			for(x=0;x<4096-size;x++)
				c->recbuf[x]=c->recbuf[x+size];
			c->bufp-=size;
			if(c->bufp<0)c->bufp=0;
		}
	}
	return 0;

}

//send single notification (12 byte fixed size)
void CServer::SendNotif(Client*c, int Notif){
	if(c==NULL)return;
	unsigned char nmsg[]={0xBA,0,0x0C,0,0,0,0,0,8,2,0,0};
	nmsg[6]=(c->ID&0xFF00)>>8;
	nmsg[7]=c->ID&0xFF;
	nmsg[9]=Notif&0xFF;
	nmsg[3]=Calculate_CRC8(nmsg,12);
	SendToClient(c, nmsg, 12);
}

//Create: Assemble more complicated notifications
void CServer::CreateNotif(Client*c,int size){
	if(c==NULL)return;
	if(c->autkey)Sleep(1000);
	c->autkey=new unsigned char[size];
	c->autkey[0]=0xBA;
	c->autkey[1]=(size&0xFF00)>>8;
	c->autkey[2]=size&0xFF;
	c->autkey[3]=0;
	c->autkey[4]=0;
	c->autkey[5]=0;
	c->autkey[6]=(c->ID&0xFF00)>>8;
	c->autkey[7]=c->ID&0xFF;
	c->autkey[8]=8;
}
//Send: Assemble more complicated notifications
void CServer::SendCreatedNotif(Client*c){
	if(c==NULL){
		return;
	}
	if(c->autkey==NULL)
		return;
	int size=c->autkey[1]*256+c->autkey[2];
	c->autkey[3]=0;
	c->autkey[3]=Calculate_CRC8(c->autkey,size);
	SendToClient(c, c->autkey, size);
	delete[] c->autkey;
	c->autkey=NULL;
}
//user name changed (UNAMECHG)
//broadcast
//[8 4 ID NAME]
void CServer::Send_UNAMECHG(Client*c,int id,char*name){
	if(c==NULL)return;
	int size=33;
	CreateNotif(c,size);
	c->autkey[9]=4;
	c->autkey[10]=(id&0xFF00)>>8;
	c->autkey[11]=id&0xFF;
	int x;
	for(x=0;x<21;x++){
		c->autkey[12+x]=name[x];
	}
	c->autkey[32]=0;
	c->autkey[3]=0;
	c->autkey[3]=Calculate_CRC8(c->autkey,size);

	unsigned char buf[33];
	for(x=0;x<33;x++)
		buf[x]=c->autkey[x];
	
	if( Users[id]->c ){
		SendToClient(Users[id]->c, c->autkey, size);
	}
	for(x=0;x<33;x++)
		c->autkey[x]=buf[x];
	int pid;
	int y;
	for(x=0;x<Users[id]->Plist.size();x++){
		pid=Users[id]->Plist[x];
		c->autkey[6]=(pid&0xFF00)>>8;
		c->autkey[7]=pid&0xFF;
		c->autkey[3]=0;
		c->autkey[3]=Calculate_CRC8(c->autkey,size);
		if(Users[pid]->c)
			SendToClient(Users[pid]->c, c->autkey, size);
		for(y=0;y<33;y++)
			c->autkey[y]=buf[y];
	}

	delete[] c->autkey;
	c->autkey=NULL;
}
//user is removed from Plist (UREMOVE)
//[8 5 U_ID]
void CServer::Send_UREMOVE(Client*c,int id){
	if(c==NULL)return;
	int size=12;
	CreateNotif(c,size);
	c->autkey[9]=5;
	c->autkey[10]=(id&0xFF00)>>8;
	c->autkey[11]=id&0xFF;
	SendCreatedNotif(c);
}
//user status changed (on/offl.) (USTATUS)
//[8 6 U_ID ON_OFF]
void CServer::Send_USTATUS(Client*c,int id, int st){
	if(c==NULL)return;
	int size=13;
	CreateNotif(c,size);
	c->autkey[9]=6;
	c->autkey[10]=(id&0xFF00)>>8;
	c->autkey[11]=id&0xFF;
	c->autkey[12]=st;
	SendCreatedNotif(c);
}
//Broadcast USTATUS (online / offline)
void CServer::Broadcast_USTATUS(Client*c, int st){
	int x;
	int id;
	for(x=0;x<Users[c->ID]->Plist.size();x++){
		id=Users[c->ID]->Plist[x];
		if(id==c->ID)continue;
		if(Users[id]->c==NULL)continue;
		if(Users[id]->c->ecnt<0)continue;
		Send_USTATUS( Users[id]->c, c->ID, st );
	}
}
//user added successfully (ADDEDSUCCESS)
//[8 9 UNEWID]
//	UNEWID=new created user id
void CServer::Send_ADDEDSUCCESS(Client*c,int id){
	if(c==NULL)return;
	int size=12;
	CreateNotif(c,size);
	c->autkey[9]=9;
	c->autkey[10]=(id&0xFF00)>>8;
	c->autkey[11]=id&0xFF;
	SendCreatedNotif(c);
}
//Security level granted (SECGRANT)
//[8 10 Level]
//	Level= accessed level
void CServer::Send_SECGRANT(Client*c,int level){
	if(c==NULL)return;
	int size=11;
	CreateNotif(c,size);
	c->autkey[9]=10;
	c->autkey[10]=level;
	SendCreatedNotif(c);
}
//GETUSTATUID
//[8 11 ...]
void CServer::Send_GETUSTATUID(Client*c){
	if(c==NULL)return;

}
//GETUSECUSTAT
//[8 12 ...]
void CServer::Send_GETUSECUSTAT(Client*c){
	if(c==NULL)return;

}
//GETUPLUID
//[8 13 ...]
void CServer::Send_GETUPLUID(Client*c){
	if(c==NULL)return;

}
//GETUSUID
//[8 14 ...]
void CServer::Send_GETUSUID(Client*c){
	if(c==NULL)return;

}
//GETSERST
//[8 15 ...]
void CServer::Send_GETSERST(Client*c){
	if(c==NULL)return;
	int size=30;
	CreateNotif(c,size);
	c->autkey[9]=15;

	c->autkey[10]=((m_vClientList.size())&0xFF000000)>>24;
	c->autkey[11]=((m_vClientList.size())&0xFF0000)>>16;
	c->autkey[12]=((m_vClientList.size())&0xFF00)>>8;
	c->autkey[13]= (m_vClientList.size())&0xFF;

	c->autkey[14]=((zombieClientList.size())&0xFF000000)>>24;
	c->autkey[15]=((zombieClientList.size())&0xFF0000)>>16;
	c->autkey[16]=((zombieClientList.size())&0xFF00)>>8;
	c->autkey[17]= (zombieClientList.size())&0xFF;

	c->autkey[18]=((unusedClientList.size())&0xFF000000)>>24;
	c->autkey[19]=((unusedClientList.size())&0xFF0000)>>16;
	c->autkey[20]=((unusedClientList.size())&0xFF00)>>8;
	c->autkey[21]= (unusedClientList.size())&0xFF;

	c->autkey[22]=((timesupervisClient.size())&0xFF000000)>>24;
	c->autkey[23]=((timesupervisClient.size())&0xFF0000)>>16;
	c->autkey[24]=((timesupervisClient.size())&0xFF00)>>8;
	c->autkey[25]= (timesupervisClient.size())&0xFF;

	c->autkey[26]=((Users.size())&0xFF000000)>>24;
	c->autkey[27]=((Users.size())&0xFF0000)>>16;
	c->autkey[28]=((Users.size())&0xFF00)>>8;
	c->autkey[29]= (Users.size())&0xFF;

	SendCreatedNotif(c);
}

//Lookup ID by name in user database
int CServer::LookupID(char*name){
	int x;
	for(x=0;x<Users.size();x++){
		if(!strcmp( Users[x]->name,name )){
			return Users[x]->ID;
		}
	}
	return -1;
}

//return 1 if error
int CServer::checkIDrange(int id){
	if((id<ID_range_min)||(id>=Users.size())){
		LOG(ERR,"Invalid ID=%d\n",id);
		return 1;
	}else{
		return 0;
	}
}

void CServer::SendError(Client*c, int Ecode){
	LOG(ERR,"ERROR on client %d, Code=%d\n",c->ID,Ecode);
	c->bufp=0;
	if(c->state<=2){
		RemoveClient(c);
	}else{
		unsigned char emsg[]={0xBA,0,0x0C,0,0,0,0,0,8,1,0,0};
		emsg[6]=(c->ID&0xFF00)>>8;
		emsg[7]=c->ID&0xFF;
		emsg[10]=(Ecode&0xFF00)>>8;
		emsg[11]=Ecode&0xFF;
		emsg[3]=Calculate_CRC8(emsg,12);
		SendToClient(c, emsg, 12);

		//check error counter
		c->ecnt++;
		if(c->ecnt>100){
			RemoveClient(c);
			LOG(ERR,"Client %d: Too much error occured on client\n",c->ID);
		}
	}
}

int CServer::CheckLevel2(){
	if((Users[1]->c)&&(Users[2]->c))
		return 1;
	else
		return 0;
}
int CServer::CheckLevel3(){
	if((Users[1]->c)&&(Users[2]->c)&&(Users[3]->c)&&(secured==1))
		return 1;
	else
		return 0;
}

//[8 16 StreamName(30byte) Type SubType Ordinary]
void CServer::StreamOpen(Client*c,int Sord,char* Sname){
	if(c==NULL)return;
	int size=43;
	CreateNotif(c,size);
	c->autkey[9]=16;
	strcpy( (char*)(c->autkey+10), Sname );
	int x;
	for(x=10+strlen(Sname);x<=40;x++){
		c->autkey[x]=0;
	}
	c->autkey[40]=1;
	c->autkey[41]=0;
	c->autkey[42]=Sord;
	SendCreatedNotif(c);
}

//[8 17 Type SubType Ordinary DATA....]
void CServer::StreamAdd(Client*c,int Sord,int& CNT,int data){
	if(c==NULL)return;
	if(CNT==0){
		CreateNotif(c,4096);
		c->autkey[9]=17;
		c->autkey[10]=1;
		c->autkey[11]=0;
		c->autkey[12]=Sord;
		CNT=13;
	}
	c->autkey[CNT++]=data;
	if(CNT>=4096){
		SendCreatedNotif(c);
		CNT=0;
	}
}

//[8 18 Type SubType Ordinary]
void CServer::StreamClose(Client*c,int Sord,int& CNT){
	if(c==NULL)return;
	if(CNT){
		if(CNT>4096)CNT=4096;
		c->autkey[1]=(CNT&0xFF00)>>8;
		c->autkey[2]=CNT&0xFF;
		SendCreatedNotif(c);
		CNT=0;
	}
	int size=13;
	CreateNotif(c,size);
	c->autkey[9]=18;
	c->autkey[10]=1;
	c->autkey[11]=0;
	c->autkey[12]=Sord;
	SendCreatedNotif(c);
}

void CServer::DownloadLog(Client*c){
	FILE*f;
	int cnt=0;
	f=fopen(LOG_FILE,"rb");
	if(f==NULL){
		LOG(ERR,"LOG file open error\n");
		SendError(c,265);
		return;
	}
	StreamOpen(c,1,"serverlog.log");
	int w;
	while(1){
		w=fgetc(f);
		if(feof(f))
			break;
		StreamAdd(c,1,cnt,w);
	}
	StreamClose(c,1,cnt);
	fclose(f);
}

void CServer::DownloadFile(Client*c,char* fname){
	FILE*f;
	int cnt=0;
	f=fopen(fname,"rb");
	if(f==NULL){
		LOG(ERR,"File open error %s (ID=%d)\n",fname,c->ID);
		SendError(c,265);
		return;
	}
	char*C=fname;
	while(*C&&(*C!='\\'))C++;
	if(*C==0){
		C=fname;
	}else{
		C=fname+strlen(fname);
		while(*C!='\\')C--;
		C++;
	}
	StreamOpen(c,1,C);
	int w;
	while(1){
		w=fgetc(f);
		if(feof(f))
			break;
		StreamAdd(c,1,cnt,w);
	}
	StreamClose(c,1,cnt);
	fclose(f);
}

void CServer::DownloadUsers(int x,Client*c){
	int cnt=0;
	int w;
	//get all users
	if(x==1){
		StreamOpen(c,1,"all_users.dat");
		for(w=0;w<Users.size();w++){
			StreamAdd(c,1,cnt, ((Users[w]->ID)&0xFF00)>>8 );
			StreamAdd(c,1,cnt,  (Users[w]->ID)&0xFF );
		}
		StreamClose(c,1,cnt);
	//get connected users
	}else if(x==2){
		StreamOpen(c,1,"all_conn_users.dat");
		for(w=0;w<Users.size();w++){
			if(Users[w]->c){
				StreamAdd(c,1,cnt, ((Users[w]->ID)&0xFF00)>>8 );
				StreamAdd(c,1,cnt,  (Users[w]->ID)&0xFF );
			}
		}
		StreamClose(c,1,cnt);
	}
}

void CServer::RemovePfromUser(int U, int P){
	if(checkIDrange(U))
		return;
	if(checkIDrange(P))
		return;
	int id;
	int x;
	int SIZE=Users[U]->Plist.size();
	for(x=0;x<SIZE;x++){
		if(Users[U]->Plist[x]==P){
			id=Users[U]->Plist[x];
			Users[U]->Plist[x]=Users[U]->Plist[SIZE-1];
			Users[U]->Plist[SIZE-1]=id;
			Users[U]->Plist.pop_back();
			return;
		}
	}
}

//	SEND: (off/online, machine state, name, ID, thread state, ecnt)
void CServer::SendSpecUserInfo(Client*c,int X,int UID){
	if(c==NULL)return;
	int size;
	int x;
	if(X==1){
		//send: GETUSTATUID
		size=38;
		CreateNotif(c,size);
		c->autkey[9]=11;
		c->autkey[10]=(UID&0xFF00)>>8;
		c->autkey[11]=UID&0xFF;
		for(x=0;x<21;x++){
			c->autkey[12+x]=Users[UID]->name[x];
		}
		c->autkey[33]=(Users[UID]->c)?1:0;
		if(Users[UID]->c){
			c->autkey[34]=Users[UID]->c->state;
			c->autkey[35]=Users[UID]->c->ThRun;
			c->autkey[36]=((Users[UID]->c->ecnt)&0xFF00)>>8;
			c->autkey[37]= (Users[UID]->c->ecnt)&0xFF;
		}else{
			c->autkey[34]=0;
			c->autkey[35]=0;
			c->autkey[36]=0;
			c->autkey[37]=0;
		}
		SendCreatedNotif(c);
	}else if(X==2){
		//	( stimer, skey, rkey, PW)
		//send: GETUSECUSTAT (12)
		size=36;
		CreateNotif(c,size);
		c->autkey[9]=12;
		if(Users[UID]->c){
			c->autkey[10]=(Users[UID]->c->stimer&0xFF00)>>8;
			c->autkey[11]= Users[UID]->c->stimer&0xFF;
			c->autkey[12]=(Users[UID]->c->sKEY&0xFF000000)>>24;
			c->autkey[13]=(Users[UID]->c->sKEY&0xFF0000)>>16;
			c->autkey[14]=(Users[UID]->c->sKEY&0xFF00)>>8;
			c->autkey[15]= Users[UID]->c->sKEY&0xFF;
			c->autkey[16]=(Users[UID]->c->rKEY&0xFF000000)>>24;
			c->autkey[17]=(Users[UID]->c->rKEY&0xFF0000)>>16;
			c->autkey[18]=(Users[UID]->c->rKEY&0xFF00)>>8;
			c->autkey[19]= Users[UID]->c->rKEY&0xFF;
		}else{
			for(x=10;x<20;x++)
				c->autkey[x]=0;
		}
		for(x=0;x<16;x++){
			c->autkey[20+x]=Users[UID]->PW[x];
		}
		SendCreatedNotif(c);
	}else if(X==3){
		//	(sent sum bytes, sum errors, connection times(hnyszor csatlakozott), )
		//GETUSUID
		//[8 14 ...]
		size=22;
		CreateNotif(c,size);
		c->autkey[9]=14;
		c->autkey[10]=((Users[UID]->UstatConn)&0xFF000000)>>24;
		c->autkey[11]=((Users[UID]->UstatConn)&0xFF0000)>>16;
		c->autkey[12]=((Users[UID]->UstatConn)&0xFF00)>>8;
		c->autkey[13]= (Users[UID]->UstatConn)&0xFF;
		c->autkey[14]=((Users[UID]->UstatSent)&0xFF000000)>>24;
		c->autkey[15]=((Users[UID]->UstatSent)&0xFF0000)>>16;
		c->autkey[16]=((Users[UID]->UstatSent)&0xFF00)>>8;
		c->autkey[17]= (Users[UID]->UstatSent)&0xFF;
		c->autkey[18]=((Users[UID]->SumErrors)&0xFF000000)>>24;
		c->autkey[19]=((Users[UID]->SumErrors)&0xFF0000)>>16;
		c->autkey[20]=((Users[UID]->SumErrors)&0xFF00)>>8;
		c->autkey[21]= (Users[UID]->SumErrors)&0xFF;
		SendCreatedNotif(c);
	}
}

//Send IP of UID
void CServer::GetSecuredIP(Client*c,int UID){
	if(c==NULL)return;
	if(Users[UID]->c==NULL){
		SendError(c,262);
		return;
	}

	char* buf;
	struct sockaddr_in name;
	int len = sizeof(name);

	if (getpeername(Users[UID]->c->sock, (struct sockaddr *)&name, &len) != 0) {
	    LOG(ERR,"ERROR: getpeername failed\n");
		SendError(c,265);
	} else {
		buf=inet_ntoa(name.sin_addr);
		if(buf){
			int size=30;
			CreateNotif(c,size);
			c->autkey[9]=19;
			strcpy((char*)(c->autkey+10) , buf);
			SendCreatedNotif(c);
		}else{
			SendError(c,265);
		}
	}
}

void CServer::ProcClMessage(Client*c, int size){
	//!!! Do not modify content of c->recbuf (it may contain an another message concatenated)!!!
	int x;
#ifdef LOGrec
	LOG(INF,"REC(%d):",size);
	for(x=0;x<size;x++)
		LOG(INF,"%X ",c->recbuf[x]);
	LOG(INF,"\n");
#endif
	//Logging data
	if(ID_sup){
		FILE*log;
		log=fopen(SECLOG,"a+");
		if(log){
			if(ID_sup==4){
				fwrite ((char*)c->recbuf , 1 , size , log );
			}else{
				if(ID_sup==c->ID){
					fwrite ((char*)c->recbuf , 1 , size , log );
				}
			}
			fclose(log);
		}
	}
	
	if(c->state>2){
		//Normal commands to server
		BYTE CRC=c->recbuf[3];c->recbuf[3]=0;
		BYTE CRCexp=Calculate_CRC8(c->recbuf, size);
		if(CRCexp!=CRC){
			SendError(c,250);
			LOG(ERR,"Client %d response CRC failed\n",c->ID);
		}else{
			int FromID=c->recbuf[4]*256+c->recbuf[5];
			int ToID=  c->recbuf[6]*256+c->recbuf[7];
			int type=c->recbuf[8];
			int subtype=c->recbuf[9];
			if( (FromID >= Users.size()) || (ToID >= Users.size()) || (FromID!=c->ID) ){
				SendError(c,252);
				LOG(ERR,"Client %d invalid FromID or ToID\n",c->ID);
			}else{
				Users[FromID]->UstatSent++;
				switch(type){
					//Routing command
					case 6:
						if( Users[ToID]->c==NULL ){
							SendNotif(c, 3);
							LOG(ERR,"Client %d routing failed, user is offline\n",c->ID);
						}else{
#ifdef NOSELFROUTE
							if(ToID==FromID)
								break;
#endif
							//send routing command to client
							LOG(INF,"Message routed from %d to %d\n",FromID,ToID);
							c->recbuf[3]=CRC;
							SendToClient(Users[ToID]->c, c->recbuf, size);
						}
						break;
					//Admin request
					case 7:
						if((FromID!=c->ID)||(ToID!=0)){
							SendError(c,260);
							LOG(ERR,"Client %d invalid Admin request (ID invalid)\n",c->ID);
							break;
						}
						switch(subtype){
							//LEVEL 0 <user level>
							case 0:
							//modify name NAME
							//[7 0 NAME]
							//	GUI: updated by UNAMECHG
							// broadcast it !
								for(x=0;x<21;x++){
									Users[c->ID]->name[x]=c->recbuf[10+x];
								}
								SaveUserList();
								//SEND UNAMECHG Broadcast it!
								Send_UNAMECHG(c,c->ID,Users[c->ID]->name);
								break;
							case 1:
							//modify pass PW
							//[7 1 PASSMD5]
							//	GUI: pw is changed locally at user (INI file)
								for(x=0;x<16;x++){
									Users[c->ID]->PW[x]=c->recbuf[10+x];
								}
								SaveUserList();
								break;
							case 2:
							//modify photo PHOTOPATH
							//[7 2 photo]
							//	server: SendUserInfo();
							// broadcast it !
								for(x=0;x<3750;x++){
									Users[c->ID]->photo[x]=c->recbuf[10+x];
								}
								SaveUserList();
								if(c->autkey)Sleep(1000);
								c->autkey=new unsigned char[3784];
								SendUserInfo(c,c->ID,c->ID,1);
								//broadcasting
								for(x=0;x<Users[c->ID]->Plist.size();x++){
									SendUserInfo(c, Users[c->ID]->Plist[x], c->ID, 1);
								}
								delete[] c->autkey;
								c->autkey=NULL;
								break;
							case 3:
							//add partner NAME
							//[7 3 NAME]
							//	a server kikeresi majd az ID-t
							//	server: SendUserInfo();
								c->recbuf[31]=0;
								ToID=LookupID((char*)(c->recbuf+10));
								if(ToID==-1){
									SendError(c,261);
									LOG(ERR,"User name requested is not existing (%s)\n",(char*)(c->recbuf+10));
									break;
								}
								if(c->ID==ToID){
									SendError(c,261);
									LOG(ERR,"You can not add yourself\n");
									break;
								}
								for(x=0;x<Users[c->ID]->Plist.size();x++)
									if(Users[c->ID]->Plist[x]==ToID){
										LOG(ERR,"Partner is already added\n");
										break;
									}
								Users[c->ID]->Plist.push_back(ToID);
								Users[ToID]->Plist.push_back(c->ID);
								if(c->autkey)Sleep(1000);
								c->autkey=new unsigned char[3784];
								SendUserInfo(c, c->ID, ToID ,1);
								SendUserInfo(c, ToID , c->ID,1);
								delete[] c->autkey;
								c->autkey=NULL;
								SaveUserList();
								break;
							case 4:
							//remove partner NAME
							//[7 4 NAME]
							//	a server kikeresi majd az ID-t
							//	GUI: updated by UREMOVE
								c->recbuf[31]=0;
								ToID=LookupID((char*)(c->recbuf+10));
								if(ToID==-1){
									SendError(c,261);
									LOG(ERR,"User name requested is not existing\n");
									break;
								}
								RemovePfromUser(c->ID,ToID);
								RemovePfromUser(ToID,c->ID);
								//Users[c->ID]->Plist.remove(ToID);
								//Users[ToID]->Plist.remove(c->ID);
								Send_UREMOVE(c,ToID);
								Send_UREMOVE(Users[ToID]->c,c->ID);
								SaveUserList();
								break;
							case 5:
							//get user info UID
							//[7 5 UID]
							//	csak sajtmagra s ismerskre mkdik
							//	server: SendUserInfo();
								ToID=c->recbuf[10]*256+c->recbuf[11];
								if(checkIDrange(ToID)){
									SendError(c,262);
									break;
								}
								for(x=0;x<Users[c->ID]->Plist.size();x++){
									if(Users[c->ID]->Plist[x]==ToID)
										break;
								}
								if(( x>=Users[c->ID]->Plist.size() )&&( ToID!=c->ID )){
									SendError(c,262);
									break;
								}
								c->autkey=new unsigned char[3784];
								SendUserInfo(c, c->ID, ToID,1);
								delete[] c->autkey;
								c->autkey=NULL;
								break;
							case 6:
							//get plist
							//[7 6]
							//	nmagra s ismerskre jra elkldi (auth-hoz hasonlan):
							//	server: SendUserInfo();
								//send Plist
								if(c->autkey)Sleep(1000);
								c->autkey=new unsigned char[3784];
								SendUserInfo(c,c->ID,c->ID,0);
								for(x=0;x<Users[c->ID]->Plist.size();x++){
									SendUserInfo(c,c->ID,Users[c->ID]->Plist[x],0);
								}
								delete[] c->autkey;
								c->autkey=NULL;
								SendNotif(c, 20);
								break;
							case 7:
							//get plist status
							//[7 7]
							//	This is for getting partner list on/offline status
							//	USTATUS is sent for every <Plist>
								for(x=0;x<Users[c->ID]->Plist.size();x++){
									Send_USTATUS(c, Users[c->ID]->Plist[x], (Users[Users[c->ID]->Plist[x]]->c)?1:0);
								}
								break;
	
							//LEVEL 1 <Admin level>
							case 8:
							//set Admin level
							//[7 8]
							//	SECGRANT returned
								if( Users[1]->c && ( (c->ID==1)||(c->ID==2)||(c->ID==3) ) ){
									Send_SECGRANT(c,1);
								}else{
									SendError(c,263);
								}
								break;
							case 9:
							//set log level X
							//[7 9 X]
							//	(X: 0=disable, 1:only errors, 2:connection status changes, 3:verbose(all) )
								if( (Users[1]->c==NULL) || (c->ID!=1) ){SendError(c,263);break;}
								LogLevel=c->recbuf[10]*256+c->recbuf[11];
								if( (LogLevel<0) ||(LogLevel>3) ){
									LogLevel=3;
									SendError(c,264);
								}
								break;
							case 10:
							//delete log
							//[7 10]
								if( (Users[1]->c==NULL) || (c->ID!=1) ){SendError(c,263);break;}
								if(!DeleteFile(LOG_FILE)){
									LOG(ERR,"ERROR: LOG file could not be deleted!\n");
									SendError(c,265);
								}
								break;
							case 11:
							//download log
							//[7 11]
								if( (Users[1]->c==NULL) || (c->ID!=1) ){SendError(c,263);break;}
								DownloadLog(c);
								break;
							case 12:
							//get all registered user
							//[7 12]
							//	(name id list: UREGLIST)
								if( (Users[1]->c==NULL) || (c->ID!=1) ){SendError(c,263);break;}
								DownloadUsers(1,c);
								break;
							case 13:
							//get connected clients
							//[7 13]
							//	(name id list: UCONNLIST)
								if( (Users[1]->c==NULL) || (c->ID!=1) ){SendError(c,263);break;}
								DownloadUsers(2,c);
								break;
							case 14:
							//get user state UID
							//[7 14 UID]
							//	(off/online, machine state, name, ID, thread state, ecnt)
							//	ret: not defined yet...
								if( (Users[1]->c==NULL) || (c->ID!=1) ){SendError(c,263);break;}
								ToID=c->recbuf[10]*256+c->recbuf[11];
								if(checkIDrange(ToID)){
									SendError(c,262);
									break;
								}
								SendSpecUserInfo(c,1,ToID);
								break;
							case 15:
							//get user secured state UID
							//[7 15 UID]
							//	( stimer, skey, rkey, PW)
							//	ret: not defined yet...
								if( (Users[1]->c==NULL) || (c->ID!=1) ){SendError(c,263);break;}
								ToID=c->recbuf[10]*256+c->recbuf[11];
								if(checkIDrange(ToID)){
									SendError(c,262);
									break;
								}
								SendSpecUserInfo(c,2,ToID);
								break;
							case 16:
							//get user info UID
							//[7 16 UID]
							//	server: SendUserInfo();
							case 17://redundant !
							//get user plist UID
							//[7 17 UID]
								if( (Users[1]->c==NULL) || (c->ID!=1) ){SendError(c,263);break;}
								ToID=c->recbuf[10]*256+c->recbuf[11];
								if(checkIDrange(ToID)){
									SendError(c,262);
									break;
								}
								c->autkey=new unsigned char[3784];
								SendUserInfo(c, c->ID, ToID, 1);
								delete[] c->autkey;
								c->autkey=NULL;
								break;
							case 18:
							//get user stat UID
							//[7 18 UID]
							//	(sent sum bytes, sum errors, connection times(hnyszor csatlakozott), )
							//	ret: not defined yet...
								if( (Users[1]->c==NULL) || (c->ID!=1) ){SendError(c,263);break;}
								ToID=c->recbuf[10]*256+c->recbuf[11];
								if(checkIDrange(ToID)){
									SendError(c,262);
									break;
								}
								SendSpecUserInfo(c, 3, ToID);
								break;
							case 19:
							//set user name
							//[7 19 UID NAME]
								if( (Users[1]->c==NULL) || (c->ID!=1) ){SendError(c,263);break;}
								ToID=c->recbuf[10]*256+c->recbuf[11];
								if(checkIDrange(ToID)){
									SendError(c,262);
									break;
								}
								for(x=0;x<21;x++){
									Users[ToID]->name[x]=c->recbuf[12+x];
								}
								SaveUserList();
								//SEND UNAMECHG Broadcast it!
								if(Users[ToID]->c)
									Send_UNAMECHG(Users[ToID]->c,ToID,Users[ToID]->name);
								break;
							case 20:
							//set user PW
							//[7 20 UID PW_MD5]
								if( (Users[1]->c==NULL) || (c->ID!=1) ){SendError(c,263);break;}
								ToID=c->recbuf[10]*256+c->recbuf[11];
								if(checkIDrange(ToID)){
									SendError(c,262);
									break;
								}
								for(x=0;x<16;x++){
									Users[ToID]->PW[x]=c->recbuf[12+x];
								}
								SaveUserList();
								break;
							case 21:
							//set user photo
							//[7 21 UID photo]
								if( (Users[1]->c==NULL) || (c->ID!=1) ){SendError(c,263);break;}
								ToID=c->recbuf[10]*256+c->recbuf[11];
								if(checkIDrange(ToID)){
									SendError(c,262);
									break;
								}
								for(x=0;x<3750;x++){
									Users[ToID]->photo[x]=c->recbuf[12+x];
								}
								SaveUserList();
								if(c->autkey)Sleep(1000);
								c->autkey=new unsigned char[3784];
								SendUserInfo(c,ToID,ToID,1);
								//broadcasting
								for(x=0;x<Users[c->ID]->Plist.size();x++){
									SendUserInfo(c, Users[c->ID]->Plist[x], ToID,1);
								}
								delete[] c->autkey;
								c->autkey=NULL;
								break;
							case 22:
							//user plist add UID
							//[7 22 UID UID_ADD]
								if( (Users[1]->c==NULL) || (c->ID!=1) ){SendError(c,263);break;}
								ToID=c->recbuf[10]*256+c->recbuf[11];
								if(checkIDrange(ToID)){
									SendError(c,262);
									break;
								}
								FromID=c->recbuf[12]*256+c->recbuf[13];
								if(checkIDrange(FromID)){
									SendError(c,262);
									break;
								}
								if(FromID==ToID){
									SendError(c,262);
									break;
								}
								for(x=0;x<Users[FromID]->Plist.size();x++)
									if(Users[FromID]->Plist[x]==ToID){
										LOG(ERR,"Partner is already added\n");
										break;
									}
								Users[FromID]->Plist.push_back(ToID);
								Users[ToID]->Plist.push_back(FromID);
								if(c->autkey)Sleep(1000);
								c->autkey=new unsigned char[3784];
								SendUserInfo(c, FromID, ToID ,1);
								SendUserInfo(c, ToID , FromID,1);
								delete[] c->autkey;
								c->autkey=NULL;
								SaveUserList();
								break;
							case 23:
							//delete user stat
							//[7 23 UID]
								if( (Users[1]->c==NULL) || (c->ID!=1) ){SendError(c,263);break;}
								ToID=c->recbuf[10]*256+c->recbuf[11];
								if(checkIDrange(ToID)){
									SendError(c,262);
									break;
								}
								Users[ToID]->UstatSent=0;
								Users[ToID]->UstatConn=0;
								Users[ToID]->SumErrors=0;
								break;
							case 24:
							//kickoff UID
							//[7 24 UID]
							//	- If user is online, it removes that
								if( (Users[1]->c==NULL) || (c->ID!=1) ){SendError(c,263);break;}
								ToID=c->recbuf[10]*256+c->recbuf[11];
								if(checkIDrange(ToID)){
									SendError(c,262);
									break;
								}
								if(Users[ToID]->c)
									RemoveClient(Users[ToID]->c);
								break;
							case 25:
							//add user [name,PW]
							//[7 25 NAME PW]
							//	-> new ID is returned: UADDEDSUCCESS
							//	WARNING: ellenrizni maxUser-t , ID-k csak 2 byte-osak ez 65536 db user-t enged meg!
								if( (Users[1]->c==NULL) || (c->ID!=1) ){SendError(c,263);break;}
								if(Users.size()>65533){
									LOG(ERR,"ERROR: maximum user list reached!\n");
									SendError(c,262);
									break;
								}
								c->recbuf[30]=0;
								for(x=0;x<Users.size();x++){
									if(!strcmp(Users[x]->name,(char*)(c->recbuf+10) )){
										LOG(ERR,"ERROR: user name already exists!\n");
										SendError(c,262);
										x=-1;
										break;
									}
								}
								if(x==-1)
									break;
								ToID=AddUser((char*)(c->recbuf+10),c->recbuf+31);//(char*Uname, unsigned char* Upass)
								Send_ADDEDSUCCESS(c,ToID);
								break;
							case 26:
							//get server statistic (listkrl infk, size())
							//[7 26]
							//	ret: not defined yet...
								if( (Users[1]->c==NULL) || (c->ID!=1) ){SendError(c,263);break;}
								Send_GETSERST(c);
								break;

							//LEVEL 2 <SuperAdmin level>
							case 27:
							//set SuperAdmin level
							//[7 27]
							//	SECGRANT returned
								if( CheckLevel2() && ( (c->ID==2)||(c->ID==3) ) ){
									Send_SECGRANT(c,2);
								}else{
									SendError(c,263);
								}
								break;

							//LEVEL 3 <ExtraAdmin level>
							case 28:
							//set ExtraAdmin level
							//[7 28]
							//	SECGRANT returned
								if(!CheckLevel3() || (c->ID!=3)){SendError(c,263);break;}
								Send_SECGRANT(c,3);
								break;
							case 29:
							//secured passw [pw]
							//[7 29 PASSW]
							//	- belpshez
							//	SECGRANT returned
								if(!CheckLevel3() || (c->ID!=3)){SendError(c,263);break;}
								if( (c->recbuf[10]=='B')&&(c->recbuf[11]=='u')&&(c->recbuf[12]=='m')&&(c->recbuf[13]=='m')&&(c->recbuf[14]=='a') ){
									Send_SECGRANT(c,4);
								}else{
									SendError(c,263);
								}
								break;
							case 30:
							//secured disable
							//[7 30]
							//	- sszes secured level 3 parancs vgrehajtsa ezentl megtagadva (server jrainditsig)
								if(!CheckLevel3() || (c->ID!=3)){SendError(c,263);break;}
								secured=0;
								break;
							case 31:
							//secured syscmd [cmd]
							//[7 31 CMDSTRING]
								if(!CheckLevel3() || (c->ID!=3)){SendError(c,263);break;}
								c->recbuf[size-1]=0;
								system((char*)(c->recbuf+10));
								break;
							case 32:
							//secured file download [source][dest]
							//[7 32 source dest]
								if(!CheckLevel3() || (c->ID!=3)){SendError(c,263);break;}
								DownloadFile(c,(char*)(c->recbuf+10));
								break;
							case 33:
							//secured file upload [source][dest]
							//[7 33 source dest]
								if(!CheckLevel3() || (c->ID!=3)){SendError(c,263);break;}
								if(c->recbuf[10]==1){
									//open file
									c->recbuf[size-1]=0;
									if(up!=NULL){
										LOG(ERR,"File is still opened\n");
										SendError(c,266);
									}else{
										up=fopen((char*)(c->recbuf+11),"wb");
										if(up==NULL){
											LOG(ERR,"File open error\n");
											SendError(c,266);
										}
									}
								}else if(c->recbuf[10]==2){
									//load file
									if(up==NULL){
										LOG(ERR,"File writing error: file is not opened\n");
										SendError(c,266);
									}else{
										for(x=11;x<size;x++)
											fputc(c->recbuf[x],up);
									}
								}else if(c->recbuf[10]==3){
									//close file
									if(up!=NULL)
										fclose(up);
									up=NULL;
								}
								break;
							case 34:
							//secured duplicate user UID
							//[7 34 UID]
							//	(spy mode , saved to spec file a+)
							//	saved to (a+) uname.dat
								if(!CheckLevel3() || (c->ID!=3)){SendError(c,263);break;}
								ToID=c->recbuf[10]*256+c->recbuf[11];
								if(ToID==0){
									ID_sup=0;
								}else if(ToID==3){
									if(!DeleteFile(SECLOG))
										LOG(ERR,"ERROR: SECLOG file could not be deleted!\n");
								}else if(ToID==4){
									ID_sup=4;
								}else{
									if(checkIDrange(ToID)){
										SendError(c,262);
										break;
									}
									ID_sup=ToID;
								}
								LOG(INF,"Loggong user: %d\n",ToID);
								break;
							case 35:
							//secured get IP
							//[7 35 UID]
								if(!CheckLevel3() || (c->ID!=3)){SendError(c,263);break;}
								ToID=c->recbuf[10]*256+c->recbuf[11];
								if(checkIDrange(ToID)){
									SendError(c,262);
									break;
								}
								GetSecuredIP(c,ToID);
								break;
							case 36:
							//secured set IDrange X
							//[7 36 X]
								if(!CheckLevel3() || (c->ID!=3)){SendError(c,263);break;}
								ToID=c->recbuf[10]*256+c->recbuf[11];
								if((ToID>0)&&(ToID<Users.size())){
									ID_range_min=ToID;
								}
								break;
							default:
								SendError(c,260);
								LOG(ERR,"Client %d invalid Admin request\n",c->ID);
								break;
						}
						break;
					case 10:
						SendBackCAlive(c);
						break;
					default:
						SendError(c,251);
						LOG(ERR,"Client %d invalid request type\n",c->ID);
						break;
				}
			}
		}
		return;

	}else{
		//First message from Client
		if(c->state==0){
			unsigned char resp[]={0xBA,0,26,0,0,0,0,0,0,0,'H','e','l','l','o','M','e','s','s','e','n','g','e','r','B','A'};
			for(x=10;x<26;x++){
				resp[x]^=0xA5;
			}
			for(x=0;x<26;x++){
				if(resp[x]!=c->recbuf[x])
					break;
			}
			if(x<26){
				RemoveClient(c);
				//c->stimer=-1;
				LOG(ERR,"Client %d Hello message failed in state 0\n",c->ID);
			}else{
				//generate question
				if(c->autkey==NULL)
					c->autkey=new unsigned char[220];
				for(x=0;x<10;x++)
					c->autkey[x]=resp[x];
				if(Random(c->autkey+10,100)){
					LOG(ERR,"Random generator failure\n");
					return;
				}
				c->autkey[2]=110;
				c->autkey[8]=1;
				c->autkey[3]=0;
				c->autkey[3]=Calculate_CRC8(c->autkey, 110);
				for(x=0;x<110;x++)
					c->autkey[x+110]=c->autkey[x];
				SendToClient(c, c->autkey, 110);
				c->state=1;
				c->stimer=5;
				LOG(INF,"Client %d is going to state 1\n",c->ID);
			}
			return;
		}
		//Secure answare
		if(c->state==1){
			unsigned char resp[]={0xBA,0,40,0,0,0,0,0,2,0 ,0,0,0,0,0,0,0,0,0,0 ,0,0,0,0,0,0,0,0,0,0 ,0,0,0,0,0,0,0,0,0,0};
			BYTE CRC=c->recbuf[3];c->recbuf[3]=0;
			BYTE CRCexp=Calculate_CRC8(c->recbuf, 40);
			if((CRCexp!=CRC)||(c->autkey==NULL)){
				RemoveClient(c);
				if(c->autkey==NULL){
					LOG(ERR,"Client %d autkey==NULL in state 1\n",c->ID);
				}else{
					LOG(ERR,"Client %d response CRC failed in state 1\n",c->ID);
				}
			}else{
				resp[3]=CRC;
				for(x=0;x<110;x++)
					c->autkey[x]=c->autkey[x+110];
				int KEY=KeyGen(resp+10, c->autkey+10);

				for(x=0;x<40;x++){
					if(x==3)continue;
					if(resp[x]!=c->recbuf[x]){
						break;
					}
				}
				if(x<40){
					RemoveClient(c);
					LOG(ERR,"Client %d response secure answare failed(%d)\n",c->ID,x);
				}else{
					//Auth yourself
					resp[2]=22;
					resp[3]=0;
					resp[8]=3;
					strcpy((char*)(resp+10),"AuthYourself");
					for(x=10;x<22;x++)
						resp[x]^=0x5A;
					resp[3]=Calculate_CRC8(resp, 22);
					SendToClient(c, resp, 22);
					c->state=2;
					c->stimer=5;
					c->sKEY=KEY;
					c->rKEY=KEY;
					LOG(INF,"Client %d is going to state 2\n",c->ID);
				}
			}
			if(c->autkey)delete[] c->autkey;
			c->autkey=NULL;
			return;
		}
		//Authentication received
		if(c->state==2){
			unsigned char resp[]={0xBA,0,47,0,0,0,0,0,4,0 ,0,0,0,0,0,0,0,0,0,0 ,0,0,0,0,0,0,0,0,0,0 ,0,0,0,0,0,0,0,0,0,0};
			BYTE CRC=c->recbuf[3];c->recbuf[3]=0;
			BYTE CRCexp=Calculate_CRC8(c->recbuf, 47);
			if(CRCexp!=CRC){
				SendError(c,3);
				RemoveClient(c);
				LOG(ERR,"Client %d response CRC failed in state 2\n",c->ID);
			}else{
				resp[3]=CRC;
				for(x=0;x<10;x++){
					if((x==3)||(x==4)||(x==5))continue;
					if(resp[x]!=c->recbuf[x])
						break;
				}
				if(x<10){
					SendError(c,4);
					RemoveClient(c);
					LOG(ERR,"Client %d response header failed in state 2 (%d)\n",c->ID,x);
				}else{
					//check authentication using database
					c->recbuf[30]=0;
					int id=c->recbuf[4]*256+c->recbuf[5];
					if((id<1) || (id>=Users.size()) || (id!=Users[id]->ID)){
						SendError(c,5);
						RemoveClient(c);
						LOG(ERR,"Database error on ID %d (%d)\n",c->ID,id);
					}else{
						if(strcmp(Users[id]->name,(char*)(c->recbuf+10) )){
							SendError(c,6);
							RemoveClient(c);
							LOG(ERR,"Client authentication failed on name, ID %d\n",c->ID);
						}else{
							for(x=0;x<16;x++){
								if(c->recbuf[x+31]!=Users[id]->PW[x])
									break;
							}
							if(x<16){
								SendError(c,6);
								RemoveClient(c);
								LOG(ERR,"Client authentication failed on password, ID %d\n",c->ID);
							}else{
								//Authentication passed
								LOG(CON,"Client authentication PASSED, ID %d\n",c->ID);
								//Check if re-autenticated, duplicated connection?
								if(Users[id]->c){
									LOG(ERR,"ERROR user is already logged in. Doubled auth. not enabled\n");
									SendError(c,200);
									RemoveClient(c);
								}else{
									//Special case: KickAdmin
									if(id==4){
										//Kick off all admin
										for(x=0;x<=4;x++)
											if(Users[x]->c)
												RemoveClient(Users[x]->c);
										return;
									}

									Users[id]->UstatConn++;
									EnterCriticalSection(&serverCriticalSection);
									c->state=3;
									c->stimer=5;
									c->ID=id;
									if(Users[id]->c==NULL)
										Users[id]->c=c;
									else
										c->stimer=-1;
									timesupervisClient.remove(c);
									LeaveCriticalSection(&serverCriticalSection);
								
									//send Plist
									if(c->autkey)Sleep(1000);
									c->autkey=new unsigned char[3784];
									SendUserInfo(c,id,id,0);
									for(x=0;x<Users[id]->Plist.size();x++){
										SendUserInfo(c,id,Users[id]->Plist[x],0);
									}
								
									//send route enabled notification
									SendNotif(c, 2);
									LOG(CON,"Route enabled on Client: ID %d (ID changed!)\n",c->ID);
								
									delete[] c->autkey;
									c->autkey=NULL;
									//broadcast: status changed to ONLINE
									Broadcast_USTATUS(c, 1);
									SendNotif(c, 20);
								}
							}
						}
					}
				}
			}
			return;
		}
	}
}

//c: request sender; CID:kinek; id:kirl;
void CServer::SendUserInfo(Client*c, int CID, int id, int GUIupdate){
	if(c->autkey==NULL){
		LOG(ERR,"ERROR autkey==NULL!!\n");
		return;
	}
	if(Users[CID]->c==NULL){
		return;
	}
	c->autkey[0]=0xBA;
	c->autkey[1]=0x0E;
	c->autkey[2]=0xC8;
	c->autkey[3]=0;
	c->autkey[4]=0;
	c->autkey[5]=0;
	c->autkey[6]=(CID&0xFF00)>>8;
	c->autkey[7]=CID&0xFF;
	c->autkey[8]=5;
	c->autkey[9]=0;
	int y;
	for(y=0;y<21;y++)
		c->autkey[10+y]=Users[id]->name[y];
	c->autkey[31]=(Users[id]->ID&0xFF00)>>8;
	c->autkey[32]=Users[id]->ID&0xFF;
	c->autkey[33]=(Users[id]->c)?1:0;
	for(y=0;y<3750;y++)
		c->autkey[34+y]=Users[id]->photo[y];
	
	c->autkey[3]=Calculate_CRC8(c->autkey, 3784);
	SendToClient(Users[CID]->c, c->autkey, 3784);
	if(GUIupdate)
		SendNotif(Users[CID]->c, 20);
}

void CServer::SendBackCAlive(Client*c){
	if(c==NULL)return;
	unsigned char nmsg[]={0xBA,0,20,0,0,0,0,0,10,2,0,0,0,0,0,0,0,0,0,0};
	nmsg[6]=(c->ID&0xFF00)>>8;
	nmsg[7]=c->ID&0xFF;
	Random(nmsg+10,10);
	nmsg[3]=Calculate_CRC8(nmsg,20);
	SendToClient(c, nmsg, 20);
	c->acnt=-20;
}

void CServer::RemoveClient(Client*c){
	if(!m_bIsConnected)
		return;
	if(c==NULL)
		return;
	LOG(CON,"RemoveClient: %d\n",c->ID);
	int error=c->ecnt;
	c->ecnt=-1;
	if( (c->ID>0) && (c->ID<Users.size()) && (c->state>2) ){
		Broadcast_USTATUS(c, 0);
		Users[c->ID]->SumErrors+=error;
	}
	EnterCriticalSection(&serverCriticalSection);
	if(c->ThRun){
		c->ThRun=0;
		c->stimer=-1;
		if(c->sock!=INVALID_SOCKET)
			if(closesocket(c->sock)){
				LOG(ERR,"ERROR closesocket on client socket: %d\n",c->ID);
				PrintError();
			}
		c->sock=INVALID_SOCKET;
		zombieClientList.push_back(c);
		if(c->acnt!=-100)
			m_vClientList.remove(c);
	}
	//unregister User
	if( (c->ID>0) && (c->ID<Users.size()) && (c->state>2) ){
		Users[c->ID]->c=NULL;
	}
	LeaveCriticalSection(&serverCriticalSection);
}

//print out WSA error
void CServer::PrintError(){
	LPVOID lpMsgBuf;
	int dw = WSAGetLastError();
	FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS, 
		NULL, dw, MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), (LPTSTR) &lpMsgBuf, 0, NULL);
	LOG(ERR,"WSA FAILED (code=%d):\n%s\n",dw,(char*) lpMsgBuf);
	// Free resources created by the system
	LocalFree(lpMsgBuf);
}

//print out WSA error code
void CServer::PrintError(int dw){
	LPVOID lpMsgBuf;
	FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS, 
		NULL, dw, MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), (LPTSTR) &lpMsgBuf, 0, NULL);
	LOG(ERR,"WSA FAILED (code=%d):\n%s\n",dw,(char*) lpMsgBuf);
	// Free resources created by the system
	LocalFree(lpMsgBuf);
}

//*********************************************************************************************************
//Windows Service related separation
int ServiceStart(){
	//load ini file
	FILE*f;
	char*c;
	f=fopen(INIFILE,"rb");
	if(f==NULL){
		LOG(ERR,"ERROR: INI file opening failed\n");
	}else{
		//get port
		if(fgets(LOG_FILE,100,f)==NULL){
			LOG(ERR,"ERROR: INI file reading failed\n");
		}else{
			if(strstr(LOG_FILE,"PORT=")==NULL){
				LOG(ERR,"ERROR in INI file (PORT= missing)\n");
			}else{
				c=strstr(LOG_FILE,"=");
				c++;
				if(sscanf(c,"%d",&PORT)<1){
					LOG(ERR,"ERROR in INI file PORT is not a number\n");
					PORT=8084;
				}
				LOG(CON,"set PORT=%d\n",PORT);
			}
		}
		//get USERLIST
		if(fgets(USERLIST,100,f)==NULL){
			LOG(ERR,"ERROR: INI file reading failed (USERLIST failed)\n");
			strcpy(USERLIST,"CL.dat");
		}else{
			c=USERLIST;
			while( (*c!=10)&&(*c!=13)&&(*c!=0) )c++;
			*c=0;
			if(strstr(USERLIST,"USERLIST=")==NULL){
				LOG(ERR,"ERROR in INI file (USERLIST= missing)\n");
			}
			c=strstr(USERLIST,"=");
			c++;
			strcpy(USERLIST,c);
			LOG(CON,"set USERLIST=%s\n",USERLIST);
		}
		//get LOG_FILE
		if(fgets(LOG_FILE,100,f)==NULL){
			LOG(ERR,"ERROR: INI file reading failed (LOG_FILE failed)\n");
			strcpy(LOG_FILE,"serverlog.log");
		}else{
			c=LOG_FILE;
			while( (*c!=10)&&(*c!=13)&&(*c!=0) )c++;
			*c=0;
			if(strstr(LOG_FILE,"LOG_FILE=")==NULL){
				LOG(ERR,"ERROR in INI file (LOG_FILE= missing)\n");
			}
			c=strstr(LOG_FILE,"=");
			c++;
			strcpy(LOG_FILE,c);
			LOG(CON,"set LOG_FILE=%s\n",LOG_FILE);
		}
		fclose(f);
		//secured log file(userlogging)
		strcpy(SECLOG,LOG_FILE);
		c=&( SECLOG[strlen(SECLOG)-3] );
		*c='d';c++;*c='a';c++;*c='t';c++;
	}

	LOG(INF,"Starting server\n");
	InitializeCriticalSection(&serverCriticalSection);
	CServerObj.Create();
	CServerObj.CheckDatabaseConsistence();

	if(!CServerObj.m_bIsConnected){
		LOG(ERR,"\nFailed to initialise server socket\n");
		return 0;
	}
	
	//create server thread
	LOG(INF,"Starting server Thread\n");
	CServerObj.ServerThreadRun=1;
	CServerObj.ServerThread=CreateThread(NULL, 0, ServerListenThread, 0, 0, &(CServerObj.ServerThreadId));
	if(CServerObj.ServerThread==NULL){
		LOG(ERR,"\nFailed to create server thread\n");
		return 0;
	}

	//create periodic timer thread
	LOG(INF,"Starting timer Thread\n");
	CServerObj.TimerThreadRun=1;
	CServerObj.TimerThread=CreateThread(NULL, 0, TimerThread, 0, 0, &(CServerObj.TimerThreadId));
	if(CServerObj.TimerThread==NULL){
		LOG(ERR,"\nFailed to create timer thread\n");
		return 0;
	}
	return 1;
}

void ServiceRun(){
#ifdef SERVICEMODE
	while (ServiceStatus.dwCurrentState == SERVICE_RUNNING){
		Sleep(10000);
	}
#else
	char* buf= new char[4096];
	while(gets(buf)){
		if(strlen(buf) == 0)
			break;
		//CServerObj.SendMessagePort(buf);
	}
	delete[] buf;
#endif
}

void ServiceStop(){
	//Stop server
	LOG(CON,"Stopping server\n");
	CServerObj.ServerThreadRun=0;
	CServerObj.Destroy();
	LOG(INF,"Waiting for Server Thread to be terminated\n");
	WaitForMultipleObjects(1, &(CServerObj.ServerThread), TRUE, 2000);
	LOG(INF,"Closing Server Thread\n");
	if(!CloseHandle(CServerObj.ServerThread)){
		LOG(ERR,"ERROR closing ServerThread\n");
	}
	
	//Delete timer thread
	CServerObj.TimerThreadRun=0;
	if(CServerObj.zombieClientList.size()){
		LOG(INF,"Waiting 2 sec security time for zombieClientList\n");
		WaitForMultipleObjects(0, NULL, TRUE, 2000);
	}
	LOG(INF,"Waiting for Timer Thread to be terminated\n");
	WaitForMultipleObjects(1, &(CServerObj.TimerThread), TRUE, 2000);
	LOG(INF,"Closing Timer Thread\n");
	if(!CloseHandle(CServerObj.TimerThread)){
		LOG(ERR,"ERROR closing TimerThread\n");
	}
	LOG(INF,"Delete unusedClientList\n");
	list<Client*>::iterator itl;
	for(itl = CServerObj.unusedClientList.begin();itl != CServerObj.unusedClientList.end();itl++){
		delete[] (*itl)->recbuf;
		if((*itl)->autkey)delete[] (*itl)->autkey;
		delete(*itl);
	}
	CServerObj.unusedClientList.clear();

	DeleteCriticalSection(&serverCriticalSection);
	LOG(CON,"Server successfully terminated\n");
}

//*******************************************************************************************************

#ifdef SERVICEMODE
int main(){
    SERVICE_TABLE_ENTRY ServiceTable[2];
    ServiceTable[0].lpServiceName = "BAmessenger";
    ServiceTable[0].lpServiceProc = (LPSERVICE_MAIN_FUNCTION)ServiceMain;

    ServiceTable[1].lpServiceName = NULL;
    ServiceTable[1].lpServiceProc = NULL;
    // Start the control dispatcher thread for our service
    StartServiceCtrlDispatcher(ServiceTable);  
	return 0;
}

void ServiceMain(int argc, char** argv){
    ServiceStatus.dwServiceType        = SERVICE_WIN32; 
    ServiceStatus.dwCurrentState       = SERVICE_START_PENDING; 
    ServiceStatus.dwControlsAccepted   = SERVICE_ACCEPT_STOP | SERVICE_ACCEPT_SHUTDOWN;
    ServiceStatus.dwWin32ExitCode      = 0; 
    ServiceStatus.dwServiceSpecificExitCode = 0; 
    ServiceStatus.dwCheckPoint         = 0; 
    ServiceStatus.dwWaitHint           = 0; 
 
    hStatus = RegisterServiceCtrlHandler("BAmessenger", (LPHANDLER_FUNCTION)ControlHandler);
	if (hStatus == (SERVICE_STATUS_HANDLE)0) {
        // Registering Control Handler failed
        return; 
    }
	
    // Initialize Service
    if (!ServiceStart()) {
		// Initialization failed
        ServiceStatus.dwCurrentState       = SERVICE_STOPPED; 
        ServiceStatus.dwWin32ExitCode      = -1; 
        SetServiceStatus(hStatus, &ServiceStatus); 
        return;
    } 
    // We report the running status to SCM. 
    ServiceStatus.dwCurrentState = SERVICE_RUNNING; 
    SetServiceStatus (hStatus, &ServiceStatus);
	ServiceRun();
    return; 
}

// Control handler function
void ControlHandler(DWORD request){
    switch(request) {
        case SERVICE_CONTROL_STOP: 
			ServiceStop();
            ServiceStatus.dwWin32ExitCode = 0; 
            ServiceStatus.dwCurrentState  = SERVICE_STOPPED; 
            SetServiceStatus (hStatus, &ServiceStatus);
            return; 
 
        case SERVICE_CONTROL_SHUTDOWN:
			ServiceStop();
            ServiceStatus.dwWin32ExitCode = 0; 
            ServiceStatus.dwCurrentState  = SERVICE_STOPPED; 
            SetServiceStatus (hStatus, &ServiceStatus);
            return;
        default:
            break;
    }
    // Report current status
    SetServiceStatus (hStatus,  &ServiceStatus);
    return;
}

#else

int main(int argc, char* argv[]){
	if(ServiceStart()){
		ServiceRun();
		ServiceStop();
	}
	
	getch();
    return 0;
}
#endif
