//TStringList Class
#include "TStringListClass.h"

TStringList::TStringList(){
	Root=NULL;
	Tail=NULL;
	Act=NULL;
	ActIndex=0;
	Strings.Own=this;
	Count=0;
}

TStringList::~TStringList(){
	Clear();
}

bool TStringList::LoadFromFile(char* fname){
	FILE*f;
	Clear();
	if(!( f=fopen(fname,"rb"))){
		ErrorAlert("ERROR: File open error!", "TStringList::LoadFromFile");
		return false;
	}
	char c;
	char str[AnsiStringMAXLEN];
	int strp=0;
	while(!feof(f)){
		c=fgetc(f);
		if((c==10)||(c==13)){
			c=fgetc(f);
			str[strp]=0;
			strp=0;
			Add(str);
			continue;
		}
		if(feof(f)){
			str[strp]=0;
			Add(str);
		}
		str[strp++]=c;
		if(strp>=AnsiStringMAXLEN-2){
			str[strp]=0;
			strp=0;
			Add(str);
		}
	}
	fclose(f);
	
	return true;
}

bool TStringList::LoadFromFile(AnsiString& fname){
	return LoadFromFile(fname.String);
}

bool TStringList::SaveToFile(char* fname){
	FILE*f;
	if(!( f=fopen(fname,"wb"))){
		ErrorAlert("ERROR: File open error!", "TStringList::SaveToFile");
		return false;
	}
	TextElem* p;
	int strp;
	for(p=Root;p!=NULL;p=p->FW){
		strp=0;
		while((strp<AnsiStringMAXLEN)&&(p->Element.String[strp])){
			fputc(p->Element.String[strp],f);
			strp++;
		}
		fputc(13,f);
		fputc(10,f);
	}
	fclose(f);
	
	Act=Root;
	ActIndex=0;
	return true;	
}

bool TStringList::SaveToFile(AnsiString& fname){
	return SaveToFile(fname.String);
}

void TStringList::Insert(long i, char* str){
	if((Count==0)||(i>=Count)){
		ErrorAlert("ERROR: index exceeds bound", "TStringList::Insert");
		return;	
	}
	if(Count>COUNTMAX-1){
		ErrorAlert("ERROR: list too long", "TStringList::Add");
		return;		
	}
	TextElem* p=GetElemByIndex(i);
	TextElem* N=new TextElem();
	N->Element=str;
	
	if(!p->BK){
		N->BK=NULL;
		N->FW=Root;
		Root->BK=N;
		Root=N;
		Act=N;
		ActIndex=0;
		Count++;
	}else{
		N->FW=p;
		N->BK=p->BK;
		p->BK->FW=N;
		p->BK=N;
		Act=N;
		ActIndex=i;
		Count++;
	}
}

void TStringList::Insert(long index, AnsiString str){
	Insert(index,str.String);
}

void TStringList::Insert(unsigned index, AnsiString str){
	Insert((long)index,str.String);
}

void TStringList::Insert(unsigned i, char* str){
	Insert((long)i,str);
}

void TStringList::Clear(){
	TextElem* p;
	TextElem* q;
	for(p=Root;p!=NULL;p=q){
		q=p->FW;
		delete p;
	}
	Root=NULL;
	Tail=NULL;
	Act=NULL;
	ActIndex=0;
	Count=0;
}

void TStringList::Delete(unsigned i){
	Delete((long)i);
}

void TStringList::Delete(int i){
	Delete((long)i);
}

void TStringList::Delete(long i){
	if((Count==0)||(i>=Count)){
		ErrorAlert("ERROR: index exceeds bound", "TStringList::Delete");
		return;	
	}
	TextElem* p=GetElemByIndex(i);
	if((!p->FW)&&(!p->BK)){
		Root=NULL;
		Tail=NULL;
		Act=NULL;
		ActIndex=0;
		Count=0;
	}else if(!p->FW){
		Tail=p->BK;
		Tail->FW=NULL;
		Act=Tail;
		Count--;
		ActIndex=Count-1;
	}else if(!p->BK){
		Root=p->FW;
		Root->BK=NULL;
		Act=Root;
		ActIndex=0;
		Count--;
	}else{
		p->BK->FW=p->FW;
		p->FW->BK=p->BK;
		Act=p->FW;
		ActIndex=i;
		Count--;
	}

	delete p;
}

void TStringList::Add(AnsiString str){
	Add(str.String);
}

void TStringList::Add(char* str){
	if(Count>COUNTMAX){
		ErrorAlert("ERROR: list too long", "TStringList::Add");
		return;		
	}
	if(!Root){
		Root=new TextElem();
		if(!Root){
			ErrorAlert("ERROR: new()", "TStringList::Add");
			return;
		}
		Tail=Root;
		Act=Root;
		ActIndex=0;
		Count=1;
		Root->Element=str;
		Root->FW=NULL;
		Root->BK=NULL;
	}else{
		Tail->FW=new TextElem();
		if(!Tail->FW){
			ErrorAlert("ERROR: new()", "TStringList::Add");
			return;
		}
		Tail->FW->BK=Tail;
		Tail=Tail->FW;
		Act=Tail;
		ActIndex++;
		Count++;
		Tail->Element=str;
		Tail->FW=NULL;
	}
}

TextElem* TStringList::GetElemByIndex(long i){
	if((Count==0)||(i>=Count)){
		ErrorAlert("ERROR: index exceeds bound", "TStringList::GetElemByIndex");
		return NULL;	
	}
	long t1,t2,t3,t4,s;
	TextElem* fp;
	t1=i;
	t2=Count-1-i;
	t3=i-ActIndex;
	t4=t3;if(t4<0)t4=(-1)*t4;
	if((t1<=t2)&&(t1<=t4)){
		fp=Root;
		for(s=0;s<t1;s++){
			fp=fp->FW;
		}
	}else if((t2<=t1)&&(t2<=t4)){
		fp=Tail;
		for(s=0;s<t2;s++){
			fp=fp->BK;
		}
	}else{
		fp=Act;
		if(t3<0){
			for(s=0;s<t4;s++)
				fp=fp->BK;
		}
		if(t3>0){
			for(s=0;s<t4;s++)
				fp=fp->FW;
		}
	}
	
	Act=fp;
	ActIndex=i;
	return fp;
}

AnsiString& StringElem::operator[] (long i){
	TextElem* tmp=Own->GetElemByIndex(i);
	return tmp->Element;
}

AnsiString& StringElem::operator[] (unsigned i){
	TextElem* tmp=Own->GetElemByIndex((long)i);
	return tmp->Element;
}

AnsiString& StringElem::operator[] (int i){
	TextElem* tmp=Own->GetElemByIndex((long)i);
	return tmp->Element;
}
