//Library code for open and save .BMP files

#include <windows.h>
#include <stdio.h>

BYTE* ConvertRGBToBMPBuffer ( BYTE* Buffer, int width, int height, long* newsize ){
	if ( ( NULL == Buffer ) || ( width == 0 ) || ( height == 0 ) )
		return NULL;
	int padding = 0;
	int scanlinebytes = width * 3;
	while ( ( scanlinebytes + padding ) % 4 != 0 )
		padding++;
	int psw = scanlinebytes + padding;
	*newsize = height * psw;
	BYTE* newbuf = new BYTE[*newsize];
	memset ( newbuf, 0, *newsize );
	long bufpos = 0;   
	long newpos = 0;
	for ( int y = 0; y < height; y++ )
		for ( int x = 0; x < 3 * width; x+=3 ){
			bufpos = y * 3 * width + x;
			newpos = ( height - y - 1 ) * psw + x;

			newbuf[newpos] = Buffer[bufpos+2];
			newbuf[newpos + 1] = Buffer[bufpos + 1];
			newbuf[newpos + 2] = Buffer[bufpos];
		}
	return newbuf;
}

BYTE* ConvertBMPToRGBBuffer ( BYTE* Buffer, int width, int height ){
	if ( ( NULL == Buffer ) || ( width == 0 ) || ( height == 0 ) )
		return NULL;
	int padding = 0;
	int scanlinebytes = width * 3;
	while ( ( scanlinebytes + padding ) % 4 != 0 )     // DWORD = 4 bytes
		padding++;
	int psw = scanlinebytes + padding;
	BYTE* newbuf = new BYTE[width*height*3];
	long bufpos = 0;   
	long newpos = 0;
	for ( int y = 0; y < height; y++ )
		for ( int x = 0; x < 3 * width; x+=3 ){
			newpos = y * 3 * width + x;     
			bufpos = ( height - y - 1 ) * psw + x;

			newbuf[newpos] = Buffer[bufpos + 2];       
			newbuf[newpos + 1] = Buffer[bufpos+1]; 
			newbuf[newpos + 2] = Buffer[bufpos];     
		}
	return newbuf;
}

bool LoadBMPIntoDC ( HDC hDC, LPCTSTR bmpfile ){
	if ( ( NULL == hDC  ) || ( NULL == bmpfile ) )
		return false;      
	HANDLE hBmp = LoadImage ( NULL, bmpfile, IMAGE_BITMAP, 0, 0,LR_LOADFROMFILE );
	if ( NULL == hBmp )
		return false;        // failed to load image
	HDC dcmem = CreateCompatibleDC ( NULL );
	if ( NULL == SelectObject ( dcmem, hBmp ) )	{	// failed to load bitmap into device context
		DeleteDC ( dcmem ); 
		return false; 
	}
	BITMAP bm;
	GetObject ( hBmp, sizeof(bm), &bm );
	if ( BitBlt ( hDC, 0, 0, bm.bmWidth, bm.bmHeight, dcmem,0, 0, SRCCOPY ) == 0 ){	// failed the blit
		DeleteDC ( dcmem ); 
		return false; 
	}
	DeleteDC ( dcmem );  // clear up the memory dc
	return true;
}

bool SaveBMP ( BYTE* Buffer, int width, int height, long paddedsize, LPCTSTR bmpfile ){
	BITMAPFILEHEADER bmfh;
	BITMAPINFOHEADER info;
	memset ( &bmfh, 0, sizeof (BITMAPFILEHEADER ) );
	memset ( &info, 0, sizeof (BITMAPINFOHEADER ) );
	bmfh.bfType = 0x4d42;       // 0x4d42 = 'BM'
	bmfh.bfReserved1 = 0;
	bmfh.bfReserved2 = 0;
	bmfh.bfSize = sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER) + paddedsize;
	bmfh.bfOffBits = 0x36;		// number of bytes to start of bitmap bits

	info.biSize = sizeof(BITMAPINFOHEADER);
	info.biWidth = width;
	info.biHeight = height;
	info.biPlanes = 1;			// we only have one bitplane
	info.biBitCount = 24;		// RGB mode is 24 bits
	info.biCompression = BI_RGB;	
	info.biSizeImage = 0;		// can be 0 for 24 bit images
	info.biXPelsPerMeter = 0x0ec4;     // paint and PSP use this values
	info.biYPelsPerMeter = 0x0ec4;     
	info.biClrUsed = 0;			// we are in RGB mode and have no palette
	info.biClrImportant = 0;    // all colors are important

	HANDLE file = CreateFile ( bmpfile , GENERIC_WRITE, FILE_SHARE_READ,
		 NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL );
	if ( file == NULL ){
		CloseHandle ( file );
		return false;
	}
	
	unsigned long bwritten;
	if ( WriteFile ( file, &bmfh, sizeof ( BITMAPFILEHEADER ), &bwritten, NULL ) == false ){
		CloseHandle ( file );
		return false;
	}
	if ( WriteFile ( file, &info, sizeof ( BITMAPINFOHEADER ), &bwritten, NULL ) == false ){
		CloseHandle ( file );
		return false;
	}
	if ( WriteFile ( file, Buffer, paddedsize, &bwritten, NULL ) == false ){
		CloseHandle ( file );
		return false;
	}
	CloseHandle ( file );

	return true;
}


BYTE* LoadBMP ( int* width, int* height, long* size, LPCTSTR bmpfile ){
	BITMAPFILEHEADER bmpheader;
	BITMAPINFOHEADER bmpinfo;
	DWORD bytesread;
	HANDLE file = CreateFile ( bmpfile , GENERIC_READ, FILE_SHARE_READ,
		 NULL, OPEN_EXISTING, FILE_FLAG_SEQUENTIAL_SCAN, NULL );
	if ( NULL == file )
		return NULL; // coudn't open file
	if ( ReadFile ( file, &bmpheader, sizeof ( BITMAPFILEHEADER ), &bytesread, NULL ) == false ){
		CloseHandle ( file );
		return NULL;
	}
	if ( ReadFile ( file, &bmpinfo, sizeof ( BITMAPINFOHEADER ), &bytesread, NULL ) == false ){
		CloseHandle ( file );
		return NULL;
	}
	if ( bmpheader.bfType != 'MB' ){
		CloseHandle ( file );
		return NULL;
	}
	*width   = bmpinfo.biWidth;
	*height  = abs ( bmpinfo.biHeight );

	if ( bmpinfo.biCompression != BI_RGB ){
		CloseHandle ( file );
		return NULL;
	}
	if ( bmpinfo.biBitCount != 24 ){
		CloseHandle ( file );
		return NULL;
	}
	*size = bmpheader.bfSize - bmpheader.bfOffBits;
	BYTE* Buffer = new BYTE[ *size ];
	SetFilePointer ( file, bmpheader.bfOffBits, NULL, FILE_BEGIN );
	if ( ReadFile ( file, Buffer, *size, &bytesread, NULL ) == false ){
		delete [] Buffer;
		CloseHandle ( file );
		return NULL;
	}
	CloseHandle ( file );
	return Buffer;
}
