//Library code for open and save .BMP files
//Linux version

#include <stdio.h>
#include "bmp_Linux.h"

BYTE* ConvertRGBToBMPBuffer ( BYTE* Buffer, int width, int height, long* newsize ){
	if ( ( NULL == Buffer ) || ( width == 0 ) || ( height == 0 ) )
		return NULL;
	int padding = 0;
	int scanlinebytes = width * 3;
	while ( ( scanlinebytes + padding ) % 4 != 0 )
		padding++;
	int psw = scanlinebytes + padding;
	*newsize = height * psw;
	BYTE* newbuf = new BYTE[*newsize];
	memset ( newbuf, 0, *newsize );
	long bufpos = 0;   
	long newpos = 0;
	for ( int y = 0; y < height; y++ )
		for ( int x = 0; x < 3 * width; x+=3 ){
			bufpos = y * 3 * width + x;
			newpos = ( height - y - 1 ) * psw + x;

			newbuf[newpos] = Buffer[bufpos+2];
			newbuf[newpos + 1] = Buffer[bufpos + 1];
			newbuf[newpos + 2] = Buffer[bufpos];
		}
	return newbuf;
}

BYTE* ConvertBMPToRGBBuffer ( BYTE* Buffer, int width, int height ){
	if ( ( NULL == Buffer ) || ( width == 0 ) || ( height == 0 ) )
		return NULL;
	int padding = 0;
	int scanlinebytes = width * 3;
	while ( ( scanlinebytes + padding ) % 4 != 0 )     // DWORD = 4 bytes
		padding++;
	int psw = scanlinebytes + padding;
	BYTE* newbuf = new BYTE[width*height*3];
	long bufpos = 0;   
	long newpos = 0;
	for ( int y = 0; y < height; y++ )
		for ( int x = 0; x < 3 * width; x+=3 ){
			newpos = y * 3 * width + x;     
			bufpos = ( height - y - 1 ) * psw + x;

			newbuf[newpos] = Buffer[bufpos + 2];       
			newbuf[newpos + 1] = Buffer[bufpos+1]; 
			newbuf[newpos + 2] = Buffer[bufpos];     
		}
	return newbuf;
}

bool SaveBMP ( BYTE* Buffer, int width, int height, long paddedsize, char* bmpfile ){
	BITMAPFILEHEADER bmfh;
	BITMAPINFOHEADER info;
	memset ( &bmfh, 0, sizeof (BITMAPFILEHEADER ) );
	memset ( &info, 0, sizeof (BITMAPINFOHEADER ) );
	bmfh.bfType = 0x4d42;       // 0x4d42 = 'BM'
	bmfh.bfReserved1 = 0;
	bmfh.bfReserved2 = 0;
	bmfh.bfSize = sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER) + paddedsize;
	bmfh.bfOffBits = 0x36;		// number of bytes to start of bitmap bits

	info.biSize = sizeof(BITMAPINFOHEADER);
	info.biWidth = width;
	info.biHeight = height;
	info.biPlanes = 1;			// we only have one bitplane
	info.biBitCount = 24;		// RGB mode is 24 bits
	info.biCompression = BI_RGB;	
	info.biSizeImage = 0;		// can be 0 for 24 bit images
	info.biXPelsPerMeter = 0x0ec4;     // paint and PSP use this values
	info.biYPelsPerMeter = 0x0ec4;     
	info.biClrUsed = 0;			// we are in RGB mode and have no palette
	info.biClrImportant = 0;    // all colors are important

    FILE* file=fopen(bmpfile,"wb");
	if ( file == NULL ){
		return false;
	}
	
	if(fwrite(&bmfh, sizeof(BITMAPFILEHEADER), 1, file)<sizeof(BITMAPFILEHEADER)){
		fclose( file );
		return false;
	}
	if(fwrite(&info, sizeof(BITMAPINFOHEADER), 1, file)<sizeof(BITMAPINFOHEADER)){
		fclose( file );
		return false;
	}
	if(fwrite(Buffer, paddedsize, 1, file)<paddedsize){
		fclose( file );
		return false;
	}
	fclose( file );

	return true;
}


BYTE* LoadBMP ( int* width, int* height, long* size, char* bmpfile ){
	BITMAPFILEHEADER bmpheader;
	BITMAPINFOHEADER bmpinfo;
	FILE* file=fopen(bmpfile,"rb");
	if ( NULL == file )
		return NULL;
	if(fread(&bmpheader,sizeof(BITMAPFILEHEADER),1,file)<sizeof(BITMAPFILEHEADER)){
		fclose ( file );
		return NULL;
	}
	if(fread(&bmpinfo,sizeof(BITMAPINFOHEADER),1,file)<sizeof(BITMAPINFOHEADER)){
		fclose ( file );
		return NULL;
	}
	if ( bmpheader.bfType != 'MB' ){
		fclose ( file );
		return NULL;
	}
	*width   = bmpinfo.biWidth;
	*height  = abs ( bmpinfo.biHeight );

	if ( bmpinfo.biCompression != BI_RGB ){
		fclose ( file );
		return NULL;
	}
	if ( bmpinfo.biBitCount != 24 ){
		fclose ( file );
		return NULL;
	}
	*size = bmpheader.bfSize - bmpheader.bfOffBits;
	BYTE* Buffer = new BYTE[ *size ];
	if(fseek(file, bmpheader.bfOffBits, SEEK_SET)){
	    fclose ( file );
		return NULL;
	}
	if(fread(Buffer,*size,1,file)<*size){
		delete [] Buffer;
		fclose ( file );
		return NULL;
	}
	fclose ( file );
	return Buffer;
}
