//Dyndns Linux daemon
//This daemon service updates the dyndns.hu dynamic DNS server with current IP
//wrutten by Andras Bognar 2014
 
#include <stdio.h>
#include <fcntl.h>
#include <signal.h>
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>

#define RUNNING_DIR	"/bin/bamessenger/dynd"
#define LOCK_FILE	"dynd.lock"
#define LOG_FILE	"dynd.log"
#define INI_FILE	"dynd.ini"

void log_message(char *message){
    FILE *logfile;
	logfile=fopen(LOG_FILE,"a");
	if(!logfile) return;
	fprintf(logfile,"%s\n",message);
	fclose(logfile);
}

void signal_handler(int sig){
	switch(sig) {
	case SIGHUP:
		log_message("SIGNAL: hangup signal catched");
		break;
	case SIGTERM:
		log_message("SIGNAL: terminate signal catched");
		exit(0);
		break;
	}
}

int main(){
    int i,lfp;
    char str[10];
	if(getppid()==1)exit(0); /* already a daemon */
	i=fork();
	if (i<0) exit(1); /* fork error */
	if (i>0) exit(0); /* parent exits */
	/* child (daemon) continues */
	if(setsid()<0)exit(1); /* obtain a new process group */
	for (i=getdtablesize();i>=0;--i) close(i); /* close all descriptors */
	i=open("/dev/null",O_RDWR); dup(i); dup(i); /* handle standart I/O */
	umask(027); /* set newly created file permissions */
	if(chdir(RUNNING_DIR)<0)exit(1); /* change running directory */
	lfp=open(LOCK_FILE,O_RDWR|O_CREAT,0640);
	if (lfp<0) exit(1); /* can not open */
	if (lockf(lfp,F_TLOCK,0)<0) exit(0); /* can not lock */
	/* first instance continues */
	sprintf(str,"%d\n",getpid());
	write(lfp,str,strlen(str)); /* record pid to lockfile */
	signal(SIGCHLD,SIG_IGN); /* ignore child */
	signal(SIGTSTP,SIG_IGN); /* ignore tty signals */
	signal(SIGTTOU,SIG_IGN);
	signal(SIGTTIN,SIG_IGN);
	signal(SIGHUP,signal_handler); /* catch hangup signal */
	signal(SIGTERM,signal_handler); /* catch kill signal */

    //initialize
    char cmd[200];
    log_message("STARTUP: Starting dyndnsd service");
    FILE *f;
	f=fopen(INI_FILE,"r");
	if(!f){
	    log_message("FATAL: could not open INI file");
	    exit(0);
	}
	if(!fgets(cmd,200,f)){
	    log_message("FATAL: could not read INI file");
	    exit(0);
	}
	fclose(f);
	
	//Updating cycle for dyndns.hu
	while(1){
	    system(cmd);
	    sleep(3600);
	}
	
	exit(0);
	return 0;
}
